//==============================================================================
//	
//	Copyright (c) 2002-
//	Authors:
//	* Carlos S. Bederián (Universidad Nacional de Córdoba)
//	* Dave Parker <david.parker@comlab.ox.ac.uk> (University of Oxford, formerly University of Birmingham)
//	
//------------------------------------------------------------------------------
//	
//	This file is part of PRISM.
//	
//	PRISM is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//	
//	PRISM is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//	
//	You should have received a copy of the GNU General Public License
//	along with PRISM; if not, write to the Free Software Foundation,
//	Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//	
//==============================================================================

package prism;

import java.util.*;

import mtbdd.PrismMTBDD;

import jdd.*;
import parser.*;
import parser.ast.*;
import parser.type.*;
import prism.NondetModelChecker.JDDFairnessConstraint;
import sparse.PrismSparse;
import jltl2ba.APElement;
import jltl2dstar.*;

/*
 * LTL model checking functionality
 */
public class LTLModelChecker
{
	protected Prism prism;
	protected PrismLog mainLog;

	public LTLModelChecker(Prism prism) throws PrismException
	{
		this.prism = prism;
		mainLog = prism.getMainLog();
	}

	/**
	 * Extract maximal state formula from an LTL path formula, model check them (with passed in model checker) and
	 * replace them with ExpressionLabel objects L0, L1, etc. Expression passed in is modified directly, but the result
	 * is also returned. As an optimisation, model checking that results in true/false for all states is converted to an
	 * actual true/false, and duplicate results are given the same label. BDDs giving the states which satisfy each label
	 * are put into the vector  labelDDs, which should be empty when this function is called.
	 */
	public Expression checkMaximalStateFormulas(ModelChecker mc, Model model, Expression expr, Vector<JDDNode> labelDDs) throws PrismException
	{
		// A state formula
		if (expr.getType() instanceof TypeBool) {
			// Model check
			JDDNode dd = mc.checkExpressionDD(expr);
			// Detect special cases (true, false) for optimisation
			if (dd.equals(JDD.ZERO)) {
				JDD.Deref(dd);
				return Expression.False();
			}
			if (dd.equals(model.getReach())) {
				JDD.Deref(dd);
				return Expression.True();
			}
			// See if we already have an identical result
			// (in which case, reuse it)
			int i = labelDDs.indexOf(dd);
			if (i != -1) {
				JDD.Deref(dd);
				return new ExpressionLabel("L" + i);
			}
			// Otherwise, add result to list, return new label
			labelDDs.add(dd);
			return new ExpressionLabel("L" + (labelDDs.size() - 1));
		}
		// A path formula (recurse, modify, return)
		else if (expr.getType() instanceof TypePathBool) {
			if (expr instanceof ExpressionBinaryOp) {
				ExpressionBinaryOp exprBinOp = (ExpressionBinaryOp) expr;
				exprBinOp.setOperand1(checkMaximalStateFormulas(mc, model, exprBinOp.getOperand1(), labelDDs));
				exprBinOp.setOperand2(checkMaximalStateFormulas(mc, model, exprBinOp.getOperand2(), labelDDs));
			} else if (expr instanceof ExpressionUnaryOp) {
				ExpressionUnaryOp exprUnOp = (ExpressionUnaryOp) expr;
				exprUnOp.setOperand(checkMaximalStateFormulas(mc, model, exprUnOp.getOperand(), labelDDs));
			} else if (expr instanceof ExpressionTemporal) {
				ExpressionTemporal exprTemp = (ExpressionTemporal) expr;
				if (exprTemp.getOperand1() != null) {
					exprTemp.setOperand1(checkMaximalStateFormulas(mc, model, exprTemp.getOperand1(), labelDDs));
				}
				if (exprTemp.getOperand2() != null) {
					exprTemp.setOperand2(checkMaximalStateFormulas(mc, model, exprTemp.getOperand2(), labelDDs));
				}
			}
		}
		return expr;
	}

	/**
	 * Construct product of DRA and DTMC/CTMC.
	 * @param dra: The DRA
	 * @param model: The  DTMC/CTMC
	 * @param labelDDs: BDDs giving the set of states for each AP in the DRA
	 */
	public ProbModel constructProductMC(DRA dra, ProbModel model, Vector<JDDNode> labelDDs) throws PrismException
	{
		return constructProductMC(dra, model, labelDDs, null, null, true);
	}

	/**
	 * Construct product of DRA and DTMC/CTMC.
	 * @param dra: The DRA
	 * @param model: The  DTMC/CTMC
	 * @param labelDDs: BDDs giving the set of states for each AP in the DRA
	 * @param draDDRowVarsCopy: (Optionally) empty JDDVars object to obtain copy of DD row vars for DRA
	 * @param draDDColVarsCopy: (Optionally) empty JDDVars object to obtain copy of DD col vars for DRA
	 */
	public ProbModel constructProductMC(DRA dra, ProbModel model, Vector<JDDNode> labelDDs, JDDVars draDDRowVarsCopy, JDDVars draDDColVarsCopy)
			throws PrismException
	{
		return constructProductMC(dra, model, labelDDs, draDDRowVarsCopy, draDDColVarsCopy, true);
	}

	/**
	 * Construct product of DRA and DTMC/CTMC.
	 * @param dra: The DRA
	 * @param model: The  DTMC/CTMC
	 * @param labelDDs: BDDs giving the set of states for each AP in the DRA
	 * @param draDDRowVarsCopy: (Optionally) empty JDDVars object to obtain copy of DD row vars for DRA
	 * @param draDDColVarsCopy: (Optionally) empty JDDVars object to obtain copy of DD col vars for DRA
	 * @param allInit: Do we assume that all states of the original model are initial states? (see below)
	 */
	public ProbModel constructProductMC(DRA dra, ProbModel model, Vector<JDDNode> labelDDs, JDDVars draDDRowVarsCopy, JDDVars draDDColVarsCopy, boolean allInit)
			throws PrismException
	{
		// Existing model - dds, vars, etc.
		JDDVars varDDRowVars[];
		JDDVars varDDColVars[];
		JDDVars allDDRowVars;
		JDDVars allDDColVars;
		Vector<String> ddVarNames;
		VarList varList;
		// New (product) model - dds, vars, etc.
		JDDNode newTrans, newStart;
		JDDVars newVarDDRowVars[], newVarDDColVars[];
		JDDVars newAllDDRowVars, newAllDDColVars;
		Vector<String> newDDVarNames;
		VarList newVarList;
		String draVar;
		// DRA stuff
		JDDVars draDDRowVars, draDDColVars;
		// Misc
		int i, j, n;
		boolean before;

		// Get details of old model
		varDDRowVars = model.getVarDDRowVars();
		varDDColVars = model.getVarDDColVars();
		allDDRowVars = model.getAllDDRowVars();
		allDDColVars = model.getAllDDColVars();
		ddVarNames = model.getDDVarNames();
		varList = model.getVarList();

		// Create a (new, unique) name for the variable that will represent DRA states
		draVar = "_dra";
		while (varList.getIndex(draVar) != -1) {
			draVar = "_" + draVar;
		}

		// See how many new dd vars will be needed for DRA
		// and whether there is room to put them before rather than after the existing vars
		// (if DRA only has one state, we add an extra dummy state)
		n = (int) Math.ceil(PrismUtils.log2(dra.size()));
		n = Math.max(n, 1);
		before = true;
		if (allDDRowVars.getMinVarIndex() - 1 < 2 * n) {
			before = false;
		}

		// If passed in var lists are null, create new lists
		// (which won't be accessible later in this case)
		draDDRowVars = (draDDRowVarsCopy == null) ? new JDDVars() : draDDRowVarsCopy;
		draDDColVars = (draDDColVarsCopy == null) ? new JDDVars() : draDDColVarsCopy;
		// Create the new dd variables
		newDDVarNames = new Vector<String>();
		newDDVarNames.addAll(ddVarNames);
		j = before ? allDDRowVars.getMinVarIndex() - 2 * n : model.getAllDDColVars().getMaxVarIndex() + 1;
		for (i = 0; i < n; i++) {
			draDDRowVars.addVar(JDD.Var(j++));
			draDDColVars.addVar(JDD.Var(j++));
			if (!before) {
				newDDVarNames.add("");
				newDDVarNames.add("");
			}
			newDDVarNames.set(j - 2, draVar + "." + i);
			newDDVarNames.set(j - 1, draVar + "'." + i);
		}

		// Create/populate new lists
		newVarDDRowVars = new JDDVars[varDDRowVars.length + 1];
		newVarDDColVars = new JDDVars[varDDRowVars.length + 1];
		newVarDDRowVars[before ? 0 : varDDRowVars.length] = draDDRowVars;
		newVarDDColVars[before ? 0 : varDDColVars.length] = draDDColVars;
		for (i = 0; i < varDDRowVars.length; i++) {
			newVarDDRowVars[before ? i + 1 : i] = new JDDVars();
			newVarDDColVars[before ? i + 1 : i] = new JDDVars();
			newVarDDRowVars[before ? i + 1 : i].addVars(varDDRowVars[i]);
			newVarDDColVars[before ? i + 1 : i].addVars(varDDColVars[i]);
		}
		newAllDDRowVars = new JDDVars();
		newAllDDColVars = new JDDVars();
		if (before) {
			newAllDDRowVars.addVars(draDDRowVars);
			newAllDDColVars.addVars(draDDColVars);
			newAllDDRowVars.addVars(allDDRowVars);
			newAllDDColVars.addVars(allDDColVars);
		} else {
			newAllDDRowVars.addVars(allDDRowVars);
			newAllDDColVars.addVars(allDDColVars);
			newAllDDRowVars.addVars(draDDRowVars);
			newAllDDColVars.addVars(draDDColVars);
		}
		newVarList = (VarList) varList.clone();
		// NB: if DRA only has one state, we add an extra dummy state
		Declaration decl = new Declaration(draVar, new DeclarationInt(Expression.Int(0), Expression.Int(Math.max(dra.size() - 1, 1))));
		newVarList.addVar(before ? 0 : varList.getNumVars(), decl, 1, model.getConstantValues());

		// Extra references (because will get derefed when new model is done with)
		allDDRowVars.refAll();
		allDDRowVars.refAll();
		allDDColVars.refAll();
		allDDColVars.refAll();
		for (i = 0; i < model.getNumModules(); i++) {
			model.getModuleDDRowVars(i).refAll();
			model.getModuleDDColVars(i).refAll();
		}
		draDDRowVars.refAll();
		draDDColVars.refAll();

		// Build transition matrix for product
		newTrans = buildTransMask(dra, labelDDs, allDDRowVars, allDDColVars, draDDRowVars, draDDColVars);
		JDD.Ref(model.getTrans());
		newTrans = JDD.Apply(JDD.TIMES, model.getTrans(), newTrans);

		// Build set of initial states for product
		// Note, by default, we take product of *all* states of the original model, not just its initial states.
		// Initial states are only used for reachability in this instance.
		// We need to ensure that the product model includes states corresponding to all
		// states of the original model (because we compute probabilities for all of them)
		// but some of these may not be reachable from the initial state of the product model.
		// Optionally (if allInit is false), we don't do this - maybe because we only care about result for the initial state
		newStart = buildStartMask(dra, labelDDs, draDDRowVars);
		JDD.Ref(allInit ? model.getReach() : model.getStart());
		newStart = JDD.And(allInit ? model.getReach() : model.getStart(), newStart);

		// Create a new model model object to store the product model
		ProbModel modelProd = new ProbModel(
		// New transition matrix/start state
				newTrans, newStart,
				// Don't pass in any rewards info
				new JDDNode[0], new JDDNode[0], new String[0],
				// New list of all row/col vars
				newAllDDRowVars, newAllDDColVars,
				// New list of var names
				newDDVarNames,
				// Module info (unchanged)
				model.getNumModules(), model.getModuleNames(), model.getModuleDDRowVars(), model.getModuleDDColVars(),
				// New var info
				model.getNumVars() + 1, newVarList, newVarDDRowVars, newVarDDColVars,
				// Constants (no change)
				model.getConstantValues());

		// Do reachability/etc. for the new model
		modelProd.doReachability(prism.getExtraReachInfo());
		modelProd.filterReachableStates();
		modelProd.findDeadlocks();
		if (modelProd.getDeadlockStates().size() > 0) {
			throw new PrismException("Model-DRA product has deadlock states");
		}

		// Reference DRA DD vars (if being returned)
		if (draDDRowVarsCopy != null)
			draDDRowVarsCopy.refAll();
		if (draDDColVarsCopy != null)
			draDDColVarsCopy.refAll();

		return modelProd;
	}

	/**
	 * Construct product of DRA and MDP.
	 * @param dra: The DRA
	 * @param model: The MDP
	 * @param labelDDs: BDDs giving the set of states for each AP in the DRA
	 */
	public NondetModel constructProductMDP(DRA dra, NondetModel model, Vector<JDDNode> labelDDs) throws PrismException
	{
		return constructProductMDP(dra, model, labelDDs, null, null, true);
	}

	/**
	 * Construct product of DRA and MDP.
	 * @param dra: The DRA
	 * @param model: The MDP
	 * @param labelDDs: BDDs giving the set of states for each AP in the DRA
	 * @param draDDRowVarsCopy: (Optionally) empty JDDVars object to obtain copy of DD row vars for DRA
	 * @param draDDColVarsCopy: (Optionally) empty JDDVars object to obtain copy of DD col vars for DRA
	 */
	public NondetModel constructProductMDP(DRA dra, NondetModel model, Vector<JDDNode> labelDDs, JDDVars draDDRowVarsCopy, JDDVars draDDColVarsCopy)
			throws PrismException
	{
		return constructProductMDP(dra, model, labelDDs, draDDRowVarsCopy, draDDColVarsCopy, true);
	}

	/**
	 * Construct product of DRA and MDP.
	 * @param dra: The DRA
	 * @param model: The MDP
	 * @param labelDDs: BDDs giving the set of states for each AP in the DRA
	 * @param draDDRowVarsCopy: (Optionally) empty JDDVars object to obtain copy of DD row vars for DRA
	 * @param draDDColVarsCopy: (Optionally) empty JDDVars object to obtain copy of DD col vars for DRA
	 * @param allInit: Do we assume that all states of the original model are initial states? (see below)
	 */
	public NondetModel constructProductMDP(DRA dra, NondetModel model, Vector<JDDNode> labelDDs, JDDVars draDDRowVarsCopy, JDDVars draDDColVarsCopy,
			boolean allInit) throws PrismException
	{
		// Existing model - dds, vars, etc.
		JDDVars varDDRowVars[];
		JDDVars varDDColVars[];
		JDDVars allDDRowVars;
		JDDVars allDDColVars;
		Vector<String> ddVarNames;
		VarList varList;
		// New (product) model - dds, vars, etc.
		JDDNode newTrans, newStart;
		JDDVars newVarDDRowVars[], newVarDDColVars[];
		JDDVars newAllDDRowVars, newAllDDColVars;
		Vector<String> newDDVarNames;
		VarList newVarList;
		String draVar;
		// DRA stuff
		JDDVars draDDRowVars, draDDColVars;
		// Misc
		int i, j, n;
		boolean before;

		// Get details of old model
		varDDRowVars = model.getVarDDRowVars();
		varDDColVars = model.getVarDDColVars();
		allDDRowVars = model.getAllDDRowVars();
		allDDColVars = model.getAllDDColVars();
		ddVarNames = model.getDDVarNames();
		varList = model.getVarList();

		// Create a (new, unique) name for the variable that will represent DRA states
		draVar = "_dra";
		while (varList.getIndex(draVar) != -1) {
			draVar = "_" + draVar;
		}

		// See how many new dd vars will be needed for DRA
		// and whether there is room to put them before rather than after the existing vars
		// (if DRA only has one state, we add an extra dummy state)
		n = (int) Math.ceil(PrismUtils.log2(dra.size()));
		n = Math.max(n, 1);
		before = true;
		if ((allDDRowVars.getMinVarIndex() - model.getAllDDNondetVars().getMaxVarIndex()) - 1 < 2 * n) {
			before = false;
		}

		// If passed in var lists are null, create new lists
		// (which won't be accessible later in this case)
		draDDRowVars = (draDDRowVarsCopy == null) ? new JDDVars() : draDDRowVarsCopy;
		draDDColVars = (draDDColVarsCopy == null) ? new JDDVars() : draDDColVarsCopy;
		// Create the new dd variables
		newDDVarNames = new Vector<String>();
		newDDVarNames.addAll(ddVarNames);
		j = before ? allDDRowVars.getMinVarIndex() - 2 * n : model.getAllDDColVars().getMaxVarIndex() + 1;
		for (i = 0; i < n; i++) {
			draDDRowVars.addVar(JDD.Var(j++));
			draDDColVars.addVar(JDD.Var(j++));
			if (!before) {
				newDDVarNames.add("");
				newDDVarNames.add("");
			}
			newDDVarNames.set(j - 2, draVar + "." + i);
			newDDVarNames.set(j - 1, draVar + "'." + i);
		}

		// Create/populate new lists
		newVarDDRowVars = new JDDVars[varDDRowVars.length + 1];
		newVarDDColVars = new JDDVars[varDDRowVars.length + 1];
		newVarDDRowVars[before ? 0 : varDDRowVars.length] = draDDRowVars;
		newVarDDColVars[before ? 0 : varDDColVars.length] = draDDColVars;
		for (i = 0; i < varDDRowVars.length; i++) {
			newVarDDRowVars[before ? i + 1 : i] = new JDDVars();
			newVarDDColVars[before ? i + 1 : i] = new JDDVars();
			newVarDDRowVars[before ? i + 1 : i].addVars(varDDRowVars[i]);
			newVarDDColVars[before ? i + 1 : i].addVars(varDDColVars[i]);
		}
		newAllDDRowVars = new JDDVars();
		newAllDDColVars = new JDDVars();
		if (before) {
			newAllDDRowVars.addVars(draDDRowVars);
			newAllDDColVars.addVars(draDDColVars);
			newAllDDRowVars.addVars(allDDRowVars);
			newAllDDColVars.addVars(allDDColVars);
		} else {
			newAllDDRowVars.addVars(allDDRowVars);
			newAllDDColVars.addVars(allDDColVars);
			newAllDDRowVars.addVars(draDDRowVars);
			newAllDDColVars.addVars(draDDColVars);
		}
		newVarList = (VarList) varList.clone();
		// NB: if DRA only has one state, we add an extra dummy state
		Declaration decl = new Declaration(draVar, new DeclarationInt(Expression.Int(0), Expression.Int(Math.max(dra.size() - 1, 1))));
		newVarList.addVar(before ? 0 : varList.getNumVars(), decl, 1, model.getConstantValues());

		// Extra references (because will get derefed when new model is done with)
		allDDRowVars.refAll();
		allDDRowVars.refAll();
		allDDColVars.refAll();
		allDDColVars.refAll();
		for (i = 0; i < model.getNumModules(); i++) {
			model.getModuleDDRowVars(i).refAll();
			model.getModuleDDColVars(i).refAll();
		}
		draDDRowVars.refAll();
		draDDColVars.refAll();
		model.getAllDDSchedVars().refAll();
		model.getAllDDSynchVars().refAll();
		model.getAllDDChoiceVars().refAll();
		model.getAllDDNondetVars().refAll();

		// Build transition matrix for product
		newTrans = buildTransMask(dra, labelDDs, allDDRowVars, allDDColVars, draDDRowVars, draDDColVars);
		JDD.Ref(model.getTrans());
		newTrans = JDD.Apply(JDD.TIMES, model.getTrans(), newTrans);

		// Build set of initial states for product
		// Note, by default, we take product of *all* states of the original model, not just its initial states.
		// Initial states are only used for reachability in this instance.
		// We need to ensure that the product model includes states corresponding to all
		// states of the original model (because we compute probabilities for all of them)
		// but some of these may not be reachable from the initial state of the product model.
		// Optionally (if allInit is false), we don't do this - maybe because we only care about result for the initial state
		newStart = buildStartMask(dra, labelDDs, draDDRowVars);
		JDD.Ref(allInit ? model.getReach() : model.getStart());
		newStart = JDD.And(allInit ? model.getReach() : model.getStart(), newStart);

		// Create a new model model object to store the product model
		NondetModel modelProd = new NondetModel(
		// New transition matrix/start state
				newTrans, newStart,
				// Don't pass in any rewards info
				new JDDNode[0], new JDDNode[0], new String[0],
				// New list of all row/col vars
				newAllDDRowVars, newAllDDColVars,
				// Nondet variables (unchanged)
				model.getAllDDSchedVars(), model.getAllDDSynchVars(), model.getAllDDChoiceVars(), model.getAllDDNondetVars(),
				// New list of var names
				newDDVarNames,
				// Module info (unchanged)
				model.getNumModules(), model.getModuleNames(), model.getModuleDDRowVars(), model.getModuleDDColVars(),
				// New var info
				model.getNumVars() + 1, newVarList, newVarDDRowVars, newVarDDColVars,
				// Constants (no change)
				model.getConstantValues());

		// Copy action info MTBDD across directly
		// If present, just needs filtering to reachable states,
		// which will get done below.
		if (model.getTransActions() != null) {
			JDD.Ref(model.getTransActions());
			modelProd.setTransActions(model.getTransActions());
		}
		// Also need to copy set of action label strings
		modelProd.setSynchs(new Vector<String>(model.getSynchs()));

		// Do reachability/etc. for the new model
		modelProd.doReachability(prism.getExtraReachInfo());
		modelProd.filterReachableStates();
		modelProd.findDeadlocks();
		if (modelProd.getDeadlockStates().size() > 0) {
			throw new PrismException("Model-DRA product has deadlock states");
		}

		//try { prism.exportStatesToFile(modelProd, Prism.EXPORT_PLAIN, new java.io.File("prod.sta")); }
		//catch (java.io.FileNotFoundException e) {}

		// Reference DRA DD vars (if being returned)
		if (draDDRowVarsCopy != null)
			draDDRowVarsCopy.refAll();
		if (draDDColVarsCopy != null)
			draDDColVarsCopy.refAll();

		return modelProd;
	}

	/**
	 * Builds a mask BDD for trans (which contains transitions between every DRA state after adding draRow/ColVars) that
	 * includes only the transitions that exist in the DRA.
	 * 
	 * @return a referenced mask BDD over trans
	 */
	public JDDNode buildTransMask(DRA dra, Vector<JDDNode> labelDDs, JDDVars allDDRowVars, JDDVars allDDColVars, JDDVars draDDRowVars, JDDVars draDDColVars)
	{
		Iterator<DA_State> it;
		DA_State state;
		JDDNode draMask, label, exprBDD, transition;
		int i, n;

		draMask = JDD.Constant(0);
		// Iterate through all (states and) transitions of DRA
		for (it = dra.iterator(); it.hasNext();) {
			state = it.next();
			for (Map.Entry<APElement, DA_State> edge : state.edges().entrySet()) {
				// Build a transition label BDD for each edge
				// System.out.println(state.getName() + " to " + edge.getValue().getName() + " through " +
				// edge.getKey().toString(dra.getAPSet(), false));
				label = JDD.Constant(1);
				n = dra.getAPSize();
				for (i = 0; i < n; i++) {
					// Get the expression BDD for label i
					exprBDD = labelDDs.get(Integer.parseInt(dra.getAPSet().getAP(i).substring(1)));
					JDD.Ref(exprBDD);
					if (!edge.getKey().get(i)) {
						exprBDD = JDD.Not(exprBDD);
					}
					label = JDD.And(label, exprBDD);
				}
				// Switch label BDD to col vars
				label = JDD.PermuteVariables(label, allDDRowVars, allDDColVars);
				// Build a BDD for the edge
				transition = JDD.SetMatrixElement(JDD.Constant(0), draDDRowVars, draDDColVars, state.getName(), edge.getValue().getName(), 1);
				// Now get the conjunction of the two
				transition = JDD.And(transition, label);
				// Add edge BDD to the DRA transition mask
				draMask = JDD.Or(draMask, transition);
			}
		}

		return draMask;
	}

	/**
	 * Builds a mask BDD for start (which contains start nodes for every DRA state after adding draRowVars) that
	 * includes only the start states (s, q) such that q = delta(q_in, label(s)) in the DRA.
	 * 
	 * @return a referenced mask BDD over start
	 */
	public JDDNode buildStartMask(DRA dra, Vector<JDDNode> labelDDs, JDDVars draDDRowVars)
	{
		JDDNode startMask, label, exprBDD, dest, tmp;

		startMask = JDD.Constant(0);
		for (Map.Entry<APElement, DA_State> edge : dra.getStartState().edges().entrySet()) {
			// Build a transition label BDD for each edge
			// System.out.println("To " + edge.getValue().getName() + " through " +
			// edge.getKey().toString(dra.getAPSet(), false));
			label = JDD.Constant(1);
			for (int i = 0; i < dra.getAPSize(); i++) {
				exprBDD = labelDDs.get(Integer.parseInt(dra.getAPSet().getAP(i).substring(1)));
				JDD.Ref(exprBDD);
				if (!edge.getKey().get(i)) {
					exprBDD = JDD.Not(exprBDD);
				}
				label = JDD.Apply(JDD.TIMES, label, exprBDD);
			}
			// Build a BDD for the DRA destination state
			dest = JDD.Constant(0);
			dest = JDD.SetVectorElement(dest, draDDRowVars, edge.getValue().getName(), 1);

			// Now get the conjunction of the two
			tmp = JDD.And(dest, label);

			// Add this destination to our start mask
			startMask = JDD.Or(startMask, tmp);
		}

		return startMask;
	}

	/**
	 * Computes sets of accepting states in a DTMC/CTMC for each Rabin acceptance pair
	 * 
	 * @returns a referenced BDD of the union of all the accepting sets
	 */
	public JDDNode findAcceptingBSCCs(DRA dra, JDDVars draDDRowVars, JDDVars draDDColVars, ProbModel model) throws PrismException
	{
		SCCComputer sccComputer;
		JDDNode acceptingStates, allAcceptingStates;
		Vector<JDDNode> vectBSCCs, newVectBSCCs;
		JDDNode tmp, tmp2;
		int i, j, n;

		// Compute BSCCs for model
		sccComputer = prism.getSCCComputer(model);
		sccComputer.computeBSCCs();
		vectBSCCs = sccComputer.getVectBSCCs();
		JDD.Deref(sccComputer.getNotInBSCCs());

		allAcceptingStates = JDD.Constant(0);

		// for each acceptance pair (H_i, L_i) in the DRA, build H'_i = S x H_i
		// and compute the BSCCs in H'_i
		for (i = 0; i < dra.acceptance().size(); i++) {

			// build the acceptance vectors H_i and L_i
			JDDNode acceptanceVector_H = JDD.Constant(0);
			JDDNode acceptanceVector_L = JDD.Constant(0);
			for (j = 0; j < dra.size(); j++) {
				/*
				 * [dA97] uses Rabin acceptance pairs (H_i, L_i) such that H_i contains Inf(ρ) The intersection of
				 * Inf(ρ) and L_i is non-empty
				 * 
				 * OTOH ltl2dstar (and our port to java) uses pairs (L_i, U_i) such that The intersection of U_i and
				 * Inf(ρ) is empty (H_i = S - U_i) The intersection of L_i and Inf(ρ) is non-empty
				 */
				if (!dra.acceptance().isStateInAcceptance_U(i, j)) {
					acceptanceVector_H = JDD.SetVectorElement(acceptanceVector_H, draDDRowVars, j, 1.0);
				}
				if (dra.acceptance().isStateInAcceptance_L(i, j)) {
					acceptanceVector_L = JDD.SetVectorElement(acceptanceVector_L, draDDRowVars, j, 1.0);
				}
			}

			// Check each BSCC for inclusion in H_i states
			// i.e. restrict each BSCC to H_i states and test if unchanged
			n = vectBSCCs.size();
			newVectBSCCs = new Vector<JDDNode>();
			for (j = 0; j < n; j++) {
				tmp = vectBSCCs.get(j);
				JDD.Ref(tmp);
				JDD.Ref(tmp);
				JDD.Ref(acceptanceVector_H);
				tmp2 = JDD.And(tmp, acceptanceVector_H);
				if (tmp.equals(tmp2)) {
					newVectBSCCs.add(tmp);
				} else {
					JDD.Deref(tmp);
				}
				JDD.Deref(tmp2);
			}
			JDD.Deref(acceptanceVector_H);

			// Compute union of BSCCs which overlap with acceptanceVector_L
			acceptingStates = filteredUnion(newVectBSCCs, acceptanceVector_L);

			// Add states to our destination BDD
			allAcceptingStates = JDD.Or(allAcceptingStates, acceptingStates);
		}

		// Deref BSCCs 
		n = vectBSCCs.size();
		for (j = 0; j < n; j++) {
			JDD.Deref(vectBSCCs.get(j));
		}

		return allAcceptingStates;
	}

	/**
	 * Computes sets of accepting states in an MDP for each Rabin acceptance pair
	 * 
	 * @returns a referenced BDD of the union of all the accepting sets
	 */
	public JDDNode findAcceptingStates(NondetModel orignalmodel, DRA dra, NondetModel model, JDDVars draDDRowVars, JDDVars draDDColVars, boolean fairness)
			throws PrismException
	{
		JDDNode acceptingStates = null, allAcceptingStates, acceptanceVector_H, acceptanceVector_L, candidateStates;
		int i, j;
		StateListMTBDD vl;

		allAcceptingStates = JDD.Constant(0);

		if (dra.acceptance().size() > 1) {
			acceptanceVector_H = JDD.Constant(0);
			acceptanceVector_L = JDD.Constant(0);
			ArrayList<JDDNode> statesH = new ArrayList<JDDNode>();
			ArrayList<JDDNode> statesL = new ArrayList<JDDNode>();

			for (i = 0; i < dra.acceptance().size(); i++) {
				JDDNode tmpH = JDD.Constant(0);
				JDDNode tmpL = JDD.Constant(0);
				for (j = 0; j < dra.size(); j++) {
					if (!dra.acceptance().isStateInAcceptance_U(i, j)) {
						tmpH = JDD.SetVectorElement(tmpH, draDDRowVars, j, 1.0);
					}
					if (dra.acceptance().isStateInAcceptance_L(i, j)) {
						tmpL = JDD.SetVectorElement(tmpL, draDDRowVars, j, 1.0);
					}
				}
				statesH.add(tmpH);
				JDD.Ref(tmpH);
				acceptanceVector_H = JDD.Or(acceptanceVector_H, tmpH);
				statesL.add(tmpL);
				JDD.Ref(tmpL);
				acceptanceVector_L = JDD.Or(acceptanceVector_L, tmpL);
			}

			JDD.Ref(model.getTrans01());
			candidateStates = JDD.Apply(JDD.TIMES, model.getTrans01(), acceptanceVector_H);
			JDD.Ref(acceptanceVector_H);
			acceptanceVector_H = JDD.PermuteVariables(acceptanceVector_H, draDDRowVars, draDDColVars);
			candidateStates = JDD.Apply(JDD.TIMES, candidateStates, acceptanceVector_H);
			candidateStates = JDD.ThereExists(candidateStates, model.getAllDDColVars());
			candidateStates = JDD.ThereExists(candidateStates, model.getAllDDNondetVars());
			// find all maximal end components
			Vector<JDDNode> allecs = findEndComponents(model, candidateStates, acceptanceVector_L);

			for (i = 0; i < dra.acceptance().size(); i++) {
				// build the acceptance vectors H_i and L_i
				acceptanceVector_H = statesH.get(i);
				acceptanceVector_L = statesL.get(i);
				for (JDDNode ec : allecs) {
					// build bdd of accepting states (under H_i) in the product model
					Vector<JDDNode> ecs;
					JDD.Ref(ec);
					JDD.Ref(acceptanceVector_H);
					candidateStates = JDD.And(ec, acceptanceVector_H);
					if (candidateStates.equals(ec)) {
						//mainLog.println(" ------------- ec is not modified ------------- ");
						ecs = new Vector<JDDNode>();
						ecs.add(ec);
					} else if (candidateStates.equals(JDD.ZERO)) {
						//mainLog.println(" ------------- ec is ZERO ------------- ");
						continue;
					} else { // recompute maximal end components
						//mainLog.println(" ------------- ec is recomputed ------------- ");
						JDD.Ref(model.getTrans01());
						JDD.Ref(candidateStates);
						JDDNode newcandidateStates = JDD.Apply(JDD.TIMES, model.getTrans01(), candidateStates);
						candidateStates = JDD.PermuteVariables(candidateStates, draDDRowVars, draDDColVars);
						newcandidateStates = JDD.Apply(JDD.TIMES, candidateStates, newcandidateStates);
						newcandidateStates = JDD.ThereExists(newcandidateStates, model.getAllDDColVars());
						candidateStates = JDD.ThereExists(newcandidateStates, model.getAllDDNondetVars());
						ecs = findEndComponents(model, candidateStates, acceptanceVector_L);
					}

					// find ECs in acceptingStates that are accepting under L_i
					acceptingStates = JDD.Constant(0);
					for (JDDNode set : ecs) {
						if (JDD.AreInterecting(set, acceptanceVector_L))
							acceptingStates = JDD.Or(acceptingStates, set);
						else
							JDD.Deref(set);
					}
					allAcceptingStates = JDD.Or(allAcceptingStates, acceptingStates);
				}
				JDD.Deref(acceptanceVector_L);
				JDD.Deref(acceptanceVector_H);
			}
			for (JDDNode ec : allecs)
				JDD.Deref(ec);
		} else {
			// for each acceptance pair (H_i, L_i) in the DRA, build H'_i = S x H_i
			// and compute the maximal ECs in H'_i
			for (i = 0; i < dra.acceptance().size(); i++) {

				// build the acceptance vectors H_i and L_i
				acceptanceVector_H = JDD.Constant(0);
				acceptanceVector_L = JDD.Constant(0);
				for (j = 0; j < dra.size(); j++) {
					/*
					 * [dA97] uses Rabin acceptance pairs (H_i, L_i) such that H_i contains Inf(ρ) The intersection of
					 * Inf(ρ) and L_i is non-empty
					 * 
					 * OTOH ltl2dstar (and our port to java) uses pairs (L_i, U_i) such that The intersection of U_i and
					 * Inf(ρ) is empty (H_i = S - U_i) The intersection of L_i and Inf(ρ) is non-empty
					 */
					if (!dra.acceptance().isStateInAcceptance_U(i, j)) {
						acceptanceVector_H = JDD.SetVectorElement(acceptanceVector_H, draDDRowVars, j, 1.0);
					}
					if (dra.acceptance().isStateInAcceptance_L(i, j)) {
						acceptanceVector_L = JDD.SetVectorElement(acceptanceVector_L, draDDRowVars, j, 1.0);
					}
				}

				/*mainLog.println(" ------------- acceptanceVector_H ------------- ");
				JDD.Ref(model.getReach());
				JDD.Ref(acceptanceVector_H);
				JDDNode ttt = JDD.And(acceptanceVector_H, model.getReach());
				vl = new StateListMTBDD(ttt, model);
				vl.print(mainLog);
				JDD.Deref(ttt);*/

				// build bdd of accepting states (under H_i) in the product model
				JDD.Ref(model.getTrans01());
				JDD.Ref(acceptanceVector_H);
				candidateStates = JDD.Apply(JDD.TIMES, model.getTrans01(), acceptanceVector_H);
				acceptanceVector_H = JDD.PermuteVariables(acceptanceVector_H, draDDRowVars, draDDColVars);
				candidateStates = JDD.Apply(JDD.TIMES, candidateStates, acceptanceVector_H);
				candidateStates = JDD.ThereExists(candidateStates, model.getAllDDColVars());
				candidateStates = JDD.ThereExists(candidateStates, model.getAllDDNondetVars());

				/*mainLog.println(" ------------- acceptanceVector_L ------------- ");
				JDD.Ref(model.getReach());
				JDD.Ref(acceptanceVector_L);
				ttt = JDD.And(acceptanceVector_L, model.getReach());
				vl = new StateListMTBDD(ttt, model);
				vl.print(mainLog);
				JDD.Deref(ttt);
				mainLog.println(" ------------- candidateStates ------------- ");
				vl = new StateListMTBDD(candidateStates, model);
				vl.print(mainLog);*/

				if (fairness) {
					// compute the backward set of S x L_i
					JDD.Ref(candidateStates);
					JDDNode tmp = JDD.And(candidateStates, acceptanceVector_L);
					JDD.Ref(model.getTrans01());
					JDDNode edges = JDD.ThereExists(model.getTrans01(), model.getAllDDNondetVars());
					JDDNode filterStates = backwardSet(model, tmp, edges);

					// Filter out states that can't reach a state in L'_i
					candidateStates = JDD.And(candidateStates, filterStates);

					// Find accepting states in S x H_i
					acceptingStates = findFairECs(model, candidateStates);
				} else {
					// find ECs in acceptingStates that are accepting under L_i
					Vector<JDDNode> ecs = findEndComponents(model, candidateStates);
					/*//StateListMTBDD vl;
					int count = 0;
					for (JDDNode set : ecs) {
						mainLog.println(" ------------- set " + count + " ------------- ");
						vl = new StateListMTBDD(set, model);
						vl.print(mainLog);
						count++;
						
						JDD.Ref(set);
						JDD.Ref(model.getTrans01());
						JDDNode newEdges1 = JDD.Apply(JDD.TIMES, model.getTrans01(), set);
						JDD.Ref(set);
						newEdges1 = JDD.Apply(JDD.TIMES, newEdges1, 
								JDD.PermuteVariables(set, model.getAllDDRowVars(), model.getAllDDColVars()));
						try {
							PrismSparse.ExportMDP(newEdges1, model.getTransActions(), model.getSynchs(), 
								model.getTransSymbol(), model.getAllDDRowVars(), model.getAllDDColVars(), 
								model.getAllDDNondetVars(), model.getODD(),
								Prism.EXPORT_DOT, ("ec" + count + ".dot"));
						} catch(Exception e){};
						//JDD.Deref(set);
						JDD.Deref(newEdges1);
					}
					acceptingStates = filteredUnion(ecs, acceptanceVector_L);
					mainLog.println(" ------------- acceptingStates ------------- ");
					vl = new StateListMTBDD(acceptingStates, model);
					vl.print(mainLog);*/
					acceptingStates = filteredUnion(ecs, acceptanceVector_L);
				}

				// Add states to our destination BDD
				allAcceptingStates = JDD.Or(allAcceptingStates, acceptingStates);
			}
		}

		return allAcceptingStates;
	}

	public ArrayList<JDDNode> findFairAcceptingStates(NondetModel orignalmodel, DRA dra, NondetModel model, JDDVars draDDRowVars, JDDVars draDDColVars,
			boolean fairness, Vector<JDDFairnessConstraint> sf, Vector<JDDFairnessConstraint> wf) throws PrismException
	{
		ArrayList<JDDNode> result = new ArrayList<JDDNode>();
		JDDNode acceptingStates = null, allAcceptingStates, acceptanceVector_H, acceptanceVector_L, candidateStates;
		int i, j;
		//StateListMTBDD vl;

		allAcceptingStates = JDD.Constant(0);

		/*********************** deal with fairness constraints ***********************/
		// find all maximal end components
		JDD.Ref(model.getReach());
		Vector<JDDNode> allecs = findEndComponents(model, model.getReach());
		Vector<JDDNode> recs = new Vector<JDDNode>();

		// identify each maximal end component
		JDDNode acc = JDD.Constant(0);
		for (JDDNode set : allecs) {
			/*mainLog.println("------------- End component: -------------");
			vl = new StateListMTBDD(set, model);
			vl.print(mainLog);*/

			boolean isfair = isEcFair(model, set, sf, wf);
			if (isfair) {
				JDD.Ref(set);
				acc = JDD.Or(acc, set);
				//recs.add(set);
				//mainLog.println("------------- is a fair component! -------------");
			} else {
				//mainLog.println("------------- is not a fair component! -------------");
			}
		}
		// compute reachability from all states to fair components
		mainLog.println("\nPerforming reachability analysis to fairness components...");
		boolean isRealisable = false;
		if (!acc.equals(JDD.ZERO)) {
			//NondetModelChecker mcProduct = new NondetModelChecker(prism, model, null);
			//StateValues probsProduct = mcProduct.checkProbUntil(model.getReach(), acc, false, false);
			//probsProduct.clear();

			// We just need to compute the "no" nodes
			JDDNode no = PrismMTBDD.Prob0A(model.getTrans01(), model.getReach(), model.getAllDDRowVars(), model.getAllDDColVars(), model.getAllDDNondetVars(),
					model.getReach(), acc);
			double numno = JDD.GetNumMinterms(no, model.getAllDDRowVars().n());
			if (numno > 0) {
				mainLog.println("number of states not reaching FMEC = " + JDD.GetNumMintermsString(no, model.getAllDDRowVars().n()));
				/*mainLog.println("------------- no: -------------");
				vl = new StateListMTBDD(no, model);
				vl.print(mainLog);*/
			} else
				isRealisable = true;
			JDD.Deref(acc);

			if (isRealisable)
				mainLog.println("Congratulations! The fairness constraint is realisable.");
			else
				mainLog.println("Warning: The fairness constraint is NOT realisable!");

			if (isRealisable) {
				JDD.Deref(no);
				recs.addAll(allecs);
			} else {
				//recs.addAll(allecs);
				for (JDDNode ec : allecs) {
					if (!JDD.AreInterecting(ec, no))
						recs.add(ec);
				}
				// remove no states
				JDD.Ref(model.getReach());
				JDDNode realisablestates = JDD.And(model.getReach(), JDD.Not(no));
				// recursively remove states that do not have probability 1 to reach fair EC
				// realisablestates = findMaximalStableSet(model, realisablestates);
				JDD.Ref(realisablestates);
				JDD.Ref(model.getTrans());
				// Select transitions starting in current
				JDDNode currTrans = JDD.Apply(JDD.TIMES, model.getTrans(), realisablestates);
				// Select transitions starting in current and ending in current
				//JDD.Ref(realisablestates);
				JDDNode tmp = JDD.PermuteVariables(realisablestates, model.getAllDDRowVars(), model.getAllDDColVars());
				tmp = JDD.Apply(JDD.TIMES, currTrans, tmp);
				// Sum all successor probabilities for each (state, action) tuple
				tmp = JDD.SumAbstract(tmp, model.getAllDDColVars());
				// If the sum for a (state,action) tuple is 1,
				// there is an action that remains in the stable set with prob 1
				tmp = JDD.GreaterThan(tmp, 1 - prism.getSumRoundOff());

				JDD.Ref(model.getTrans());
				JDD.Ref(tmp);
				//JDDNode newtrans = JDD.Apply(JDD.TIMES, model.getTrans(), tmp);
				JDDNode newtrans = JDD.And(model.getTrans(), tmp);

				JDD.Ref(model.getTrans01());
				JDDNode newtrans01 = JDD.And(model.getTrans01(), tmp);

				result.add(newtrans);
				result.add(newtrans01);
				/*// get the initial state
				JDDNode initial = model.getStart();
				// build reachable states from the initial state under the new transition relation
				JDDNode statespace = initial;
				JDD.Ref(statespace);
				JDDNode frontier = initial;
				JDD.Ref(frontier);
				mainLog.println("------------- frontier: -------------");
				vl = new StateListMTBDD(frontier, model);
				vl.print(mainLog);					
				while(!frontier.equals(JDD.ZERO)) {
					JDD.Ref(newtrans);
					frontier = JDD.Apply(JDD.TIMES, newtrans, frontier);
					frontier = JDD.PermuteVariables(frontier, model.getAllDDRowVars(), model.getAllDDColVars());
					frontier = JDD.ThereExists(frontier, model.getAllDDNondetVars());
					frontier = JDD.ThereExists(frontier, model.getAllDDColVars()); // new frontier
					JDD.Ref(realisablestates);
					frontier = JDD.And(frontier, realisablestates); // remove no states
					mainLog.println("------------- frontier (1): -------------");
					vl = new StateListMTBDD(frontier, model);
					vl.print(mainLog);
					JDD.Ref(statespace);
					frontier = JDD.And(frontier, JDD.Not(statespace)); // remove states that have been visited 
					mainLog.println("------------- frontier (2): -------------");
					vl = new StateListMTBDD(frontier, model);
					vl.print(mainLog);
					JDD.Ref(frontier);
					statespace = JDD.Or(statespace, frontier); // new state space
				}
				JDD.Deref(newtrans);
				JDD.Deref(frontier);
				JDD.Deref(realisablestates);
				
				for(JDDNode ec: allecs) {
					if(JDD.AreInterecting(ec, statespace))
						recs.add(ec);
				}
				JDD.Deref(statespace); */
			}

			for (i = 0; i < dra.acceptance().size(); i++) {
				acceptanceVector_H = JDD.Constant(0);
				acceptanceVector_L = JDD.Constant(0);
				for (j = 0; j < dra.size(); j++) {
					if (!dra.acceptance().isStateInAcceptance_U(i, j)) {
						acceptanceVector_H = JDD.SetVectorElement(acceptanceVector_H, draDDRowVars, j, 1.0);
					}
					if (dra.acceptance().isStateInAcceptance_L(i, j)) {
						acceptanceVector_L = JDD.SetVectorElement(acceptanceVector_L, draDDRowVars, j, 1.0);
					}
				}
				acceptingStates = JDD.Constant(0);
				for (JDDNode ec : recs) {
					if (JDD.AreInterecting(ec, acceptanceVector_L)) {
						JDD.Ref(acceptanceVector_H);
						JDD.Ref(ec);
						JDDNode tmp = JDD.And(ec, JDD.Not(acceptanceVector_H));
						Vector<JDDNode> newecs;
						if (tmp.equals(JDD.ZERO)) {
							if (isEcFairAccept(model, ec, sf, wf, acceptanceVector_L)) {
								JDD.Ref(ec);
								acceptingStates = JDD.Or(acceptingStates, ec);
								//mainLog.println("------------- acceptance ec: -------------");
								//vl = new StateListMTBDD(ec, model);
								//vl.print(mainLog);
							}
						} else {
							JDD.Ref(acceptanceVector_H);
							JDD.Ref(ec);
							JDDNode Hstates = JDD.And(ec, acceptanceVector_H);

							JDD.Ref(model.getTrans01());
							JDD.Ref(Hstates);
							candidateStates = JDD.Apply(JDD.TIMES, model.getTrans01(), Hstates);
							Hstates = JDD.PermuteVariables(Hstates, draDDRowVars, draDDColVars);
							candidateStates = JDD.Apply(JDD.TIMES, candidateStates, Hstates);
							candidateStates = JDD.ThereExists(candidateStates, model.getAllDDColVars());
							candidateStates = JDD.ThereExists(candidateStates, model.getAllDDNondetVars());

							// find all maximal end components in ec
							newecs = findEndComponents(model, candidateStates, acceptanceVector_L);

							//JDD.Deref(candidateStates);
							for (JDDNode tmpec : newecs) {
								if (isEcFairAccept(model, tmpec, sf, wf, acceptanceVector_L)) {
									JDD.Ref(ec);
									acceptingStates = JDD.Or(acceptingStates, ec);
									//mainLog.println("------------- acceptance ec: -------------");
									//vl = new StateListMTBDD(ec, model);
									//vl.print(mainLog);
									break;
								}
							}
							for (JDDNode tmpec : newecs)
								JDD.Deref(tmpec);
						}
						JDD.Deref(tmp);
					}
				}
				JDD.Deref(acceptanceVector_H);
				JDD.Deref(acceptanceVector_L);
				allAcceptingStates = JDD.Or(allAcceptingStates, acceptingStates);
			}
		} else {
			JDD.Deref(acc);
			mainLog.println("Warning: The fairness constraint is NOT realisable!");
		}

		for (JDDNode ec : allecs)
			JDD.Deref(ec);

		// get the set of ECs that are realisable
		result.add(0, allAcceptingStates);
		return result;
	}

	/*public JDDNode findMultiAcceptingStates(DRA dra, NondetModel model, JDDVars draDDRowVars, JDDVars draDDColVars, 
			boolean fairness, Vector<JDDFairnessConstraint> sf, Vector<JDDFairnessConstraint> wf) throws PrismException
	{
		JDDNode acceptingStates, allAcceptingStates;
		JDDNode acceptanceVector_H, acceptanceVector_L;
		int i, j;
		
		allAcceptingStates = JDD.Constant(0);

		// for each acceptance pair (H_i, L_i) in the DRA, build H'_i = S x H_i
		// and compute the maximal ECs in H'_i
		for (i = 0; i < dra.acceptance().size(); i++) {

			// build the acceptance vectors H_i and L_i
			acceptanceVector_H = JDD.Constant(0);
			acceptanceVector_L = JDD.Constant(0);
			for (j = 0; j < dra.size(); j++) {
				if (!dra.acceptance().isStateInAcceptance_U(i, j)) {
					acceptanceVector_H = JDD.SetVectorElement(acceptanceVector_H, draDDRowVars, j, 1.0);
				}
				if (dra.acceptance().isStateInAcceptance_L(i, j)) {
					acceptanceVector_L = JDD.SetVectorElement(acceptanceVector_L, draDDRowVars, j, 1.0);
				}
			}

			// build bdd of accepting states (under H_i) in the product model
			JDD.Ref(model.getTrans01());
			JDD.Ref(acceptanceVector_H);
			JDDNode candidateStates = JDD.Apply(JDD.TIMES, model.getTrans01(), acceptanceVector_H);
			acceptanceVector_H = JDD.PermuteVariables(acceptanceVector_H, draDDRowVars, draDDColVars);
			JDD.Ref(acceptanceVector_H);
			candidateStates = JDD.Apply(JDD.TIMES, candidateStates, acceptanceVector_H);
			candidateStates = JDD.ThereExists(candidateStates, model.getAllDDColVars());
			candidateStates = JDD.ThereExists(candidateStates, model.getAllDDNondetVars());

			if (fairness) {
				// compute the backward set of S x L_i
				JDD.Ref(candidateStates);
				JDDNode tmp = JDD.And(candidateStates, acceptanceVector_L);
				JDD.Ref(model.getTrans01());
				JDDNode edges = JDD.ThereExists(model.getTrans01(), model.getAllDDNondetVars());
				JDDNode filterStates = backwardSet(model, tmp, edges);
				// Filter out states that can't reach a state in L'_i
				candidateStates = JDD.And(candidateStates, filterStates);
				// Find accepting states in S x H_i
				acceptingStates = findFairECs(model, candidateStates);
			} else {
				// find ECs in acceptingStates that are accepting under L_i
				//acceptingStates = filteredUnion(findEndComponents(model, candidateStates), acceptanceVector_L);
				Vector<JDDNode> ecs = findEndComponents(model, candidateStates);
				//StateListMTBDD vl;
				if(sf == null && wf == null)
					acceptingStates = filteredUnion(ecs, acceptanceVector_L);
				else { // dealing with fairness constraints
					// test if fairness constraints are realisable
					JDDNode acc = JDD.Constant(0);
					acceptingStates = JDD.Constant(0);
					for (JDDNode set : ecs) {
						/ *mainLog.println("------------- End component: -------------");
						vl = new StateListMTBDD(set, model);
						vl.print(mainLog);* /
						boolean isfair = isEcFair(model, set, sf, wf);
						if(isfair) {
							JDD.Ref(set);
							acc = JDD.Or(acc, set);
							//mainLog.println("------------- is a fair component! -------------");
							if(JDD.AreInterecting(set, acceptanceVector_L)) {
								JDD.Ref(set);
								acceptingStates = JDD.Or(acceptingStates, set);
							}
						} else {
							//mainLog.println("------------- is not a fair component! -------------");
						}
						JDD.Deref(set);
					}
					// compute reachability from all states to fair components
					mainLog.println("\nPerforming reachability analysis to fairness components...");
					boolean isRealisable = false;
					if(!acc.equals(JDD.ZERO)) {						
						NondetModelChecker mcProduct = new NondetModelChecker(prism, model, null);
						//probsProduct.clear();
						
						// We just need to compute the "no" nodes
						JDDNode no = PrismMTBDD.Prob0A(model.getTrans01(), model.getReach(), model.getAllDDRowVars(), 
								model.getAllDDColVars(), model.getAllDDNondetVars(), model.getReach(), acc);
						double numno = JDD.GetNumMinterms(no, model.getAllDDRowVars().n());
						if(numno > 0)
							mainLog.println("no = " + JDD.GetNumMintermsString(no, model.getAllDDRowVars().n()));
						else
							isRealisable = true;
						JDD.Deref(no);
					}
					JDD.Deref(acc);
					if(isRealisable)
						mainLog.println("Congratulations! The fairness constraint is realisable.");
					else
						mainLog.println("Warning: The fairness constraint is NOT realisable!");
					// get fair end components
					JDD.Deref(acceptanceVector_L);
				}
				// Deref (others are deref-ed elsewhere, e.g. by call to filteredUnion)
				JDD.Deref(acceptanceVector_H);
			}

			// Add states to our destination BDD
			allAcceptingStates = JDD.Or(allAcceptingStates, acceptingStates);
		}
		
		return allAcceptingStates;
	}*/

	public JDDNode findMultiAcceptingStates(DRA dra, NondetModel model, JDDVars draDDRowVars, JDDVars draDDColVars, boolean fairness, Vector<JDDNode> allecs,
			ArrayList<JDDNode> statesH, ArrayList<JDDNode> statesL) throws PrismException
	{
		JDDNode acceptingStates = null, allAcceptingStates, candidateStates;
		JDDNode acceptanceVector_H, acceptanceVector_L;
		int i;

		allAcceptingStates = JDD.Constant(0);

		// for each acceptance pair (H_i, L_i) in the DRA, build H'_i = S x H_i
		// and compute the maximal ECs in H'_i
		for (i = 0; i < dra.acceptance().size(); i++) {
			// build the acceptance vectors H_i and L_i
			acceptanceVector_H = statesH.get(i);
			acceptanceVector_L = statesL.get(i);
			for (JDDNode ec : allecs) {
				// build bdd of accepting states (under H_i) in the product model
				Vector<JDDNode> ecs = null;
				JDD.Ref(ec);
				JDD.Ref(acceptanceVector_H);
				candidateStates = JDD.And(ec, acceptanceVector_H);
				if (candidateStates.equals(ec)) {
					//mainLog.println(" ------------- ec is not modified ------------- ");
					ecs = new Vector<JDDNode>();
					//JDDNode ec1 = ec;
					//JDD.Ref(ec);
					ecs.add(ec);
					//JDD.Deref(candidateStates);
				} else if (candidateStates.equals(JDD.ZERO)) {
					//mainLog.println(" ------------- ec is ZERO ------------- ");
					JDD.Deref(candidateStates);
					continue;
				} else { // recompute maximal end components
					//mainLog.println(" ------------- ec is recomputed ------------- ");
					JDD.Ref(model.getTrans01());
					JDD.Ref(candidateStates);
					JDDNode newcandidateStates = JDD.Apply(JDD.TIMES, model.getTrans01(), candidateStates);
					candidateStates = JDD.PermuteVariables(candidateStates, draDDRowVars, draDDColVars);
					newcandidateStates = JDD.Apply(JDD.TIMES, candidateStates, newcandidateStates);
					newcandidateStates = JDD.ThereExists(newcandidateStates, model.getAllDDColVars());
					candidateStates = JDD.ThereExists(newcandidateStates, model.getAllDDNondetVars());
					ecs = findEndComponents(model, candidateStates, acceptanceVector_L);
				}

				//StateListMTBDD vl;
				//int count = 0;
				acceptingStates = JDD.Constant(0);
				for (JDDNode set : ecs) {
					if (JDD.AreInterecting(set, acceptanceVector_L))
						acceptingStates = JDD.Or(acceptingStates, set);
					else
						JDD.Deref(set);
				}
				// Add states to our destination BDD
				allAcceptingStates = JDD.Or(allAcceptingStates, acceptingStates);
			}
			JDD.Deref(acceptanceVector_L);
			JDD.Deref(acceptanceVector_H);
		}

		return allAcceptingStates;
	}

	public void findMultiConflictAcceptingStates(DRA[] dra, NondetModel model, JDDVars[] draDDRowVars, JDDVars[] draDDColVars, List<JDDNode> targetDDs,
			List<List<JDDNode>> allstatesH, List<List<JDDNode>> allstatesL, List<JDDNode> combinations, List<List<Integer>> combinationIDs)
			throws PrismException
	{
		List<queueElement> queue = new ArrayList<queueElement>();
		int sp = 0;

		for (int i = 0; i < dra.length; i++) {
			List<Integer> ids = new ArrayList<Integer>();
			ids.add(i);
			queueElement e = new queueElement(allstatesH.get(i), allstatesL.get(i), targetDDs.get(i), ids, i + 1);
			queue.add(e);
		}

		while (sp < queue.size()) {
			computeCombinations(dra, model, draDDRowVars, draDDColVars, targetDDs, allstatesH, allstatesL, queue, sp);
			sp++;
		}

		// subtract children from targetDD
		for (queueElement e : queue)
			if (e.children != null) {
				JDDNode newtarget = e.targetDD;
				//JDD.Ref(newtarget);
				for (queueElement e1 : e.children) {
					JDD.Ref(e1.targetDD);
					newtarget = JDD.And(newtarget, JDD.Not(e1.targetDD));
				}
				//JDD.Deref(e.targetDD);
				e.targetDD = newtarget;
			}
		targetDDs.clear();
		for (int i = 0; i < dra.length; i++) {
			targetDDs.add(queue.get(i).targetDD);
		}
		for (int i = dra.length; i < queue.size(); i++) {
			combinations.add(queue.get(i).targetDD);
			combinationIDs.add(queue.get(i).draIDs);
		}
	}

	private void computeCombinations(DRA[] dra, NondetModel model, JDDVars[] draDDRowVars, JDDVars[] draDDColVars, List<JDDNode> targetDDs,
			List<List<JDDNode>> allstatesH, List<List<JDDNode>> allstatesL, List<queueElement> queue, int sp) throws PrismException
	{
		queueElement e = queue.get(sp);
		int bound = queue.size();
		//StateListMTBDD vl = null;
		//mainLog.println("  ------------- Processing " + e.draIDs + ": -------------");

		for (int i = e.next; i < dra.length; i++) {
			List<JDDNode> newstatesH = new ArrayList<JDDNode>();
			List<JDDNode> newstatesL = new ArrayList<JDDNode>();
			//if(e.draIDs.size() >= 2 || sp > 0 /*|| queue.size() > 3*/)
			//	break;
			//mainLog.println("             combinations " + e.draIDs + ", " + i + ": ");
			JDDNode allAcceptingStates = JDD.Constant(0);
			// compute conjunction of e and next
			List<JDDNode> nextstatesH = allstatesH.get(i);
			List<JDDNode> nextstatesL = allstatesL.get(i);
			JDD.Ref(e.targetDD);
			JDD.Ref(targetDDs.get(i));
			JDDNode intersection = JDD.And(e.targetDD, targetDDs.get(i));
			for (int j = 0; j < e.statesH.size(); j++) {
				JDD.Ref(intersection);
				JDD.Ref(e.statesH.get(j));
				JDDNode candidateStates = JDD.And(intersection, e.statesH.get(j));
				for (int k = 0; k < nextstatesH.size(); k++) {
					JDD.Ref(candidateStates);
					JDD.Ref(nextstatesH.get(k));
					JDDNode candidateStates1 = JDD.And(candidateStates, nextstatesH.get(k));

					// Find end components in candidateStates1
					JDD.Ref(model.getTrans01());
					JDD.Ref(candidateStates1);
					JDDNode newcandidateStates = JDD.Apply(JDD.TIMES, model.getTrans01(), candidateStates1);
					/*for(int x=0; x<e.draIDs.size(); x++)
						candidateStates1 = JDD.PermuteVariables(candidateStates1, draDDRowVars[e.draIDs.get(x)], draDDColVars[e.draIDs.get(x)]);
					candidateStates1 = JDD.PermuteVariables(candidateStates1, draDDRowVars[i], draDDColVars[i]);*/
					candidateStates1 = JDD.PermuteVariables(candidateStates1, model.getAllDDRowVars(), model.getAllDDColVars());
					newcandidateStates = JDD.Apply(JDD.TIMES, candidateStates1, newcandidateStates);
					newcandidateStates = JDD.ThereExists(newcandidateStates, model.getAllDDColVars());
					candidateStates1 = JDD.ThereExists(newcandidateStates, model.getAllDDNondetVars());
					JDD.Ref(e.statesL.get(j));
					JDD.Ref(nextstatesL.get(k));
					JDDNode acceptanceVector_L = JDD.And(e.statesL.get(j), nextstatesL.get(k));
					Vector<JDDNode> ecs = null;
					ecs = findEndComponents(model, candidateStates1, acceptanceVector_L);

					// For each ec, test if it has non-empty intersection with L states
					if (ecs != null) {
						boolean valid = false;
						for (JDDNode set : ecs) {
							if (JDD.AreInterecting(set, acceptanceVector_L)) {
								allAcceptingStates = JDD.Or(allAcceptingStates, set);
								valid = true;
							} else
								JDD.Deref(set);
						}
						if (valid) {
							//mainLog.println("          adding j = " + j + ", k = " + k + " to nextstateH & L ");
							JDD.Ref(e.statesH.get(j));
							JDD.Ref(nextstatesH.get(k));
							JDDNode ttt = JDD.And(e.statesH.get(j), nextstatesH.get(k));
							newstatesH.add(ttt);
							JDD.Ref(acceptanceVector_L);
							newstatesL.add(acceptanceVector_L);
						}
					}
					//if(!valid)
					JDD.Deref(acceptanceVector_L);
				}
				JDD.Deref(candidateStates);
			}
			JDD.Deref(intersection);

			if (!newstatesH.isEmpty() /*&& i+1 < dra.length*/) {
				// generate a new element and put it into queue
				List<Integer> ids = new ArrayList<Integer>(e.draIDs);
				ids.add(i);
				queueElement e1 = new queueElement(newstatesH, newstatesL, allAcceptingStates, ids, i + 1);
				queue.add(e1);
				// add link to e
				e.addChildren(e1);
			} else
				JDD.Deref(allAcceptingStates);

			/*String s = "";
			for(int j=0; j<e.draIDs.size(); j++) 
				s += e.draIDs.*/
			/*vl = new StateListMTBDD(allAcceptingStates, model);
			vl.print(mainLog);
			mainLog.flush();*/
		}

		// add children generated by other elements to e
		for (int i = bound - 1; i > sp; i--) {
			queueElement e2 = queue.get(i);
			if (e2.draIDs.size() <= e.draIDs.size())
				break;
			if (e2.draIDs.containsAll(e.draIDs))
				e.addChildren(e2);
		}

		if (e.draIDs.size() > 1) {
			//mainLog.println("          releaseing statesH & L ");
			for (int i = 0; i < e.statesH.size(); i++) {
				JDD.Deref(e.statesH.get(i));
				JDD.Deref(e.statesL.get(i));
			}
		}

	}

	public JDDNode findFairMultiAcceptingStates(DRA dra, NondetModel model, JDDVars draDDRowVars, JDDVars draDDColVars, boolean fairness,
			Vector<JDDFairnessConstraint> sf, Vector<JDDFairnessConstraint> wf, Vector<JDDNode> allecs, ArrayList<JDDNode> statesH, ArrayList<JDDNode> statesL)
			throws PrismException
	{
		JDDNode acceptingStates = null, allAcceptingStates, candidateStates;
		JDDNode acceptanceVector_H, acceptanceVector_L;
		int i;

		allAcceptingStates = JDD.Constant(0);
		for (i = 0; i < dra.acceptance().size(); i++) {
			acceptanceVector_H = statesH.get(i);
			acceptanceVector_L = statesL.get(i);
			acceptingStates = JDD.Constant(0);
			for (JDDNode ec : allecs) {
				if (JDD.AreInterecting(ec, acceptanceVector_L)) {
					JDD.Ref(acceptanceVector_H);
					JDD.Ref(ec);
					JDDNode tmp = JDD.And(ec, JDD.Not(acceptanceVector_H));
					if (tmp.equals(JDD.ZERO)) {
						JDD.Ref(ec);
						acceptingStates = JDD.Or(acceptingStates, ec);
						/*mainLog.println("------------- acceptance ec: -------------");
						vl = new StateListMTBDD(ec, model);
						vl.print(mainLog);*/
					} else {
						JDD.Ref(acceptanceVector_H);
						JDD.Ref(ec);
						candidateStates = JDD.And(ec, acceptanceVector_H);
						// find all maximal end components in ec
						Vector<JDDNode> newecs = findEndComponents(model, candidateStates, acceptanceVector_L);
						//JDD.Deref(candidateStates);
						for (JDDNode tmpec : newecs) {
							if (isEcFairAccept(model, tmpec, sf, wf, acceptanceVector_L/*, this*/)) {
								JDD.Ref(ec);
								acceptingStates = JDD.Or(acceptingStates, ec);
								/*mainLog.println("------------- acceptance ec: -------------");
								vl = new StateListMTBDD(ec, model);
								vl.print(mainLog);*/
								break;
							}
						}
						for (JDDNode tmpec : newecs)
							JDD.Deref(tmpec);
					}
					JDD.Deref(tmp);
				}
			}
			JDD.Deref(acceptanceVector_H);
			JDD.Deref(acceptanceVector_L);
			allAcceptingStates = JDD.Or(allAcceptingStates, acceptingStates);
		}

		return allAcceptingStates;
	}

	/**
	 * Returns all end components in candidateStates under fairness assumptions
	 * 
	 * @param candidateStates
	 *            set of candidate states S x H_i (dereferenced after calling this function)
	 * @return a referenced BDD with the maximal stable set in c
	 */
	private JDDNode findFairECs(NondetModel model, JDDNode candidateStates)
	{

		JDDNode old = JDD.Constant(0);
		JDDNode current = candidateStates;

		while (!current.equals(old)) {
			JDD.Deref(old);
			JDD.Ref(current);
			old = current;

			JDD.Ref(current);
			JDD.Ref(model.getTrans01());
			// Select transitions starting in current
			JDDNode currTrans01 = JDD.And(model.getTrans01(), current);
			JDD.Ref(current);
			// mask of transitions that end outside current
			JDDNode mask = JDD.Not(JDD.PermuteVariables(current, model.getAllDDRowVars(), model.getAllDDColVars()));
			mask = JDD.And(currTrans01, mask);
			// mask of states that have bad transitions
			mask = JDD.ThereExists(mask, model.getAllDDColVars());
			mask = JDD.ThereExists(mask, model.getAllDDNondetVars());
			// Filter states with bad transitions
			current = JDD.And(current, JDD.Not(mask));
		}
		JDD.Deref(old);
		return current;
	}

	/**
	 * Return the set of states that reach nodes through edges Refs: result Derefs: nodes, edges
	 */
	private JDDNode backwardSet(NondetModel model, JDDNode nodes, JDDNode edges)
	{
		JDDNode old = JDD.Constant(0);
		JDDNode current = nodes;
		while (!current.equals(old)) {
			JDD.Deref(old);
			JDD.Ref(current);
			old = current;
			JDD.Ref(current);
			JDD.Ref(edges);
			current = JDD.Or(current, preimage(model, current, edges));
		}
		JDD.Deref(edges);
		JDD.Deref(old);
		return current;
	}

	/**
	 * Return the preimage of nodes in edges Refs: result Derefs: edges, nodes
	 */
	// FIXME: Refactor this out (duplicated in SCCComputers)
	private JDDNode preimage(NondetModel model, JDDNode nodes, JDDNode edges)
	{
		JDDNode tmp;

		// Get transitions that end at nodes
		tmp = JDD.PermuteVariables(nodes, model.getAllDDRowVars(), model.getAllDDColVars());
		tmp = JDD.And(edges, tmp);
		// Get pre(nodes)
		tmp = JDD.ThereExists(tmp, model.getAllDDColVars());
		return tmp;
	}

	/**
	 * Computes maximal accepting end components in states
	 * 
	 * @param states
	 *            BDD of a set of states (dereferenced after calling this function)
	 * @return a vector of referenced BDDs containing all the ECs in states
	 */
	public Vector<JDDNode> findEndComponents(NondetModel model, JDDNode states) throws PrismException
	{

		//boolean initialCandidate = true;
		Stack<JDDNode> candidates = new Stack<JDDNode>();
		candidates.push(states);
		Vector<JDDNode> ecs = new Vector<JDDNode>();
		SCCComputer sccComputer;

		while (!candidates.isEmpty()) {
			JDDNode candidate = candidates.pop();

			// Compute the stable set
			JDD.Ref(candidate);
			JDDNode stableSet = findMaximalStableSet(model, candidate);

			// Drop empty sets
			if (stableSet.equals(JDD.ZERO)) {
				JDD.Deref(stableSet);
				JDD.Deref(candidate);
				continue;
			}

			/*if (!initialCandidate) {
				// candidate is an SCC, check if it's stable
				if (stableSet.equals(candidate)) {
					ecs.add(candidate);
					JDD.Deref(stableSet);
					continue;
				}
			} else {
				initialCandidate = false;
			}
			JDD.Deref(candidate);*/

			if (stableSet.equals(candidate) && JDD.GetNumMinterms(stableSet, model.getNumDDRowVars()) == 1) {
				ecs.add(candidate);
				JDD.Deref(stableSet);
				continue;
			}

			// Filter bad transitions
			JDD.Ref(stableSet);
			JDDNode stableSetTrans = maxStableSetTrans(model, stableSet);

			// now find the maximal SCCs in (stableSet, stableSetTrans)
			Vector<JDDNode> sccs;
			sccComputer = prism.getSCCComputer(stableSet, stableSetTrans, model.getAllDDRowVars(), model.getAllDDColVars());
			sccComputer.computeSCCs();
			JDD.Deref(stableSet);
			JDD.Deref(stableSetTrans);
			sccs = sccComputer.getVectSCCs();
			//mainLog.println("SCC computer found " + sccs.size() + " SCCs");
			JDD.Deref(sccComputer.getNotInSCCs());
			if (sccs.size() > 0) {
				if (sccs.size() > 1 || !sccs.get(0).equals(candidate)) {
					candidates.addAll(sccs);
					JDD.Deref(candidate);
				} else {
					ecs.add(candidate);
					JDD.Deref(sccs.get(0));
				}
			} else
				JDD.Deref(candidate);
		}
		return ecs;
	}

	public Vector<JDDNode> findEndComponents(NondetModel model, JDDNode states, JDDNode filter) throws PrismException
	{

		//boolean initialCandidate = true;
		Stack<JDDNode> candidates = new Stack<JDDNode>();
		candidates.push(states);
		Vector<JDDNode> ecs = new Vector<JDDNode>();
		SCCComputer sccComputer;

		//int round = 0;
		while (!candidates.isEmpty()) {
			//System.out.println("round = " + round++);
			JDDNode candidate = candidates.pop();
			
			// Compute the stable set
			JDD.Ref(candidate);			
			JDDNode stableSet = findMaximalStableSet(model, candidate);
			// Drop empty sets
			if (stableSet.equals(JDD.ZERO)) {
				JDD.Deref(stableSet);
				JDD.Deref(candidate);
				continue;
			}

			/*if (!initialCandidate) {
				// candidate is an SCC, check if it's stable
				if (stableSet.equals(candidate)) {
					ecs.add(candidate);
					JDD.Deref(stableSet);
					continue;
				}
			} else {
				initialCandidate = false;
			}
			JDD.Deref(candidate);*/

			if (stableSet.equals(candidate) && JDD.GetNumMinterms(stableSet, model.getNumDDRowVars()) == 1) {
				ecs.add(candidate);
				JDD.Deref(stableSet);
				continue;
			}

			// Filter bad transitions
			JDD.Ref(stableSet);
			JDDNode stableSetTrans = maxStableSetTrans(model, stableSet);

			// now find the maximal SCCs in (stableSet, stableSetTrans)
			Vector<JDDNode> sccs;
			sccComputer = prism.getSCCComputer(stableSet, stableSetTrans, model.getAllDDRowVars(), model.getAllDDColVars());
			//System.out.println("Computing SCCs...");
			sccComputer.computeSCCs(filter);			
			JDD.Deref(stableSet);
			JDD.Deref(stableSetTrans);
			sccs = sccComputer.getVectSCCs();
			//mainLog.println("SCC computer found " + sccs.size() + " SCCs");
			JDD.Deref(sccComputer.getNotInSCCs());
			if(sccs.size() > 0) {
				if(sccs.size() > 1 || !sccs.get(0).equals(candidate)) {
					candidates.addAll(sccs);
					JDD.Deref(candidate);
				} else {
					ecs.add(candidate);
					JDD.Deref(sccs.get(0));
				}
			} else
				JDD.Deref(candidate);
		}
		return ecs;
	}

	public Vector<JDDNode> findBottomEndComponents(NondetModel model, JDDNode states) throws PrismException
	{
		Vector<JDDNode> ecs = findEndComponents(model, states);
		Vector<JDDNode> becs = new Vector<JDDNode>();
		JDDNode out;

		for (JDDNode scc : ecs) {
			JDD.Ref(model.getTrans01());
			JDD.Ref(scc);
			out = JDD.And(model.getTrans01(), scc);
			JDD.Ref(scc);
			out = JDD.And(out, JDD.Not(JDD.PermuteVariables(scc, model.getAllDDRowVars(), model.getAllDDColVars())));
			if (out.equals(JDD.ZERO)) {
				becs.addElement(scc);
			} else {
				JDD.Deref(scc);
			}
			JDD.Deref(out);
		}
		return becs;
	}

	/**
	 * Returns a stable set of states contained in candidateStates
	 * 
	 * @param candidateStates
	 *            set of candidate states S x H_i (dereferenced after calling this function)
	 * @return a referenced BDD with the maximal stable set in c
	 */
	private JDDNode findMaximalStableSet(NondetModel model, JDDNode candidateStates)
	{

		JDDNode old = JDD.Constant(0);
		JDDNode current = candidateStates;

		while (!current.equals(old)) {
			JDD.Deref(old);
			JDD.Ref(current);
			old = current;

			JDD.Ref(current);
			JDD.Ref(model.getTrans());
			// Select transitions starting in current
			JDDNode currTrans = JDD.Apply(JDD.TIMES, model.getTrans(), current);
			// Select transitions starting in current and ending in current
			JDDNode tmp = JDD.PermuteVariables(current, model.getAllDDRowVars(), model.getAllDDColVars());
			tmp = JDD.Apply(JDD.TIMES, currTrans, tmp);
			// Sum all successor probabilities for each (state, action) tuple
			tmp = JDD.SumAbstract(tmp, model.getAllDDColVars());
			// If the sum for a (state,action) tuple is 1,
			// there is an action that remains in the stable set with prob 1
			tmp = JDD.GreaterThan(tmp, 1 - prism.getSumRoundOff());
			// Without fairness, we just need one action per state
			current = JDD.ThereExists(tmp, model.getAllDDNondetVars());
		}
		JDD.Deref(old);
		return current;
	}

	/**
	 * Returns the transition relation of a stable set
	 * 
	 * @param b
	 *            BDD of a stable set (dereferenced after calling this function)
	 * @return referenced BDD of the transition relation restricted to the stable set
	 */
	public JDDNode maxStableSetTrans(NondetModel model, JDDNode b)
	{

		JDD.Ref(b);
		JDD.Ref(model.getTrans());
		// Select transitions starting in b
		JDDNode currTrans = JDD.Apply(JDD.TIMES, model.getTrans(), b);
		JDDNode mask = JDD.PermuteVariables(b, model.getAllDDRowVars(), model.getAllDDColVars());
		// Select transitions starting in current and ending in current
		mask = JDD.Apply(JDD.TIMES, currTrans, mask);
		// Sum all successor probabilities for each (state, action) tuple
		mask = JDD.SumAbstract(mask, model.getAllDDColVars());
		// If the sum for a (state,action) tuple is 1,
		// there is an action that remains in the stable set with prob 1
		mask = JDD.GreaterThan(mask, 1 - prism.getSumRoundOff());
		// select the transitions starting in these tuples
		JDD.Ref(model.getTrans01());
		JDDNode stableTrans01 = JDD.And(model.getTrans01(), mask);
		// Abstract over actions
		return JDD.ThereExists(stableTrans01, model.getAllDDNondetVars());
	}

	public JDDNode maxStableSetTrans1(NondetModel model, JDDNode b)
	{

		JDD.Ref(b);
		JDD.Ref(model.getTrans());
		// Select transitions starting in b
		JDDNode currTrans = JDD.Apply(JDD.TIMES, model.getTrans(), b);
		JDDNode mask = JDD.PermuteVariables(b, model.getAllDDRowVars(), model.getAllDDColVars());
		// Select transitions starting in current and ending in current
		mask = JDD.Apply(JDD.TIMES, currTrans, mask);
		// Sum all successor probabilities for each (state, action) tuple
		mask = JDD.SumAbstract(mask, model.getAllDDColVars());
		// If the sum for a (state,action) tuple is 1,
		// there is an action that remains in the stable set with prob 1
		mask = JDD.GreaterThan(mask, 1 - prism.getSumRoundOff());
		// select the transitions starting in these tuples
		JDD.Ref(model.getTrans01());
		JDDNode stableTrans01 = JDD.And(model.getTrans01(), mask);
		// Abstract over actions
		return stableTrans01;
	}

	/**
	 * Returns the union of each set in the vector that has nonempty intersection with the filter BDD.
	 * 
	 * @param sets
	 *            Vector with sets which are dereferenced after calling this function
	 * @param filter
	 *            filter BDD against which each set is checked for nonempty intersection also dereferenced after calling
	 *            this function
	 * @return Referenced BDD with the filtered union
	 * 
	 * modified by Hongyang for testing
	 */
	private JDDNode filteredUnion(Vector<JDDNode> sets, JDDNode filter)
	{
		JDDNode union = JDD.Constant(0);
		for (JDDNode set : sets) {
			if (JDD.AreInterecting(set, filter))
				union = JDD.Or(union, set);
			else
				JDD.Deref(set);
		}
		JDD.Deref(filter);
		return union;
	}

	/**
	 * Check whether an end component contains a fair sub-componet
	 * See Quantitative Analysis under Fairness Constraints by
	 * C. Baier, M. Groesser and F. Ciesinski in ATVA 2009
	 * 
	 * @param ec End componenet, represented as a set of states
	 * @param SF Set of strong fairness constraints, each constraint 
	 * is a set of states, in which an action is enabled infinitely often
	 * @param WF Set of weak fairness constraints, each constraint is 
	 * a set of states, in which an action is continuously enabled
	 * 
	 * Here, each constraint is an index of an action
	 * 
	 * @return true if the end component contains a fair sub-componet 
	 * 
	 * @author Hongyang Qu
	 */
	public boolean isEcFair(NondetModel model, JDDNode ec, Vector<JDDFairnessConstraint> sf, Vector<JDDFairnessConstraint> wf/*, LTLModelChecker mcLtl*/)
	{
		// compute the restricted transition relation in the end component
		// copied from maxStableSetTrans in LTLModelChecker
		JDD.Ref(ec);
		JDD.Ref(model.getTrans());
		JDDNode currTrans = JDD.Apply(JDD.TIMES, model.getTrans(), ec);
		JDD.Ref(ec);
		JDDNode mask = JDD.PermuteVariables(ec, model.getAllDDRowVars(), model.getAllDDColVars());
		mask = JDD.Apply(JDD.TIMES, currTrans, mask);
		mask = JDD.SumAbstract(mask, model.getAllDDColVars());
		mask = JDD.GreaterThan(mask, 1 - prism.getSumRoundOff());
		JDD.Ref(model.getTrans01());
		JDDNode ecTrans = JDD.And(model.getTrans01(), mask);

		if (wf != null)
			// Check if every wf holds in ec
			for (int i = 0; i < wf.size(); i++) {
				JDDFairnessConstraint ddfc = wf.get(i);
				// First test if X \ H =/= emptyset
				JDDNode tmpec = ec;
				JDD.Ref(tmpec);
				boolean sat = true;
				// test states
				for (int j = 0; j < ddfc.H.size(); j++) {
					if (ddfc.bH.get(j).booleanValue()) {
						JDD.Ref(ddfc.H.get(j));
						tmpec = JDD.Apply(JDD.MINUS, tmpec, ddfc.H.get(j));
						//if(true) {
						if (tmpec.equals(JDD.ZERO)) {
							sat = false;
							break;
						}
					}
				}
				JDD.Deref(tmpec);
				if (sat)
					continue;

				// test actions
				JDD.Ref(ecTrans);
				JDDNode tmpecTrans = JDD.ThereExists(ecTrans, model.getAllDDColVars());
				for (int j = 0; j < ddfc.H.size(); j++) {
					if (!ddfc.bH.get(j).booleanValue()) {
						JDD.Ref(tmpecTrans);
						JDD.Ref(ddfc.H.get(j));
						tmpec = JDD.Apply(JDD.MINUS, tmpecTrans, ddfc.H.get(j));
						if (tmpec.equals(JDD.Constant(0))) {
							sat = false;
							JDD.Deref(tmpec);
							break;
						}
						JDD.Deref(tmpec);
					}
				}
				JDD.Deref(tmpecTrans);
				if (sat)
					continue;
				// Now X \ H == emptyset. We need to test X \land K =/= emptyset 
				sat = isNonEmptyIntersection(ddfc.K, ddfc.bK, model, ec, ecTrans);

				if (!sat) {
					JDD.Deref(ecTrans);
					return false;
				}
			}

		if (sf != null)
			// Check if every sf holds in ec
			for (int i = 0; i < sf.size(); i++) {
				JDDFairnessConstraint ddfc = sf.get(i);
				// First test if X \land H = emptyset
				boolean isfair = true;
				boolean sat = isNonEmptyIntersection(ddfc.H, ddfc.bH, model, ec, ecTrans);
				// Next test if X \land K = emptyset
				if (sat) {
					isfair = isNonEmptyIntersection(ddfc.K, ddfc.bK, model, ec, ecTrans);
				}
				if (!isfair) {// sf does not hold in ec
					//mainLog.println(" looking for fair subcomponent ...");
					JDDNode states = ec;
					JDD.Ref(states);
					JDDNode actions = JDD.Constant(0);
					for (int j = 0; j < ddfc.H.size(); j++) {
						JDD.Ref(ddfc.H.get(j));
						if (ddfc.bH.get(j).booleanValue()) {
							states = JDD.And(states, JDD.Not(ddfc.H.get(j)));
						} else {
							actions = JDD.Or(actions, ddfc.H.get(j));
						}
					}

					// ------------------- findEndComponents ------------------
					boolean initialCandidate = true;
					Stack<JDDNode> candidates = new Stack<JDDNode>();
					candidates.push(states);
					Vector<JDDNode> ecs = new Vector<JDDNode>();
					SCCComputer sccComputer;

					while (!candidates.isEmpty()) {
						JDDNode candidate = candidates.pop();

						// Compute the stable set
						JDD.Ref(candidate);
						//JDDNode stableSet = findMaximalStableSet(model, candidate);

						// ------------------- findMaximalStableSet ------------------
						JDDNode old = JDD.Constant(0);
						JDDNode current = candidate;

						while (!current.equals(old)) {
							JDD.Deref(old);
							JDD.Ref(current);
							old = current;

							JDD.Ref(current);
							JDD.Ref(model.getTrans());
							// Select transitions starting in current
							JDDNode currTrans1 = JDD.Apply(JDD.TIMES, model.getTrans(), current);
							// Select transitions starting in current and ending in current
							JDDNode tmp = JDD.PermuteVariables(current, model.getAllDDRowVars(), model.getAllDDColVars());
							tmp = JDD.Apply(JDD.TIMES, currTrans1, tmp);

							// get rid of actions in H
							if (!actions.equals(JDD.ZERO)) {
								mainLog.println("Removing actions in H");
								JDD.Ref(actions);
								tmp = JDD.ITE(actions, JDD.Constant(0), tmp);
							}

							// Sum all successor probabilities for each (state, action) tuple
							tmp = JDD.SumAbstract(tmp, model.getAllDDColVars());
							// If the sum for a (state,action) tuple is 1,
							// there is an action that remains in the stable set with prob 1
							tmp = JDD.GreaterThan(tmp, 1 - prism.getSumRoundOff());
							// Without fairness, we just need one action per state
							current = JDD.ThereExists(tmp, model.getAllDDNondetVars());
						}
						JDD.Deref(old);
						JDDNode stableSet = current;
						// --------------------- end of findMaximalStableSet ----------------

						// Drop empty sets
						if (stableSet.equals(JDD.ZERO)) {
							JDD.Deref(stableSet);
							JDD.Deref(candidate);
							continue;
						}

						if (!initialCandidate) {
							// candidate is an SCC, check if it's stable
							if (stableSet.equals(candidate)) {
								ecs.add(candidate);
								JDD.Deref(stableSet);
								continue;
							}
						} else {
							initialCandidate = false;
						}
						JDD.Deref(candidate);

						// Filter bad transitions
						JDD.Ref(stableSet);
						JDDNode stableSetTrans = maxStableSetTrans(model, stableSet);

						// now find the maximal SCCs in (stableSet, stableSetTrans)
						Vector<JDDNode> sccs;
						sccComputer = prism.getSCCComputer(stableSet, stableSetTrans, model.getAllDDRowVars(), model.getAllDDColVars());
						sccComputer.computeSCCs();
						JDD.Deref(stableSet);
						JDD.Deref(stableSetTrans);
						sccs = sccComputer.getVectSCCs();
						mainLog.println("In isEcFair: SCC computer found " + sccs.size() + " SCCs");
						JDD.Deref(sccComputer.getNotInSCCs());
						candidates.addAll(sccs);
					}
					// ------------ end of findEndComponents --------------------------

					Vector<JDDFairnessConstraint> newsf = new Vector<JDDFairnessConstraint>(sf);
					newsf.remove(i);
					boolean isecfair = false;
					for (JDDNode newec : ecs) {
						if (isEcFair(model, newec, newsf, wf)) {
							isecfair = true;
							break;
						}
					}
					for (JDDNode newec : ecs)
						JDD.Deref(newec);
					JDD.Deref(actions);
					JDD.Deref(ecTrans);
					return isecfair;
				}
			}

		JDD.Deref(ecTrans);
		return true;
	}

	public boolean isEcFairAccept(NondetModel model, JDDNode ec, Vector<JDDFairnessConstraint> sf, Vector<JDDFairnessConstraint> wf, JDDNode acceptanceVector_L/*, LTLModelChecker mcLtl*/)
	{
		// compute the restricted transition relation in the end component
		// copied from maxStableSetTrans in LTLModelChecker
		JDD.Ref(ec);
		JDD.Ref(model.getTrans());
		JDDNode currTrans = JDD.Apply(JDD.TIMES, model.getTrans(), ec);
		JDD.Ref(ec);
		JDDNode mask = JDD.PermuteVariables(ec, model.getAllDDRowVars(), model.getAllDDColVars());
		mask = JDD.Apply(JDD.TIMES, currTrans, mask);
		mask = JDD.SumAbstract(mask, model.getAllDDColVars());
		mask = JDD.GreaterThan(mask, 1 - prism.getSumRoundOff());
		JDD.Ref(model.getTrans01());
		JDDNode ecTrans = JDD.And(model.getTrans01(), mask);

		if (wf != null)
			// Check if every wf holds in ec
			for (int i = 0; i < wf.size(); i++) {
				JDDFairnessConstraint ddfc = wf.get(i);
				// First test if X \ H =/= emptyset
				JDDNode tmpec = ec;
				JDD.Ref(tmpec);
				boolean sat = true;
				// test states
				for (int j = 0; j < ddfc.H.size(); j++) {
					if (ddfc.bH.get(j).booleanValue()) {
						JDD.Ref(ddfc.H.get(j));
						tmpec = JDD.Apply(JDD.MINUS, tmpec, ddfc.H.get(j));
						//if(true) {
						if (tmpec.equals(JDD.ZERO)) {
							sat = false;
							break;
						}
					}
				}
				JDD.Deref(tmpec);
				if (sat)
					continue;

				// test actions
				JDD.Ref(ecTrans);
				JDDNode tmpecTrans = JDD.ThereExists(ecTrans, model.getAllDDColVars());
				for (int j = 0; j < ddfc.H.size(); j++) {
					if (!ddfc.bH.get(j).booleanValue()) {
						JDD.Ref(tmpecTrans);
						JDD.Ref(ddfc.H.get(j));
						tmpec = JDD.Apply(JDD.MINUS, tmpecTrans, ddfc.H.get(j));
						if (tmpec.equals(JDD.Constant(0))) {
							sat = false;
							JDD.Deref(tmpec);
							break;
						}
						JDD.Deref(tmpec);
					}
				}
				JDD.Deref(tmpecTrans);
				if (sat)
					continue;
				// Now X \ H == emptyset. We need to test X \land K =/= emptyset 
				sat = isNonEmptyIntersection(ddfc.K, ddfc.bK, model, ec, ecTrans);

				if (!sat) {
					JDD.Deref(ecTrans);
					return false;
				}
			}

		if (sf != null)
			// Check if every sf holds in ec
			for (int i = 0; i < sf.size(); i++) {
				JDDFairnessConstraint ddfc = sf.get(i);
				// First test if X \land H = emptyset
				boolean isfair = true;
				boolean sat = isNonEmptyIntersection(ddfc.H, ddfc.bH, model, ec, ecTrans);
				// Next test if X \land K =/= emptyset
				if (sat) {
					isfair = isNonEmptyIntersection(ddfc.K, ddfc.bK, model, ec, ecTrans);
				}
				if (!isfair) {// sf does not hold in ec
					//mainLog.println(" looking for fair subcomponent ...");
					JDDNode states = ec;
					JDD.Ref(states);
					JDDNode actions = JDD.Constant(0);
					for (int j = 0; j < ddfc.H.size(); j++) {
						JDD.Ref(ddfc.H.get(j));
						if (ddfc.bH.get(j).booleanValue()) {
							states = JDD.And(states, JDD.Not(ddfc.H.get(j)));
						} else {
							actions = JDD.Or(actions, ddfc.H.get(j));
						}
					}

					// ------------------- findEndComponents ------------------
					boolean initialCandidate = true;
					Stack<JDDNode> candidates = new Stack<JDDNode>();
					candidates.push(states);
					Vector<JDDNode> ecs = new Vector<JDDNode>();
					SCCComputer sccComputer;

					while (!candidates.isEmpty()) {
						JDDNode candidate = candidates.pop();

						// Compute the stable set
						JDD.Ref(candidate);
						//JDDNode stableSet = findMaximalStableSet(model, candidate);

						// ------------------- findMaximalStableSet ------------------
						JDDNode old = JDD.Constant(0);
						JDDNode current = candidate;

						while (!current.equals(old)) {
							JDD.Deref(old);
							JDD.Ref(current);
							old = current;

							JDD.Ref(current);
							JDD.Ref(model.getTrans());
							// Select transitions starting in current
							JDDNode currTrans1 = JDD.Apply(JDD.TIMES, model.getTrans(), current);
							// Select transitions starting in current and ending in current
							JDDNode tmp = JDD.PermuteVariables(current, model.getAllDDRowVars(), model.getAllDDColVars());
							tmp = JDD.Apply(JDD.TIMES, currTrans1, tmp);

							// get rid of actions in H
							if (!actions.equals(JDD.ZERO)) {
								mainLog.println("Removing actions in H");
								JDD.Ref(actions);
								tmp = JDD.ITE(actions, JDD.Constant(0), tmp);
							}

							// Sum all successor probabilities for each (state, action) tuple
							tmp = JDD.SumAbstract(tmp, model.getAllDDColVars());
							// If the sum for a (state,action) tuple is 1,
							// there is an action that remains in the stable set with prob 1
							tmp = JDD.GreaterThan(tmp, 1 - prism.getSumRoundOff());
							// Without fairness, we just need one action per state
							current = JDD.ThereExists(tmp, model.getAllDDNondetVars());
						}
						JDD.Deref(old);
						JDDNode stableSet = current;
						// --------------------- end of findMaximalStableSet ----------------

						// Drop empty sets
						if (stableSet.equals(JDD.ZERO)) {
							JDD.Deref(stableSet);
							JDD.Deref(candidate);
							continue;
						}

						if (!initialCandidate) {
							// candidate is an SCC, check if it's stable
							if (stableSet.equals(candidate)) {
								ecs.add(candidate);
								JDD.Deref(stableSet);
								continue;
							}
						} else {
							initialCandidate = false;
						}
						JDD.Deref(candidate);

						// Filter bad transitions
						JDD.Ref(stableSet);
						JDDNode stableSetTrans = maxStableSetTrans(model, stableSet);

						// now find the maximal SCCs in (stableSet, stableSetTrans)
						Vector<JDDNode> sccs;
						sccComputer = prism.getSCCComputer(stableSet, stableSetTrans, model.getAllDDRowVars(), model.getAllDDColVars());
						sccComputer.computeSCCs();
						JDD.Deref(stableSet);
						JDD.Deref(stableSetTrans);
						sccs = sccComputer.getVectSCCs();
						mainLog.println("In isEcFair: SCC computer found " + sccs.size() + " SCCs");
						JDD.Deref(sccComputer.getNotInSCCs());
						candidates.addAll(sccs);
					}
					// ------------ end of findEndComponents --------------------------

					//Vector<JDDNode> newecs = mcLtl.findEndComponents(model, candidateStates);
					Vector<JDDFairnessConstraint> newsf = new Vector<JDDFairnessConstraint>(sf);
					newsf.remove(i);
					boolean isecfair = false;
					for (JDDNode newec : ecs) {
						if (JDD.AreInterecting(newec, acceptanceVector_L) && isEcFairAccept(model, newec, newsf, wf, acceptanceVector_L/*, mcLtl*/)) {
							isecfair = true;
							break;
						}
					}
					/*if(derefStates)
						JDD.Deref(states);*/
					for (JDDNode newec : ecs)
						JDD.Deref(newec);
					JDD.Deref(actions);
					JDD.Deref(ecTrans);
					return isecfair;
				}
			}

		JDD.Deref(ecTrans);
		return true;
	}

	/*private boolean[] isEcFairAccept(NondetModel model, JDDNode ec, 
			Vector<JDDFairnessConstraint> sf, Vector<JDDFairnessConstraint> wf, 
			ArrayList<JDDNode> statesH, ArrayList<JDDNode> statesL,
			LTLModelChecker mcLtl) throws PrismException {
		boolean[] result = null;
		// compute the restricted transition relation in the end component
		// copied from maxStableSetTrans in LTLModelChecker
		JDD.Ref(ec);
		JDD.Ref(model.getTrans());
		JDDNode currTrans = JDD.Apply(JDD.TIMES, model.getTrans(), ec);
		JDD.Ref(ec);
		JDDNode mask = JDD.PermuteVariables(ec, model.getAllDDRowVars(), model.getAllDDColVars());
		mask = JDD.Apply(JDD.TIMES, currTrans, mask);
		mask = JDD.SumAbstract(mask, model.getAllDDColVars());
		mask = JDD.GreaterThan(mask, 1 - prism.getSumRoundOff());
		JDD.Ref(model.getTrans01());
		JDDNode ecTrans = JDD.And(model.getTrans01(), mask);
		
			
		if(wf != null)
			// Check if every wf holds in ec
			for (int i=0; i<wf.size(); i++) {
				JDDFairnessConstraint ddfc = wf.get(i);
				// First test if X \ H =/= emptyset
				JDDNode tmpec = ec;
				JDD.Ref(tmpec);
				boolean sat = true;
				// test states
				for(int j=0; j<ddfc.H.size(); j++) {
					if(ddfc.bH.get(j).booleanValue()) { 
						JDD.Ref(ddfc.H.get(j));
						tmpec = JDD.Apply(JDD.MINUS, tmpec, ddfc.H.get(j));
						//if(true) {
						if(tmpec.equals(JDD.ZERO)) {
							sat = false;
							break;
						}
					}
				}
				JDD.Deref(tmpec);
				if(sat) 
					continue;
		
				// test actions
				JDD.Ref(ecTrans);
				JDDNode tmpecTrans = JDD.ThereExists(ecTrans, model.getAllDDColVars());
				for(int j=0; j<ddfc.H.size(); j++) {
					if(!ddfc.bH.get(j).booleanValue()) { 
						JDD.Ref(tmpecTrans);
						JDD.Ref(ddfc.H.get(j));
						tmpec = JDD.Apply(JDD.MINUS, tmpecTrans, ddfc.H.get(j));
						if(tmpec.equals(JDD.Constant(0))) {
							sat = false;
							JDD.Deref(tmpec);
							break;
						}
						JDD.Deref(tmpec);
					}
				}
				JDD.Deref(tmpecTrans);
				if(sat)
					continue;
				// Now X \ H == emptyset. We need to test X \land K =/= emptyset 
				sat = isNonEmptyIntersection(ddfc.K, ddfc.bK, model, ec, ecTrans);
				
				if(!sat) {
					JDD.Deref(ecTrans);
					result = new boolean[1];
					result[0] = false;
					return result;
				}
			}
			
		if(sf != null)
			// Check if every sf holds in ec
			for (int i=0; i<sf.size(); i++) {
				JDDFairnessConstraint ddfc = sf.get(i);
				// First test if X \land H = emptyset
				boolean isfair = true;
				boolean sat = isNonEmptyIntersection(ddfc.H, ddfc.bH, model, ec, ecTrans);
				// Next test if X \land K =/= emptyset
				if(sat) {
					isfair = isNonEmptyIntersection(ddfc.K, ddfc.bK, model, ec, ecTrans);
				}
				if(!isfair) {// sf does not hold in ec
					//mainLog.println(" looking for fair subcomponent ...");
					JDDNode states = ec;
					JDD.Ref(states);
					JDDNode actions = JDD.Constant(0);
					for(int j=0; j<ddfc.H.size(); j++) {
						JDD.Ref(ddfc.H.get(j));
						if(ddfc.bH.get(j).booleanValue()) {
							states = JDD.And(states, JDD.Not(ddfc.H.get(j)));
						} else {
							actions = JDD.Or(actions, ddfc.H.get(j));
						}
					}
					
					// ------------------- findEndComponents ------------------
					boolean initialCandidate = true;
					Stack<JDDNode> candidates = new Stack<JDDNode>();
					candidates.push(states);
					Vector<JDDNode> ecs = new Vector<JDDNode>();
					SCCComputer sccComputer;

					while (!candidates.isEmpty()) {
						JDDNode candidate = candidates.pop();
						
						// Compute the stable set
						JDD.Ref(candidate);			
						//JDDNode stableSet = findMaximalStableSet(model, candidate);
						
						// ------------------- findMaximalStableSet ------------------
						JDDNode old = JDD.Constant(0);
						JDDNode current = candidate;

						while (!current.equals(old)) {
							JDD.Deref(old);
							JDD.Ref(current);
							old = current;

							JDD.Ref(current);
							JDD.Ref(model.getTrans());
							// Select transitions starting in current
							JDDNode currTrans1 = JDD.Apply(JDD.TIMES, model.getTrans(), current);
							// Select transitions starting in current and ending in current
							JDDNode tmp = JDD.PermuteVariables(current, model.getAllDDRowVars(), model.getAllDDColVars());
							tmp = JDD.Apply(JDD.TIMES, currTrans1, tmp);
							
							// get rid of actions in H
							if(!actions.equals(JDD.ZERO)) {
								mainLog.println("Removing actions in H");
								JDD.Ref(actions);
								tmp = JDD.ITE(actions, JDD.Constant(0), tmp);
							}
							
							// Sum all successor probabilities for each (state, action) tuple
							tmp = JDD.SumAbstract(tmp, model.getAllDDColVars());
							// If the sum for a (state,action) tuple is 1,
							// there is an action that remains in the stable set with prob 1
							tmp = JDD.GreaterThan(tmp, 1 - prism.getSumRoundOff());
							// Without fairness, we just need one action per state
							current = JDD.ThereExists(tmp, model.getAllDDNondetVars());
						}
						JDD.Deref(old);
						JDDNode stableSet = current;
						// --------------------- end of findMaximalStableSet ----------------
						
						// Drop empty sets
						if (stableSet.equals(JDD.ZERO)) {
							JDD.Deref(stableSet);
							JDD.Deref(candidate);
							continue;
						}

						if (!initialCandidate) {
							// candidate is an SCC, check if it's stable
							if (stableSet.equals(candidate)) {
								ecs.add(candidate);
								JDD.Deref(stableSet);
								continue;
							}
						} else {
							initialCandidate = false;
						}
						JDD.Deref(candidate);

						// Filter bad transitions
						JDD.Ref(stableSet);
						JDDNode stableSetTrans = maxStableSetTrans(model, stableSet);

						// now find the maximal SCCs in (stableSet, stableSetTrans)
						Vector<JDDNode> sccs;
						sccComputer = prism.getSCCComputer(stableSet, stableSetTrans, model.getAllDDRowVars(), model
								.getAllDDColVars());
						sccComputer.computeSCCs();			
						JDD.Deref(stableSet);
						JDD.Deref(stableSetTrans);
						sccs = sccComputer.getVectSCCs();
						mainLog.println("In isEcFair: SCC computer found " + sccs.size() + " SCCs");
						JDD.Deref(sccComputer.getNotInSCCs());
						candidates.addAll(sccs);
					}
					// ------------ end of findEndComponents --------------------------
					
					//Vector<JDDNode> newecs = mcLtl.findEndComponents(model, candidateStates);
					Vector<JDDFairnessConstraint> newsf = new Vector<JDDFairnessConstraint>(sf);
					newsf.remove(i);
					for(JDDNode newec : ecs) {
						boolean[] isAFMEC = isEcFairAccept(model, newec, newsf, wf, statesH, statesL, mcLtl);
						if(isAFMEC[0]) {
							isfair = true;
							if(isAFMEC[1]) {
								JDD.Deref(actions);
								JDD.Deref(ecTrans);
								result = new boolean[2];
								result[0] = true;
								result[1] = true;
								return result;
							}
						} / *else
							JDD.Deref(newec);* /
					}
					/ *if(derefStates)
						JDD.Deref(states);* /
					//if(derefActions)
					JDD.Deref(actions);
					JDD.Deref(ecTrans);
					if(isfair) {
						result = new boolean[2];
						result[0] = true;
						result[1] = false;
					} else {
						result = new boolean[1];
						result[0] = false;
					}
					return result;						
				} 
			}
		// this EC is fair. Now search for acceptance states
		// test if ec has not-empty intersection with L states
		Vector<JDDNode> newH = new Vector<JDDNode>();
		Vector<JDDNode> newL = new Vector<JDDNode>();
		for(int i=0; i<statesL.size(); i++) {
			JDDNode acceptanceVector_L = statesL.get(i);
			if(JDD.AreInterecting(ec, acceptanceVector_L)) {
				JDDNode acceptanceVector_H = statesH.get(i);
				JDD.Ref(acceptanceVector_H);
				JDD.Ref(ec);
				JDDNode tmp = JDD.And(ec, JDD.Not(acceptanceVector_H));
				if(tmp.equals(JDD.ZERO)) {
					JDD.Deref(tmp);
					JDD.Deref(ecTrans);
					result = new boolean[2];
					result[0] = result[1] = true;
					return result;
				} else {
					JDD.Deref(tmp);
					newH.add(acceptanceVector_H);
					newL.add(acceptanceVector_L);
				}
			}
		}
		// Now we need to search for small EC that is fair and accepting
		for(int i=0; i<newH.size(); i++) {
			JDDNode acceptanceVector_H = newH.get(i);
			JDDNode acceptanceVector_L = newL.get(i);
			JDD.Ref(acceptanceVector_H);
			JDD.Ref(ec);
			JDDNode tmp = JDD.And(ec, acceptanceVector_H);
			// find all maximal end components in ec
			Vector<JDDNode> allecs = findEndComponents(model, tmp, acceptanceVector_L);
			ArrayList<JDDNode> tmpH = new ArrayList<JDDNode>();
			tmpH.add(acceptanceVector_H);
			ArrayList<JDDNode> tmpL = new ArrayList<JDDNode>();
			tmpL.add(acceptanceVector_L);
			for(JDDNode newec: allecs) {
				boolean[] newresult = isEcFairAccept(model, newec, sf, wf, tmpH, tmpL, mcLtl);
				if(newresult[0] && newresult[1]) {
					JDD.Deref(ecTrans);
					result = new boolean[2];
					result[0] = result[1] = true;
					return result;
				}
			}
		}
		
		JDD.Deref(ecTrans);
		result = new boolean[1];
		result[0] = false;
		return result;
	}*/

	private boolean isNonEmptyIntersection(ArrayList<JDDNode> HK, ArrayList<Boolean> bHK, NondetModel model, JDDNode ec, JDDNode ecTrans)
	{
		boolean sat = false;
		for (int j = 0; j < HK.size(); j++) {
			if (bHK.get(j).booleanValue()) {
				if (JDD.AreInterecting(ec, HK.get(j))) {
					sat = true;
					break;
				}
			}
		}
		if (!sat) {
			// test actions
			JDD.Ref(ecTrans);
			JDDNode tmpecTrans = JDD.ThereExists(ecTrans, model.getAllDDColVars());
			for (int j = 0; j < HK.size(); j++) {
				if (!bHK.get(j).booleanValue()) {
					JDD.Ref(HK.get(j));
					JDD.Ref(tmpecTrans);
					JDDNode tmp = JDD.And(tmpecTrans, HK.get(j));
					if (!tmp.equals(JDD.ZERO)) {
						sat = true;
					}
					JDD.Deref(tmp);
					if (sat)
						break;
				}
			}
			JDD.Deref(tmpecTrans);
		}
		return sat;
	}

}

class queueElement
{
	List<JDDNode> statesH;
	List<JDDNode> statesL;
	JDDNode targetDD;
	List<Integer> draIDs;
	int next;
	List<queueElement> children;

	public queueElement(List<JDDNode> statesH, List<JDDNode> statesL, JDDNode targetDD, List<Integer> draIDs, int next)
	{
		this.statesH = statesH;
		this.statesL = statesL;
		this.targetDD = targetDD;
		this.draIDs = draIDs;
		this.next = next;
	}

	public void addChildren(queueElement child)
	{
		if (children == null)
			children = new ArrayList<queueElement>();
		children.add(child);
	}
}
