//==============================================================================
//	
//	Copyright (c) 2002-
//	Authors:
//	* Dave Parker <david.parker@comlab.ox.ac.uk> (University of Oxford, formerly University of Birmingham)
//	* Carlos S. Bederian (Universidad Nacional de Cordoba)
//	
//------------------------------------------------------------------------------
//	
//	This file is part of PRISM.
//	
//	PRISM is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//	
//	PRISM is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//	
//	You should have received a copy of the GNU General Public License
//	along with PRISM; if not, write to the Free Software Foundation,
//	Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//	
//==============================================================================

package prism;

import java.io.FileNotFoundException;
import java.util.*;

import jdd.*;
import dv.*;
import mtbdd.*;
import sparse.*;
import hybrid.*;
import parser.type.*;
import parser.ast.*;
import parser.visitor.ASTTraverse;
import jltl2dstar.*;

/*
 * Model checker for MDPs
 */
public class NondetModelChecker extends NonProbModelChecker
{
	// Model (MDP)
	protected NondetModel model;

	// Extra (MDP) model info
	protected JDDNode nondetMask;
	protected JDDVars allDDNondetVars;

	// Options (in addition to those inherited from StateModelChecker):

	// Use 0,1 precomputation algorithms?
	// if 'precomp' is false, this disables all (non-essential) use of prob0/prob1
	// if 'precomp' is true, the values of prob0/prob1 determine what is used
	// (currently prob0/prob are not under user control)
	protected boolean precomp;
	protected boolean prob0;
	protected boolean prob1;
	// Use fairness?
	protected boolean fairness;
	
	// SCC-based value iteration
	protected boolean sccBased;

	class JDDFairnessConstraint
	{
		public ArrayList<JDDNode> H;
		public ArrayList<JDDNode> K;
		public ArrayList<Boolean> bH;
		public ArrayList<Boolean> bK;

		public JDDFairnessConstraint(ExpressionFairness e, NondetModel modelProduct)
		{
			Vector<Expression> veh = e.getH();
			if (veh != null) {
				H = new ArrayList<JDDNode>();
				bH = new ArrayList<Boolean>();
				for (Expression fe : veh) {
					JDDNode dd;
					boolean isStates = false;
					if (fe instanceof ExpressionAction) {
						JDD.Ref(model.getTransActions());
						dd = JDD.Equals(model.getTransActions(), ((ExpressionAction) fe).getIndex() + 1);
					} else {
						try {
							dd = checkExpressionDD(fe);
							JDD.Ref(modelProduct.getReach());
							dd = JDD.And(dd, modelProduct.getReach());
						} catch (PrismException pe) {
							pe.printStackTrace();
							dd = JDD.Constant(1);
						}
						isStates = true;
					}
					H.add(dd);
					bH.add(new Boolean(isStates));
				}
			}
			veh = e.getK();
			if (veh != null) {
				K = new ArrayList<JDDNode>();
				bK = new ArrayList<Boolean>();
				for (Expression fe : veh) {
					JDDNode dd;
					boolean isStates = false;
					if (fe instanceof ExpressionAction) {
						JDD.Ref(model.getTransActions());
						dd = JDD.Equals(model.getTransActions(), ((ExpressionAction) fe).getIndex() + 1);
					} else {
						try {
							dd = checkExpressionDD(fe);
							JDD.Ref(modelProduct.getReach());
							dd = JDD.And(dd, modelProduct.getReach());
						} catch (PrismException pe) {
							pe.printStackTrace();
							dd = JDD.Constant(1);
						}
						isStates = true;
					}
					K.add(dd);
					bK.add(new Boolean(isStates));
				}
			}
		}

		public void derefHK()
		{
			if (H != null) {
				for (int i = 0; i < H.size(); i++)
					JDD.Deref(H.get(i));
			}
			if (K != null) {
				for (int i = 0; i < K.size(); i++)
					JDD.Deref(K.get(i));
			}
		}
	}

	// Constructor

	public NondetModelChecker(Prism prism, Model m, PropertiesFile pf) throws PrismException
	{
		// Initialise
		super(prism, m, pf);
		if (!(m instanceof NondetModel)) {
			throw new PrismException("Wrong model type passed to NondetModelChecker.");
		}
		model = (NondetModel) m;
		nondetMask = model.getNondetMask();
		allDDNondetVars = model.getAllDDNondetVars();

		// Inherit some options from parent Prism object and store locally.
		precomp = prism.getPrecomp();
		prob0 = true;
		prob1 = true;
		fairness = prism.getFairness();

		// Display warning and/or make changes for some option combinations
		if (prism.getExportAdv() != Prism.EXPORT_ADV_NONE) {
			if (engine != Prism.SPARSE) {
				mainLog.println("\nWarning: Adversary generation is only implemented for the sparse engine");
			}
			if (precomp && prob1) {
				mainLog.println("\nWarning: Disabling Prob1 since this is needed for adversary generation");
				prob1 = false;
			}
		}

		// Pass some options onto engines/native code.
		PrismNative.setExportAdv(prism.getExportAdv());
		PrismNative.setExportAdvFilename(prism.getExportAdvFilename());
		switch (engine) {
		case Prism.MTBDD:
			PrismMTBDD.setTermCrit(prism.getTermCrit());
			PrismMTBDD.setTermCritParam(prism.getTermCritParam());
			PrismMTBDD.setMaxIters(prism.getMaxIters());
			break;
		case Prism.SPARSE:
			PrismSparse.setTermCrit(prism.getTermCrit());
			PrismSparse.setTermCritParam(prism.getTermCritParam());
			PrismSparse.setMaxIters(prism.getMaxIters());
			PrismSparse.setCompact(prism.getCompact());
		case Prism.HYBRID:
			PrismHybrid.setTermCrit(prism.getTermCrit());
			PrismHybrid.setTermCritParam(prism.getTermCritParam());
			PrismHybrid.setMaxIters(prism.getMaxIters());
			PrismHybrid.setCompact(prism.getCompact());
			PrismHybrid.setSBMaxMem(prism.getSBMaxMem());
			PrismHybrid.setNumSBLevels(prism.getNumSBLevels());
		}
	}

	// -----------------------------------------------------------------------------------
	// Check a property, i.e. an expression
	// -----------------------------------------------------------------------------------

	// Check expression (recursive)

	public StateValues checkExpression(Expression expr) throws PrismException
	{
		StateValues res;

		// P operator
		if (expr instanceof ExpressionProb) {
			res = checkExpressionProb((ExpressionProb) expr);
		}
		// R operator
		else if (expr instanceof ExpressionReward) {
			res = checkExpressionReward((ExpressionReward) expr);
		}
		// Multi-objective
		else if (expr instanceof ExpressionFunc) {
			// Detect "multi" function
			if (((ExpressionFunc) expr).getName().equals("multi")) {
				res = checkExpressionMultiObjective((ExpressionFunc) expr);
			}
			// For any other function, check as normal
			else {
				res = super.checkExpression(expr);
			}
		}
		// Otherwise, use the superclass
		else {
			res = super.checkExpression(expr);
		}

		// Filter out non-reachable states from solution
		// (only necessary for symbolically stored vectors)
		if (res instanceof StateValuesMTBDD)
			res.filter(reach);

		return res;
	}

	// -----------------------------------------------------------------------------------
	// Check method for each operator
	// -----------------------------------------------------------------------------------

	// P operator

	protected StateValues checkExpressionProb(ExpressionProb expr) throws PrismException
	{
		Expression pb; // probability bound (expression)
		double p = 0; // probability bound (actual value)
		String relOp; // relational operator
		boolean min; // are we finding min (true) or max (false) probs

		JDDNode sol;
		StateValues probs = null;

		// Get info from prob operator
		relOp = expr.getRelOp();
		pb = expr.getProb();
		if (pb != null) {
			p = pb.evaluateDouble(constantValues, null);
			if (p < 0 || p > 1)
				throw new PrismException("Invalid probability bound " + p + " in P operator");
		}

		// Check for trivial (i.e. stupid) cases
		if (pb != null) {
			if ((p == 0 && relOp.equals(">=")) || (p == 1 && relOp.equals("<="))) {
				mainLog.print("\nWarning: checking for probability " + relOp + " " + p + " - formula trivially satisfies all states\n");
				JDD.Ref(reach);
				return new StateValuesMTBDD(reach, model);
			} else if ((p == 0 && relOp.equals("<")) || (p == 1 && relOp.equals(">"))) {
				mainLog.print("\nWarning: checking for probability " + relOp + " " + p + " - formula trivially satisfies no states\n");
				return new StateValuesMTBDD(JDD.Constant(0), model);
			}
		}

		// Determine whether min or max probabilities needed
		if (relOp.equals(">") || relOp.equals(">=") || relOp.equals("min=")) {
			// min
			min = true;
		} else if (relOp.equals("<") || relOp.equals("<=") || relOp.equals("max=")) {
			// max
			min = false;
		} else {
			throw new PrismException("Can't use \"P=?\" for MDPs; use \"Pmin=?\" or \"Pmax=?\"");
		}

		// Compute probabilities
		boolean qual = pb != null && ((p == 0) || (p == 1)) && precomp && prob0 && prob1;
		probs = checkProbPathFormula(expr.getExpression(), qual, min);

		// Print out probabilities
		if (verbose) {
			mainLog.print("\n" + (min ? "Minimum" : "Maximum") + " probabilities (non-zero only) for all states:\n");
			probs.print(mainLog);
		}

		// For =? properties, just return values
		if (pb == null) {
			return probs;
		}
		// Otherwise, compare against bound to get set of satisfying states
		else {
			sol = probs.getBDDFromInterval(relOp, p);
			// remove unreachable states from solution
			JDD.Ref(reach);
			sol = JDD.And(sol, reach);
			// free vector
			probs.clear();
			return new StateValuesMTBDD(sol, model);
		}
	}

	// R operator

	protected StateValues checkExpressionReward(ExpressionReward expr) throws PrismException
	{
		Object rs; // reward struct index
		Expression rb; // reward bound (expression)
		double r = 0; // reward bound (actual value)
		String relOp; // relational operator
		boolean min; // are we finding min (true) or max (false) rewards
		Expression expr2; // expression

		JDDNode stateRewards = null, transRewards = null, sol;
		StateValues rewards = null;
		int i;

		// get info from reward operator
		rs = expr.getRewardStructIndex();
		relOp = expr.getRelOp();
		rb = expr.getReward();
		if (rb != null) {
			r = rb.evaluateDouble(constantValues, null);
			if (r < 0)
				throw new PrismException("Invalid reward bound " + r + " in R operator");
		}

		// get reward info
		if (model.getNumRewardStructs() == 0)
			throw new PrismException("Model has no rewards specified");
		if (rs == null) {
			stateRewards = model.getStateRewards(0);
			transRewards = model.getTransRewards(0);
		} else if (rs instanceof Expression) {
			i = ((Expression) rs).evaluateInt(constantValues, null);
			rs = new Integer(i); // for better error reporting below
			stateRewards = model.getStateRewards(i - 1);
			transRewards = model.getTransRewards(i - 1);
		} else if (rs instanceof String) {
			stateRewards = model.getStateRewards((String) rs);
			transRewards = model.getTransRewards((String) rs);
		}
		if (stateRewards == null || transRewards == null)
			throw new PrismException("Invalid reward structure index \"" + rs + "\"");

		// check for trivial (i.e. stupid) cases
		if (rb != null) {
			if (r == 0 && relOp.equals(">=")) {
				mainLog.print("\nWarning: checking for reward " + relOp + " " + r + " - formula trivially satisfies all states\n");
				JDD.Ref(reach);
				return new StateValuesMTBDD(reach, model);
			} else if (r == 0 && relOp.equals("<")) {
				mainLog.print("\nWarning: checking for reward " + relOp + " " + r + " - formula trivially satisfies no states\n");
				return new StateValuesMTBDD(JDD.Constant(0), model);
			}
		}

		// determine whether min or max rewards needed
		if (relOp.equals(">") || relOp.equals(">=") || relOp.equals("min=")) {
			// min
			min = true;
		} else if (relOp.equals("<") || relOp.equals("<=") || relOp.equals("max=")) {
			// max
			min = false;
		} else {
			throw new PrismException("Can't use \"R=?\" for MDPs; use \"Rmin=?\" or \"Rmax=?\"");
		}

		// compute rewards
		expr2 = expr.getExpression();
		if (expr2 instanceof ExpressionTemporal) {
			switch (((ExpressionTemporal) expr2).getOperator()) {
			case ExpressionTemporal.R_I:
				rewards = checkRewardInst((ExpressionTemporal) expr2, stateRewards, transRewards, min);
				break;
			case ExpressionTemporal.R_F:
				sccBased = true;
				rewards = checkRewardReach((ExpressionTemporal) expr2, stateRewards, transRewards, min);
				break;
			}
		}
		if (rewards == null)
			throw new PrismException("Unrecognised operator in R operator");

		// print out rewards
		if (verbose) {
			mainLog.print("\n" + (min ? "Minimum" : "Maximum") + " rewards (non-zero only) for all states:\n");
			rewards.print(mainLog);
		}

		// For =? properties, just return values
		if (rb == null) {
			return rewards;
		}
		// Otherwise, compare against bound to get set of satisfying states
		else {
			sol = rewards.getBDDFromInterval(relOp, r);
			// remove unreachable states from solution
			JDD.Ref(reach);
			sol = JDD.And(sol, reach);
			// free vector
			rewards.clear();
			return new StateValuesMTBDD(sol, model);
		}
	}

	// Multi-objective

	protected StateValues checkExpressionMultiObjective(ExpressionFunc expr) throws PrismException
	{
		// Objective/target info
		int numTargets;
		List<Expression> targetExprs;
		List<JDDNode> targetDDs;
		List<Integer> relOps;
		List<Double> bounds;
		List<Integer> relOpsProb;
		List<Double> boundsProb;
		List<Integer> relOpsReward;
		List<Double> boundsReward;
		boolean[] reachExpr; // whether target is just reachability (true) or LTL (false)
		// LTL/product model stuff
		NondetModel modelProduct = null;
		NondetModel modelNew = null;
		JDDVars[] draDDRowVars = null;
		JDDVars[] draDDColVars = null;
		LTLModelChecker mcLtl;
		Expression[] ltl;
		DRA[] dra;
		// Misc
		ExpressionProb exprProb = null;
		ExpressionReward exprReward = null;
		ExpressionTemporal exprTemp;
		Expression pb; // probability bound (expression)
		double p = 0; // probability bound (actual value)
		String relOp; // relational operator
		JDDNode dd;
		int i, j;
		double prob;
		long l;
		boolean negateresult = false;
		int conflictformulae = 0;
		List<JDDNode> multitargetDDs = null;
		List<Integer> multitargetIDs = null;
		boolean hasMaxReward = false;
		boolean hasLTLconstraint = false;
		//boolean checkReach = true; // checking reachability or LTL model checking 

		// Check format and extract bounds/etc.
		numTargets = expr.getNumOperands();
		relOps = new ArrayList<Integer>(numTargets);
		bounds = new ArrayList<Double>(numTargets);
		relOpsProb = new ArrayList<Integer>(numTargets);
		boundsProb = new ArrayList<Double>(numTargets);
		relOpsReward = new ArrayList<Integer>(numTargets);
		boundsReward = new ArrayList<Double>(numTargets);
		List<JDDNode> rewards = new ArrayList<JDDNode>(numTargets);
		List<JDDNode> rewardsIndex = new ArrayList<JDDNode>(numTargets);
		//checkReach = true;
		reachExpr = new boolean[numTargets];
		for (i = 0; i < numTargets; i++) {
			// Only do P or R operators
			if (!(expr.getOperand(i) instanceof ExpressionProb || expr.getOperand(i) instanceof ExpressionReward))
				throw new PrismException("Multiple objectives must be P or R operators");
			//if (expr.getOperand(i) instanceof ExpressionReward && i > 0)
			//	throw new PrismException("R operator can only be used as first argument of multi function");
			if (expr.getOperand(i) instanceof ExpressionProb) {
				exprProb = (ExpressionProb) expr.getOperand(i);
				exprReward = null;
			} else {
				exprReward = (ExpressionReward) expr.getOperand(i);
				exprProb = null;
				Object rs = exprReward.getRewardStructIndex();
				JDDNode transRewards = null;
				if (model.getNumRewardStructs() == 0)
					throw new PrismException("Model has no rewards specified");
				if (rs == null) {
					transRewards = model.getTransRewards(0);
				} else if (rs instanceof Expression) {
					i = ((Expression) rs).evaluateInt(constantValues, null);
					rs = new Integer(i); // for better error reporting below
					transRewards = model.getTransRewards(i - 1);
				} else if (rs instanceof String) {
					transRewards = model.getTransRewards((String) rs);
				}
				if (transRewards == null)
					throw new PrismException("Invalid reward structure index \"" + rs + "\"");
				rewardsIndex.add(transRewards);
			}
			// Only do F operator for rewards
			if (exprReward != null) {
				exprTemp = (ExpressionTemporal) exprReward.getExpression();
				/*if (exprTemp.getOperator() != ExpressionTemporal.R_F) {
					throw new PrismException("Multi-objective only supports F operator for rewards");
				}*/
			}
			// No time-bounded variants of the temporal operators allowed for multi-objective (or LTL)
			try {
				expr.accept(new ASTTraverse()
				{
					public void visitPre(ExpressionTemporal e) throws PrismLangException
					{
						if (e.getLowerBound() != null)
							throw new PrismLangException(e.getOperatorSymbol());
						if (e.getUpperBound() != null)
							throw new PrismLangException(e.getOperatorSymbol());
					}
				});
			} catch (PrismLangException e) {
				String s = "Temporal operators (like " + e.getMessage() + ")";
				s += " cannot have time bounds for multi-objective properties";
				throw new PrismException(s);
			}
			// Get info from P/R operator
			// Store relational operator as an integer
			// Only do max=? or lower bounds
			relOp = (exprProb != null) ? exprProb.getRelOp() : exprReward.getRelOp();
			if (relOp.equals("max=")) {
				if(i>0)
					throw new PrismException("Only the first objective can be max/min query.");
				relOps.add((exprProb != null) ? 0 : 3);
				if(exprProb != null)
					relOpsProb.add(0);
				else {
					relOpsReward.add(3);
					hasMaxReward = true;
				}
			} else if (relOp.equals(">")) // currently do not support
				//relOps.add(1);
				throw new PrismException("Strict inequalities are not supported.");
			else if (relOp.equals(">=")) {
				relOps.add((exprProb != null) ? 2 : 4);
				if(exprProb != null) {
					relOpsProb.add(2);
					hasLTLconstraint = true;
				} else {
					relOpsReward.add(4);
					hasMaxReward = true;
				}
			} else if (relOp.equals("min=")) {
				if(i>0)
					throw new PrismException("Only the first objective can be max/min query.");
				relOps.add((exprProb != null) ? 5 : 8);
				if(exprProb != null) {
					relOpsProb.add(5);
					negateresult = true;
				} else
					relOpsReward.add(8);
			} else if (relOp.equals("<")) // currently do not support
				//relOps.add(6);
				throw new PrismException("Strict inequalities are not supported.");
			else if (relOp.equals("<=")) {
				relOps.add((exprProb != null) ? 7 : 9);
				if(exprProb != null) {
					relOpsProb.add(7);
					hasLTLconstraint = true;
				} else
					relOpsReward.add(9);
			} else
				throw new PrismException("Multiple objectives must be P/R operators with max/min=? or lower/upper probability bounds");
			// Store probability/rewards bound
			pb = (exprProb != null) ? exprProb.getProb() : exprReward.getReward();
			if (pb != null) {
				p = pb.evaluateDouble(constantValues, null);
				if ((exprProb != null) && (p < 0 || p > 1))
					throw new PrismException("Invalid probability bound " + p + " in P operator");
				if ((exprProb == null) && (p < 0))
					throw new PrismException("Invalid reward bound " + p + " in P operator");
				bounds.add(p);
				if(exprProb != null) {
					if(relOp.equals("<="))
						p = 1 - p;
					boundsProb.add(p);
				} else
					boundsReward.add(p);
			} else {
				bounds.add(-1.0);
				if(exprProb != null)
					boundsProb.add(-1.0);
				else
					boundsReward.add(-1.0);
			}
		}

		// Now go round again and extract targets
		// Also store which ones are just reachability (in reachExpr)
		targetExprs = new ArrayList<Expression>(numTargets);
		String[] targetName = new String[numTargets];
		for (i = 0; i < numTargets; i++) {
			if (expr.getOperand(i) instanceof ExpressionProb) {
				reachExpr[i] = false;
				conflictformulae++;
				exprProb = (ExpressionProb) expr.getOperand(i);
				/*if (exprProb.getExpression() instanceof ExpressionTemporal) {
					exprTemp = (ExpressionTemporal) exprProb.getExpression();
					if (relOps.get(i) < 5 && exprTemp.getOperator() == ExpressionTemporal.P_F && exprTemp.getOperand2().getType() instanceof TypeBool) {
						dd = checkExpressionDD(exprTemp.getOperand2());
						reachExpr[i] = checkWeakAbsorbing(dd, model);
						JDD.Deref(dd);
					}
				}*/
				if (reachExpr[i]) {
					targetExprs.add(((ExpressionTemporal) exprProb.getExpression()).getOperand2());
				} else {
					targetExprs.add(exprProb.getExpression());
				}
				targetName[i] = exprProb.getExpression().toString();
			} else {
				/*reachExpr[i] = false;
				exprReward = (ExpressionReward) expr.getOperand(i);
				if (exprReward.getExpression() instanceof ExpressionTemporal) {
					exprTemp = (ExpressionTemporal) exprReward.getExpression();
					if (exprTemp.getOperator() == ExpressionTemporal.R_F) {
						dd = checkExpressionDD(exprTemp.getOperand2());
						reachExpr[i] = checkWeakAbsorbing(dd, model);
						JDD.Deref(dd);
					}
				}
				if (reachExpr[i]) {
					targetExprs.add(((ExpressionTemporal) exprReward.getExpression()).getOperand2());
				} else {
					throw new PrismException("Can't handle reward properties that are not weak absorbing");
				}
				targetName[i] = exprReward.getExpression().toString();*/
				targetExprs.add(null);
				reachExpr[i] = true;
				targetName[i] = "";
			}
		}

		// Create array to store target DDs
		targetDDs = new ArrayList<JDDNode>(numTargets);

		// Multi-objective model checking

		// Create arrays to store LTL/DRA info
		ltl = new Expression[numTargets];
		dra = new DRA[numTargets];
		draDDRowVars = new JDDVars[numTargets];
		draDDColVars = new JDDVars[numTargets];

		// For LTL model checking routines
		mcLtl = new LTLModelChecker(prism);

		// Product is initially just the original model (we build it recursively)
		modelProduct = model;

		/*try{
			//prism.exportStatesToFile(modelProduct, Prism.EXPORT_PLAIN, null);
			prism.exportTransToFile(modelProduct, true, Prism.EXPORT_DOT_STATES, new java.io.File("model-multi.dot"));
		} catch(Exception e) {}*/
		
		// Go through objectives...
		boolean originalmodel = true;
		for (i = 0; i < numTargets; i++) {
			if (!reachExpr[i]) {
				// Model check maximal state formulas
				Vector<JDDNode> labelDDs = new Vector<JDDNode>();
				ltl[i] = mcLtl.checkMaximalStateFormulas(this, model, targetExprs.get(i).deepCopy(), labelDDs);

				// Convert LTL formula to deterministic Rabin automaton (DRA)
				// (always do max probs for multi-objective so never need to negate)
				mainLog.println("\nBuilding deterministic Rabin automaton (for " + targetName[i] + ")...");
				l = System.currentTimeMillis();
				dra[i] = relOps.get(i)<5 ? LTL2Rabin.ltl2rabin(ltl[i].convertForJltl2ba()) :
					LTL2Rabin.ltl2rabin(ltl[i].convertForJltl2ba().negate());
				mainLog.print("\nDRA has " + dra[i].size() + " states");
				mainLog.println(", " + dra[i].acceptance().size() + " pairs.");
				//dra[i].print(System.out);
				l = System.currentTimeMillis() - l;
				mainLog.println("\nTime for Rabin translation: " + l / 1000.0 + " seconds.");

				// Build product of MDP and automaton
				mainLog.println("\nConstructing MDP-DRA product...");
				draDDRowVars[i] = new JDDVars();
				draDDColVars[i] = new JDDVars();
				modelNew = mcLtl.constructProductMDP(dra[i], modelProduct, labelDDs, draDDRowVars[i], draDDColVars[i], false);
				// Deref old product (unless is the original model)
				if (i > 0 & !originalmodel)
					modelProduct.clear();
				// Store new product
				modelProduct = modelNew;
				originalmodel = false;

				modelProduct.printTransInfo(mainLog, prism.getExtraDDInfo());
				// Deref label BDDs
				for (j = 0; j < labelDDs.size(); j++) {
					JDD.Deref(labelDDs.get(j));
				}
			}
		}
		/*try {
			mainLog.println("Printing modelProduct...   ");
			prism.exportTransToFile(modelProduct, true, Prism.EXPORT_DOT_STATES, new java.io.File("product-multi.dot"));
			//prism.exportStatesToFile(modelProduct, Prism.EXPORT_PLAIN, null);
			//prism.exportTransToFile(modelProduct, true, Prism.EXPORT_PLAIN, null);
			mainLog.println("done...   ");
		} catch (Exception e) {
		}*/
		// Replace min by max and <= by >=
		for(i=0; i<relOpsProb.size(); i++)
			if(relOpsProb.get(i) == 5) {
				relOpsProb.remove(i);
				relOpsProb.add(i, 0);
			} else if(relOpsProb.get(i) == 7) {
				relOpsProb.remove(i);
				relOpsProb.add(i, 2);
			}
		// Print product info
		mainLog.println();
		modelProduct.printTransInfo(mainLog, prism.getExtraDDInfo());

		for(JDDNode rindex : rewardsIndex) {
			/*mainLog.println("Reward structure: ");
			JDD.PrintDebug(rindex, modelProduct.getNumDDRowVars() + modelProduct.getNumDDColVars() + modelProduct.getNumDDNondetVars());
			mainLog.println("Trans01: ");
			JDD.PrintDebug(modelProduct.getTrans01(), modelProduct.getNumDDRowVars() + modelProduct.getNumDDColVars() + modelProduct.getNumDDNondetVars());
			mainLog.println("Trans: ");
			JDD.PrintDebug(modelProduct.getTrans(), modelProduct.getNumDDRowVars() + modelProduct.getNumDDColVars() + modelProduct.getNumDDNondetVars());*/
			JDD.Ref(rindex);
			JDD.Ref(modelProduct.getTrans01());
			rewards.add(JDD.Apply(JDD.TIMES, rindex, modelProduct.getTrans01()));
			//mainLog.println("Combined: ");
			//JDD.PrintDebug(rewards.get(rewards.size()-1), modelProduct.getNumDDRowVars() + modelProduct.getNumDDColVars() + modelProduct.getNumDDNondetVars());
		} 
		
		// Removing actions with non-zero reward from the product for maximum cases
		if(hasMaxReward /*& hasLTLconstraint*/) {
			JDD.Ref(modelProduct.getReach());
			Vector<JDDNode> mecs = mcLtl.findEndComponents(modelProduct, modelProduct.getReach());
			JDDNode removedActions = JDD.Constant(0);
			JDDNode rmecs = JDD.Constant(0);
			boolean mecflags[] = new boolean[mecs.size()];
			int ec_num = 0;
			for(i=0; i<rewardsIndex.size(); i++) 
				if(relOpsReward.get(i) == 3 || relOpsReward.get(i) == 4) {
					JDD.Ref(rewardsIndex.get(i));
					JDDNode actions = JDD.GreaterThan(rewardsIndex.get(i), 0.0);
					//JDD.PrintDebug(rewardsIndex.get(i), 2);
					if(!actions.equals(JDD.ZERO)) 
						for(j = 0; j<mecs.size(); j++)	{
							JDDNode mec = mecs.get(j);
							JDD.Ref(mec);
							JDDNode mecactions = mcLtl.maxStableSetTrans1(modelProduct, mec);
							JDD.Ref(actions);
							mecactions = JDD.And(actions, mecactions);
							if(!mecactions.equals(JDD.ZERO)) {
								mecflags[j] = true;
								JDD.Ref(mec);
								rmecs = JDD.Or(rmecs, mec);
								ec_num++;
							}
							removedActions = JDD.Or(removedActions, mecactions);
						}
					JDD.Deref(actions);
				}
			for(JDDNode mec : mecs)	
				JDD.Deref(mec);
			// TODO: check if the model satisfies the LTL constraints 
			if(!rmecs.equals(JDD.ZERO)) {
				boolean constraintViolated = false;
				if(JDD.AreInterecting(modelProduct.getStart(), rmecs)) {
					constraintViolated = true;
					JDD.Deref(rmecs);
				} else {
					// find all action pointing to mecs from outside a
					//JDD.PrintDebug(rmecs, 2);
					JDD.Ref(rmecs);
					JDDNode rtarget = JDD.PermuteVariables(rmecs, modelProduct.getAllDDRowVars(), modelProduct.getAllDDColVars());
					//JDD.PrintDebug(rtarget, 2);
					JDD.Ref(modelProduct.getTrans01());
					rtarget = JDD.And(modelProduct.getTrans01(), rtarget);
					//JDD.PrintDebug(rtarget, 2);
					rtarget = JDD.And(rtarget, JDD.Not(rmecs));
					//JDD.PrintDebug(rtarget, 2);
					//rtarget = JDD.ThereExists(rtarget, modelProduct.getAllDDColVars());
					//JDD.PrintDebug(rtarget, 2);
					// find target states for LTL formulae
					Vector<JDDNode> tmptargetDDs = new Vector<JDDNode>();
					List<JDDNode> tmpmultitargetDDs = new ArrayList<JDDNode>();
					List<Integer> tmpmultitargetIDs = new ArrayList<Integer>();
					ArrayList<DRA> tmpdra = new ArrayList<DRA>();
					ArrayList<JDDVars> tmpdraDDRowVars = new ArrayList<JDDVars>();
					ArrayList<JDDVars> tmpdraDDColVars = new ArrayList<JDDVars>();
					int count = 0;
					for(i=0; i<numTargets; i++)
						if (!reachExpr[i] && relOps.get(i) != 0 && relOps.get(i) != 5) {
							tmpdra.add(dra[i]);
							tmpdraDDRowVars.add(draDDRowVars[i]);
							tmpdraDDColVars.add(draDDColVars[i]);
							count++;
						}
					if(count > 0) {
						// TUDO: distinguish whether rtarget is empty
						DRA newdra[] = new DRA[count];
						tmpdra.toArray(newdra);
						JDDVars newdraDDRowVars[] = new JDDVars[count];
						tmpdraDDRowVars.toArray(newdraDDRowVars);
						JDDVars newdraDDColVars[] = new JDDVars[count];
						tmpdraDDColVars.toArray(newdraDDColVars);
						
						findTargetStates(modelProduct, mcLtl, count, count, 
								new boolean[count], newdra,
								newdraDDRowVars, newdraDDColVars,
								tmptargetDDs, tmpmultitargetDDs, tmpmultitargetIDs);
						
						ArrayList<Integer> tmprelOpsProb = new ArrayList<Integer>(relOpsProb);
						ArrayList<Double> tmpboundsProb = new ArrayList<Double>(boundsProb);
						if(relOpsProb.get(0) == 0) {
							tmprelOpsProb.remove(0);
							tmpboundsProb.remove(0);
						}
						ArrayList<JDDNode> tmprewards = new ArrayList<JDDNode>(1);
						ArrayList<Integer> tmprelOpsReward = new ArrayList<Integer>(1);
						ArrayList<Double> tmpboundsReward = new ArrayList<Double>(1);
						tmprewards.add(rtarget);
						tmprelOpsReward.add(3);
						tmpboundsReward.add(-1.0);
						prob = computeMultiReachProbs(modelProduct, mcLtl, tmprewards, modelProduct.getStart(), tmptargetDDs, 
								tmpmultitargetDDs, tmpmultitargetIDs, tmprelOpsProb, tmpboundsProb, tmprelOpsReward, tmpboundsReward, count > 1);
						if(prob > 0.0) { // LTL formulae can be satisfied
							constraintViolated = true;
						} else if(Double.isNaN(prob))
							throw new PrismException("The LTL formulae in multi-objective query cannot be satisfied!\n");
					} else {
						// end components with non-zero rewards can always be reached
						constraintViolated = true;
					}
				
					//JDD.Deref(rtarget);
					for(JDDNode tt : tmptargetDDs)
						JDD.Deref(tt);
					for(JDDNode tt : tmpmultitargetDDs)
						JDD.Deref(tt);
				}
				if(constraintViolated) {
					mainLog.println("-----------------------------------------------------------------");
					mainLog.println("WARNING: There are reachable end compoments with non-zero reward!");
					mainLog.println("-----------------------------------------------------------------");
					mainLog.flush();
				}
				
				/*try {
					mainLog.println("Printing modelProduct...   ");
					prism.exportTransToFile(modelProduct, true, Prism.EXPORT_DOT_STATES, new java.io.File("product-multi.dot"));
					//prism.exportStatesToFile(modelProduct, Prism.EXPORT_PLAIN, null);
					//prism.exportTransToFile(modelProduct, true, Prism.EXPORT_PLAIN, null);
					mainLog.println("done...   ");
				} catch (Exception e) {
				}*/
				JDD.Ref(removedActions);
				modelProduct.trans = JDD.Apply(JDD.TIMES, modelProduct.trans, JDD.Not(removedActions));
				modelProduct.trans01 = JDD.Apply(JDD.TIMES, modelProduct.trans01, JDD.Not(removedActions));
				/*try {
					mainLog.println("Printing modelProduct...   ");
					prism.exportTransToFile(modelProduct, true, Prism.EXPORT_DOT_STATES, new java.io.File("product-multi-new.dot"));
					//prism.exportStatesToFile(modelProduct, Prism.EXPORT_PLAIN, null);
					//prism.exportTransToFile(modelProduct, true, Prism.EXPORT_PLAIN, null);
					mainLog.println("done...   ");
				} catch (Exception e) {
				}*/
			} else {
				JDD.Deref(rmecs);
				JDD.Deref(removedActions);
			}
		}
		
		// Remove all non-zero reward from trans in order to search for zero reward end components
		JDDNode tmptrans = modelProduct.getTrans();
		//JDD.Ref(tmptrans);
		JDDNode tmptrans01 = modelProduct.getTrans01();
		//JDD.Ref(tmptrans01);
		boolean transchanged = false;
		for(i=0; i<rewardsIndex.size(); i++) 
			if(relOpsReward.get(i) == 8 || relOpsReward.get(i) == 9) {
				// Get non-zero reward actions
				JDD.Ref(rewardsIndex.get(i));
				JDDNode actions = JDD.GreaterThan(rewardsIndex.get(i), 0.0);
				//mainLog.println("actions: ");
				//JDD.PrintDebug(actions, modelProduct.getNumDDRowVars() + modelProduct.getNumDDColVars() + modelProduct.getNumDDNondetVars());
				if(!actions.equals(JDD.ZERO)) {
					//mainLog.println("Removing non-zero reward actions in reward #" + i);
					JDD.Ref(actions);
					if(!transchanged)
						JDD.Ref(modelProduct.getTrans());
					JDDNode tmp = JDD.ITE(actions, JDD.Constant(0), modelProduct.getTrans());
					modelProduct.trans = tmp;
					//mainLog.println("new trans: ");
					//JDD.PrintDebug(modelProduct.getTrans(), modelProduct.getNumDDRowVars() + modelProduct.getNumDDColVars() + modelProduct.getNumDDNondetVars());
					if(!transchanged)
						JDD.Ref(modelProduct.getTrans01());
					tmp = JDD.ITE(actions, JDD.Constant(0), modelProduct.getTrans01());
					modelProduct.trans01 = tmp;
					transchanged = true;
					//mainLog.println("New trans01: ");
					//JDD.PrintDebug(modelProduct.getTrans01(), modelProduct.getNumDDRowVars() + modelProduct.getNumDDColVars() + modelProduct.getNumDDNondetVars());
				}
			}

		// Compute all maximal end components
		ArrayList<ArrayList<JDDNode>> allstatesH = new ArrayList<ArrayList<JDDNode>>(numTargets);
		ArrayList<ArrayList<JDDNode>> allstatesL = new ArrayList<ArrayList<JDDNode>>(numTargets);
		JDDNode acceptanceVector_H = JDD.Constant(0);
		JDDNode acceptanceVector_L = JDD.Constant(0);
		for (i = 0; i < numTargets; i++) {
			if (!reachExpr[i]) {
				ArrayList<JDDNode> statesH = new ArrayList<JDDNode>();
				ArrayList<JDDNode> statesL = new ArrayList<JDDNode>();
				for (int k = 0; k < dra[i].acceptance().size(); k++) {
					JDDNode tmpH = JDD.Constant(0);
					JDDNode tmpL = JDD.Constant(0);
					for (j = 0; j < dra[i].size(); j++) {
						if (!dra[i].acceptance().isStateInAcceptance_U(k, j)) {
							tmpH = JDD.SetVectorElement(tmpH, draDDRowVars[i], j, 1.0);
						}
						if (dra[i].acceptance().isStateInAcceptance_L(k, j)) {
							tmpL = JDD.SetVectorElement(tmpL, draDDRowVars[i], j, 1.0);
						}
					}
					statesH.add(tmpH);
					JDD.Ref(tmpH);
					acceptanceVector_H = JDD.Or(acceptanceVector_H, tmpH);
					statesL.add(tmpL);
					JDD.Ref(tmpL);
					acceptanceVector_L = JDD.Or(acceptanceVector_L, tmpL);
				}
				allstatesH.add(i, statesH);
				allstatesL.add(i, statesL);
			} else {
				allstatesH.add(i, null);
				allstatesL.add(i, null);
			}
		}

		// Find accepting maximum end components for each LTL formula
		Vector<JDDNode> allecs = null;
		JDDNode newtrans = null;
		JDDNode newtrans01 = null;
		Vector<JDDFairnessConstraint> sf = null;
		Vector<JDDFairnessConstraint> wf = null;
		if (model.getFairnessConstraints() != null) {
			ExpressionFunc fe = (ExpressionFunc) model.getFairnessConstraints();
			sf = new Vector<JDDFairnessConstraint>();
			wf = new Vector<JDDFairnessConstraint>();
			for (int t = 0; t < fe.getNumOperands(); t++) {
				ExpressionFairness operand = (ExpressionFairness) fe.getOperand(t);
				JDDFairnessConstraint ddfc = new JDDFairnessConstraint(operand, modelProduct);
				if (operand.isStrongFairness())
					sf.add(ddfc);
				else
					wf.add(ddfc);
			}
			// we have to find all maximal end components in the original product
			JDD.Ref(modelProduct.getReach());
			allecs = mcLtl.findEndComponents(modelProduct, modelProduct.getReach());

			Vector<JDDNode> recs = new Vector<JDDNode>();
			// identify each maximal end component
			JDDNode acc = JDD.Constant(0);
			for (JDDNode set : allecs) {
				//mainLog.println("------------- End component: -------------");
				//StateListMTBDD vl = new StateListMTBDD(set, model);
				//vl.print(mainLog);

				boolean isfair = mcLtl.isEcFair(modelProduct, set, sf, wf);
				if (isfair) {
					JDD.Ref(set);
					acc = JDD.Or(acc, set);
					//recs.add(set);
					//mainLog.println("------------- is a fair component! -------------");
				} else {
					//mainLog.println("------------- is not a fair component! -------------");
				}
			}
			// compute reachability from all states to fair components
			mainLog.println("\nPerforming reachability analysis to fairness components...");
			boolean isRealisable = false;
			if (!acc.equals(JDD.ZERO)) {
				//NondetModelChecker mcProduct = new NondetModelChecker(prism, modelProduct, null);

				// We just need to compute the "no" nodes
				JDDNode no = PrismMTBDD.Prob0A(modelProduct.getTrans01(), modelProduct.getReach(), modelProduct.getAllDDRowVars(), modelProduct
						.getAllDDColVars(), modelProduct.getAllDDNondetVars(), modelProduct.getReach(), acc);
				double numno = JDD.GetNumMinterms(no, modelProduct.getAllDDRowVars().n());
				if (numno > 0) {
					mainLog.println("number of states not reaching FMEC = " + JDD.GetNumMintermsString(no, modelProduct.getAllDDRowVars().n()));
					//mainLog.println("------------- no: -------------");
					//vl = new StateListMTBDD(no, model);
					//vl.print(mainLog);
				} else
					isRealisable = true;
				JDD.Deref(acc);

				if (isRealisable)
					mainLog.println("Congratulations! The fairness constraint is realisable.");
				else
					mainLog.println("Warning: The fairness constraint is NOT realisable!");

				if (isRealisable) {
					JDD.Deref(no);
					//recs.addAll(allecs);
				} else {
					//recs.addAll(allecs);
					for (JDDNode ec : allecs) {
						if (!JDD.AreInterecting(ec, no)) {
							JDD.Ref(ec);
							recs.add(ec);
						}
					}
					for (JDDNode ec : allecs)
						JDD.Deref(ec);
					allecs.clear();
					allecs.addAll(recs);

					// remove no states
					JDD.Ref(modelProduct.getReach());
					JDDNode realisablestates = JDD.And(modelProduct.getReach(), JDD.Not(no));
					// recursively remove states that do not have probability 1 to reach fair EC
					// realisablestates = findMaximalStableSet(model, realisablestates);
					JDD.Ref(realisablestates);
					JDD.Ref(modelProduct.getTrans());
					// Select transitions starting in current
					JDDNode currTrans = JDD.Apply(JDD.TIMES, modelProduct.getTrans(), realisablestates);
					// Select transitions starting in current and ending in current
					//JDD.Ref(realisablestates);
					JDDNode tmp = JDD.PermuteVariables(realisablestates, modelProduct.getAllDDRowVars(), modelProduct.getAllDDColVars());
					tmp = JDD.Apply(JDD.TIMES, currTrans, tmp);
					// Sum all successor probabilities for each (state, action) tuple
					tmp = JDD.SumAbstract(tmp, modelProduct.getAllDDColVars());
					// If the sum for a (state,action) tuple is 1,
					// there is an action that remains in the stable set with prob 1
					tmp = JDD.GreaterThan(tmp, 1 - prism.getSumRoundOff());

					JDD.Ref(modelProduct.getTrans());
					JDD.Ref(tmp);
					//JDDNode newtrans = JDD.Apply(JDD.TIMES, model.getTrans(), tmp);
					newtrans = JDD.And(modelProduct.getTrans(), tmp);

					JDD.Ref(modelProduct.getTrans01());
					newtrans01 = JDD.And(modelProduct.getTrans01(), tmp);

					JDD.Deref(modelProduct.getTrans());
					JDD.Deref(modelProduct.getTrans01());
					modelProduct.trans = newtrans;
					modelProduct.trans01 = newtrans01;

				}
			} else {
				//JDD.Deref(acc);
				for (JDDNode ec : allecs)
					JDD.Deref(ec);
				allecs.clear();
			}
			JDD.Deref(acceptanceVector_H);
			JDD.Deref(acceptanceVector_L);
		} else { //use acceptanceVector_H and acceptanceVector_L to speed up scc computation
			/*// check number of states in each scc
			allecs = mcLtl.findEndComponents(modelProduct, modelProduct.getReach());
			StateListMTBDD vl;
			int totalNum = 0;
			for(JDDNode set : allecs) {
				vl = new StateListMTBDD(set, modelProduct);
				totalNum += vl.size() - 1;
			}
			mainLog.println("Total number of states can be saved: " + totalNum);*/

			JDD.Ref(acceptanceVector_H);
			JDD.Ref(modelProduct.getTrans01());
			JDDNode candidateStates = JDD.Apply(JDD.TIMES, modelProduct.getTrans01(), acceptanceVector_H);
			for (i = 0; i < numTargets; i++)
				if (!reachExpr[i]) {
					acceptanceVector_H = JDD.PermuteVariables(acceptanceVector_H, draDDRowVars[i], draDDColVars[i]);
				}
			candidateStates = JDD.Apply(JDD.TIMES, candidateStates, acceptanceVector_H);
			candidateStates = JDD.ThereExists(candidateStates, modelProduct.getAllDDColVars());
			candidateStates = JDD.ThereExists(candidateStates, modelProduct.getAllDDNondetVars());
			// find all maximal end components
			allecs = mcLtl.findEndComponents(modelProduct, candidateStates, acceptanceVector_L);
			JDD.Deref(acceptanceVector_L);
		}

		for (i = 0; i < numTargets; i++) {
			if (!reachExpr[i]) {
				mainLog.println("\nFinding accepting end components for " + targetName[i] + "...");
				l = System.currentTimeMillis();
				// increase ref count for checking conflict formulas
				if(conflictformulae > 1) {
					List<JDDNode> tmpLH = allstatesH.get(i);
					for(JDDNode n : tmpLH)
						JDD.Ref(n);
					tmpLH = allstatesL.get(i);
					for(JDDNode n : tmpLH)
						JDD.Ref(n);
				}
				if (model.getFairnessConstraints() == null) {
					targetDDs.add(mcLtl.findMultiAcceptingStates(dra[i], modelProduct, draDDRowVars[i], draDDColVars[i], false, allecs, allstatesH.get(i),
							allstatesL.get(i)));
					//targetDDs.add(mcLtl.findMultiAcceptingStates(dra[i], modelProduct, draDDRowVars[i], draDDColVars[i], false));
				} else {
					targetDDs.add(mcLtl.findFairMultiAcceptingStates(dra[i], modelProduct, draDDRowVars[i], draDDColVars[i], false, sf, wf, allecs, allstatesH
							.get(i), allstatesL.get(i)));
				}
				//targetDDs.add(mcLtl.findMultiAcceptingStates(dra[i], modelProduct, draDDRowVars[i], draDDColVars[i], 
				//	false, sf, wf));
				l = System.currentTimeMillis() - l;
				mainLog.println("Time for end component identification: " + l / 1000.0 + " seconds.");
			} else {
				// Fixme: maybe not efficient
				if(targetExprs.get(i) != null) {
					dd = checkExpressionDD(targetExprs.get(i));
					JDD.Ref(modelProduct.getReach());
					dd = JDD.And(dd, modelProduct.getReach());
					targetDDs.add(dd);
				}
			}
			//mainLog.print("\n    number of targets = " + JDD.GetNumMintermsString(targetDDs.get(i), modelProduct.getAllDDRowVars().n()));
			//if(i>0)
			//	mainLog.print("\n  total number of targets = " + 
			//			JDD.GetNumMintermsString(JDD.Or(targetDDs.get(i), targetDDs.get(i-1)), modelProduct.getAllDDRowVars().n()));
		}
		
		// check if there are conflicts in objectives
		if(conflictformulae > 1) {
			DRA[] tmpdra = new DRA[conflictformulae];
			JDDVars[] tmpdraDDRowVars = new JDDVars[conflictformulae];
			JDDVars[] tmpdraDDColVars = new JDDVars[conflictformulae];
			List<JDDNode> tmptargetDDs = new ArrayList<JDDNode>(conflictformulae);
			List<List<JDDNode>> tmpallstatesH = new ArrayList<List<JDDNode>>(conflictformulae);
			List<List<JDDNode>> tmpallstatesL = new ArrayList<List<JDDNode>>(conflictformulae);
			int count = 0;
			for(i=0; i<numTargets; i++)
				if (!reachExpr[i]) {
					tmpdra[count] = dra[i];
					tmpdraDDRowVars[count] = draDDRowVars[i];
					tmpdraDDColVars[count] = draDDColVars[i];
					// tricky part; double check when reachExpr is set for efficiency
					tmptargetDDs.add(targetDDs.get(count));
					tmpallstatesH.add(allstatesH.get(i));
					tmpallstatesL.add(allstatesL.get(i));
					count++;
				}
			multitargetDDs = new ArrayList<JDDNode>();
			List<List<Integer>> tmpmultitargetIDs = new ArrayList<List<Integer>>();
			
			mcLtl.findMultiConflictAcceptingStates(tmpdra, modelProduct, 
					tmpdraDDRowVars, tmpdraDDColVars, tmptargetDDs, tmpallstatesH, tmpallstatesL,
					multitargetDDs, tmpmultitargetIDs);
			// again. double check when reachExpr is set
			count = 0;
			for(i=0; i<numTargets; i++)
				if (!reachExpr[i]) {
					targetDDs.remove(count);
					targetDDs.add(count, tmptargetDDs.get(count));
					count++;
				}
			
			// deal with reachability targets
			multitargetIDs = new ArrayList<Integer>(tmpmultitargetIDs.size());
			for(i=0; i<tmpmultitargetIDs.size(); i++) {
				multitargetIDs.add(changeToInteger(tmpmultitargetIDs.get(i)));
				//System.out.println("multitargetIDs["+i+"] = " + multitargetIDs.get(i));
			}
			
			for(i=0; i<numTargets; i++)
				if (!reachExpr[i]) {
					List<JDDNode> tmpLH = allstatesH.get(i);
					for(JDDNode n : tmpLH)
						JDD.Deref(n);
					tmpLH = allstatesL.get(i);
					for(JDDNode n : tmpLH)
						JDD.Deref(n);
				}
		}

		for (JDDNode ec : allecs)
			JDD.Deref(ec);
		/*for(i = 0; i < numTargets; i++) 
			if(!reachExpr[i]) {
				ArrayList<JDDNode> statesH = allstatesH.get(i);
				ArrayList<JDDNode> statesL = allstatesL.get(i);
				for (int k = 0; k < statesH.size(); k++) {
					JDD.Deref(statesL.get(k));
					JDD.Deref(statesH.get(k));
				}
			}*/
		if (sf != null) {
			for (int t = 0; t < sf.size(); t++)
				sf.get(t).derefHK();
			for (int t = 0; t < wf.size(); t++)
				wf.get(t).derefHK();
		}

		//new StateListMTBDD(modelProduct.getReach(), modelProduct).print(mainLog);
		//try { prism.exportTransToFile(modelProduct, true, Prism.EXPORT_DOT_STATES, new java.io.File("product.dot")); } catch(Exception e) {}

		// Note: for multi-objective model checking, we construct the product MDP for only a single initial state
		// (unlike for normal LTL model checking) so it is safe to use modelProduct.getStart() here to pass in the initial states.
		
		// Add a dummy LTL formula to get generate target states when there is no LTL formula in the query
		if(targetDDs.isEmpty()) {
			JDD.Ref(modelProduct.getReach());
			Vector<JDDNode> tmpecs = mcLtl.findEndComponents(modelProduct, modelProduct.getReach());
			JDDNode acceptingStates = JDD.Constant(0);
			for (JDDNode set : tmpecs) 
				acceptingStates = JDD.Or(acceptingStates, set);
			targetDDs.add(acceptingStates);
			relOpsProb.add(2);
			boundsProb.add(0.0);
		}
		
		// Put unmodified trans and trans01 to modelProduct
		if(transchanged) {
			JDD.Deref(modelProduct.getTrans());
			JDD.Deref(modelProduct.getTrans01());
			modelProduct.trans = tmptrans;
			modelProduct.trans01 = tmptrans01;
		}
		
		prob = computeMultiReachProbs(modelProduct, mcLtl, rewards, modelProduct.getStart(), targetDDs, multitargetDDs, multitargetIDs, relOpsProb, boundsProb, relOpsReward, boundsReward, conflictformulae > 1);
		/*prob = 1.0;
		for(JDDNode t : rewards)
			JDD.Deref(t);*/
		// THIS SHOULD BE COVERED BELOW NOW
		// Deref, clean up
		/*for (i = 0; i < targetDDs.size(); i++) {
			draDDRowVars[i].derefAll();
			draDDColVars[i].derefAll();
		}*/

		// Deref, clean up
		if (modelProduct != null && modelProduct != model)
			modelProduct.clear();
		for (i = 0; i < numTargets; i++) {
			if (!reachExpr[i]) {
				draDDRowVars[i].derefAll();
				draDDColVars[i].derefAll();
			}
		}
		for(JDDNode t : targetDDs)
			JDD.Deref(t);
		if(multitargetDDs != null)
			for(JDDNode t : multitargetDDs)
				JDD.Deref(t);

		// Return result. Note: we only compute the value for the (single) initial state.
		if(negateresult) {
			prob = 1 - prob;
		}
		
		return new StateValuesMTBDD(JDD.Constant(prob), model);
	}

	private int computeCombinations(int n, int r) {
		if(n<r || r < 1 || n < 1)
			return 0;
		if(n==r)
			return 1;
		if(r==1)
			return n;
		return computeCombinations(n, r-1)*(n-r+1)/r;
	}
	
	
	protected void findTargetStates(NondetModel modelProduct, LTLModelChecker mcLtl, int numTargets, int conflictformulae, 
			boolean reachExpr[], DRA dra[], JDDVars draDDRowVars[], JDDVars draDDColVars[],
			Vector<JDDNode> targetDDs, List<JDDNode> multitargetDDs, List<Integer> multitargetIDs)  throws PrismException {
		int i, j;
		long l;

		// Compute all maximal end components
		ArrayList<ArrayList<JDDNode>> allstatesH = new ArrayList<ArrayList<JDDNode>>(numTargets);
		ArrayList<ArrayList<JDDNode>> allstatesL = new ArrayList<ArrayList<JDDNode>>(numTargets);
		JDDNode acceptanceVector_H = JDD.Constant(0);
		JDDNode acceptanceVector_L = JDD.Constant(0);
		for (i = 0; i < numTargets; i++) {
			if (!reachExpr[i]) {
				ArrayList<JDDNode> statesH = new ArrayList<JDDNode>();
				ArrayList<JDDNode> statesL = new ArrayList<JDDNode>();
				for (int k = 0; k < dra[i].acceptance().size(); k++) {
					JDDNode tmpH = JDD.Constant(0);
					JDDNode tmpL = JDD.Constant(0);
					for (j = 0; j < dra[i].size(); j++) {
						if (!dra[i].acceptance().isStateInAcceptance_U(k, j)) {
							tmpH = JDD.SetVectorElement(tmpH, draDDRowVars[i], j, 1.0);
						}
						if (dra[i].acceptance().isStateInAcceptance_L(k, j)) {
							tmpL = JDD.SetVectorElement(tmpL, draDDRowVars[i], j, 1.0);
						}
					}
					statesH.add(tmpH);
					JDD.Ref(tmpH);
					acceptanceVector_H = JDD.Or(acceptanceVector_H, tmpH);
					statesL.add(tmpL);
					JDD.Ref(tmpL);
					acceptanceVector_L = JDD.Or(acceptanceVector_L, tmpL);
				}
				allstatesH.add(i, statesH);
				allstatesL.add(i, statesL);
			} else {
				allstatesH.add(i, null);
				allstatesL.add(i, null);
			}
		}

		// Find accepting maximum end components for each LTL formula
		Vector<JDDNode> allecs = null;
		JDDNode newtrans = null;
		JDDNode newtrans01 = null;
		Vector<JDDFairnessConstraint> sf = null;
		Vector<JDDFairnessConstraint> wf = null;
		if (modelProduct.getFairnessConstraints() != null) {
			ExpressionFunc fe = (ExpressionFunc) modelProduct.getFairnessConstraints();
			sf = new Vector<JDDFairnessConstraint>();
			wf = new Vector<JDDFairnessConstraint>();
			for (int t = 0; t < fe.getNumOperands(); t++) {
				ExpressionFairness operand = (ExpressionFairness) fe.getOperand(t);
				JDDFairnessConstraint ddfc = new JDDFairnessConstraint(operand, modelProduct);
				if (operand.isStrongFairness())
					sf.add(ddfc);
				else
					wf.add(ddfc);
			}
			// we have to find all maximal end components in the original product
			JDD.Ref(modelProduct.getReach());
			allecs = mcLtl.findEndComponents(modelProduct, modelProduct.getReach());

			Vector<JDDNode> recs = new Vector<JDDNode>();
			// identify each maximal end component
			JDDNode acc = JDD.Constant(0);
			for (JDDNode set : allecs) {
				//mainLog.println("------------- End component: -------------");
				//StateListMTBDD vl = new StateListMTBDD(set, model);
				//vl.print(mainLog);

				boolean isfair = mcLtl.isEcFair(modelProduct, set, sf, wf);
				if (isfair) {
					JDD.Ref(set);
					acc = JDD.Or(acc, set);
					//recs.add(set);
					//mainLog.println("------------- is a fair component! -------------");
				} else {
					//mainLog.println("------------- is not a fair component! -------------");
				}
			}
			// compute reachability from all states to fair components
			mainLog.println("\nPerforming reachability analysis to fairness components...");
			boolean isRealisable = false;
			if (!acc.equals(JDD.ZERO)) {
				//NondetModelChecker mcProduct = new NondetModelChecker(prism, modelProduct, null);

				// We just need to compute the "no" nodes
				JDDNode no = PrismMTBDD.Prob0A(modelProduct.getTrans01(), modelProduct.getReach(), modelProduct.getAllDDRowVars(), modelProduct
						.getAllDDColVars(), modelProduct.getAllDDNondetVars(), modelProduct.getReach(), acc);
				double numno = JDD.GetNumMinterms(no, modelProduct.getAllDDRowVars().n());
				if (numno > 0) {
					mainLog.println("number of states not reaching FMEC = " + JDD.GetNumMintermsString(no, modelProduct.getAllDDRowVars().n()));
					//mainLog.println("------------- no: -------------");
					//vl = new StateListMTBDD(no, model);
					//vl.print(mainLog);
				} else
					isRealisable = true;
				JDD.Deref(acc);

				if (isRealisable)
					mainLog.println("Congratulations! The fairness constraint is realisable.");
				else
					mainLog.println("Warning: The fairness constraint is NOT realisable!");

				if (isRealisable) {
					JDD.Deref(no);
					//recs.addAll(allecs);
				} else {
					//recs.addAll(allecs);
					for (JDDNode ec : allecs) {
						if (!JDD.AreInterecting(ec, no)) {
							JDD.Ref(ec);
							recs.add(ec);
						}
					}
					for (JDDNode ec : allecs)
						JDD.Deref(ec);
					allecs.clear();
					allecs.addAll(recs);

					// remove no states
					JDD.Ref(modelProduct.getReach());
					JDDNode realisablestates = JDD.And(modelProduct.getReach(), JDD.Not(no));
					// recursively remove states that do not have probability 1 to reach fair EC
					// realisablestates = findMaximalStableSet(model, realisablestates);
					JDD.Ref(realisablestates);
					JDD.Ref(modelProduct.getTrans());
					// Select transitions starting in current
					JDDNode currTrans = JDD.Apply(JDD.TIMES, modelProduct.getTrans(), realisablestates);
					// Select transitions starting in current and ending in current
					//JDD.Ref(realisablestates);
					JDDNode tmp = JDD.PermuteVariables(realisablestates, modelProduct.getAllDDRowVars(), modelProduct.getAllDDColVars());
					tmp = JDD.Apply(JDD.TIMES, currTrans, tmp);
					// Sum all successor probabilities for each (state, action) tuple
					tmp = JDD.SumAbstract(tmp, modelProduct.getAllDDColVars());
					// If the sum for a (state,action) tuple is 1,
					// there is an action that remains in the stable set with prob 1
					tmp = JDD.GreaterThan(tmp, 1 - prism.getSumRoundOff());

					JDD.Ref(modelProduct.getTrans());
					JDD.Ref(tmp);
					//JDDNode newtrans = JDD.Apply(JDD.TIMES, model.getTrans(), tmp);
					newtrans = JDD.And(modelProduct.getTrans(), tmp);

					JDD.Ref(modelProduct.getTrans01());
					newtrans01 = JDD.And(modelProduct.getTrans01(), tmp);

					JDD.Deref(modelProduct.getTrans());
					JDD.Deref(modelProduct.getTrans01());
					modelProduct.trans = newtrans;
					modelProduct.trans01 = newtrans01;

				}
			} else {
				//JDD.Deref(acc);
				for (JDDNode ec : allecs)
					JDD.Deref(ec);
				allecs.clear();
			}
			JDD.Deref(acceptanceVector_H);
			JDD.Deref(acceptanceVector_L);
		} else { //use acceptanceVector_H and acceptanceVector_L to speed up scc computation
			/*// check number of states in each scc
			allecs = mcLtl.findEndComponents(modelProduct, modelProduct.getReach());
			StateListMTBDD vl;
			int totalNum = 0;
			for(JDDNode set : allecs) {
				vl = new StateListMTBDD(set, modelProduct);
				totalNum += vl.size() - 1;
			}
			mainLog.println("Total number of states can be saved: " + totalNum);*/

			JDD.Ref(acceptanceVector_H);
			JDD.Ref(modelProduct.getTrans01());
			JDDNode candidateStates = JDD.Apply(JDD.TIMES, modelProduct.getTrans01(), acceptanceVector_H);
			for (i = 0; i < numTargets; i++)
				if (!reachExpr[i]) {
					acceptanceVector_H = JDD.PermuteVariables(acceptanceVector_H, draDDRowVars[i], draDDColVars[i]);
				}
			candidateStates = JDD.Apply(JDD.TIMES, candidateStates, acceptanceVector_H);
			candidateStates = JDD.ThereExists(candidateStates, modelProduct.getAllDDColVars());
			candidateStates = JDD.ThereExists(candidateStates, modelProduct.getAllDDNondetVars());
			// find all maximal end components
			allecs = mcLtl.findEndComponents(modelProduct, candidateStates, acceptanceVector_L);
			JDD.Deref(acceptanceVector_L);
		}

		for (i = 0; i < numTargets; i++) {
			if (!reachExpr[i]) {
				//mainLog.println("\nFinding accepting end components for " + targetName[i] + "...");
				l = System.currentTimeMillis();
				// increase ref count for checking conflict formulas
				if(conflictformulae > 1) {
					List<JDDNode> tmpLH = allstatesH.get(i);
					for(JDDNode n : tmpLH)
						JDD.Ref(n);
					tmpLH = allstatesL.get(i);
					for(JDDNode n : tmpLH)
						JDD.Ref(n);
				}
				if (model.getFairnessConstraints() == null) {
					targetDDs.add(mcLtl.findMultiAcceptingStates(dra[i], modelProduct, draDDRowVars[i], draDDColVars[i], false, allecs, allstatesH.get(i),
							allstatesL.get(i)));
					//targetDDs.add(mcLtl.findMultiAcceptingStates(dra[i], modelProduct, draDDRowVars[i], draDDColVars[i], false));
				} else {
					targetDDs.add(mcLtl.findFairMultiAcceptingStates(dra[i], modelProduct, draDDRowVars[i], draDDColVars[i], false, sf, wf, allecs, allstatesH
							.get(i), allstatesL.get(i)));
				}
				//targetDDs.add(mcLtl.findMultiAcceptingStates(dra[i], modelProduct, draDDRowVars[i], draDDColVars[i], 
				//	false, sf, wf));
				l = System.currentTimeMillis() - l;
				mainLog.println("Time for end component identification: " + l / 1000.0 + " seconds.");
			} /*else {
				// Fixme: maybe not efficient
				if(targetExprs.get(i) != null) {
					dd = checkExpressionDD(targetExprs.get(i));
					JDD.Ref(modelProduct.getReach());
					dd = JDD.And(dd, modelProduct.getReach());
					targetDDs.add(dd);
				}
			}*/
			//mainLog.print("\n    number of targets = " + JDD.GetNumMintermsString(targetDDs.get(i), modelProduct.getAllDDRowVars().n()));
			//if(i>0)
			//	mainLog.print("\n  total number of targets = " + 
			//			JDD.GetNumMintermsString(JDD.Or(targetDDs.get(i), targetDDs.get(i-1)), modelProduct.getAllDDRowVars().n()));
		}
		
		// check if there are conflicts in objectives
		if(conflictformulae > 1) {
			DRA[] tmpdra = new DRA[conflictformulae];
			JDDVars[] tmpdraDDRowVars = new JDDVars[conflictformulae];
			JDDVars[] tmpdraDDColVars = new JDDVars[conflictformulae];
			List<JDDNode> tmptargetDDs = new ArrayList<JDDNode>(conflictformulae);
			List<List<JDDNode>> tmpallstatesH = new ArrayList<List<JDDNode>>(conflictformulae);
			List<List<JDDNode>> tmpallstatesL = new ArrayList<List<JDDNode>>(conflictformulae);
			int count = 0;
			for(i=0; i<numTargets; i++)
				if (!reachExpr[i]) {
					tmpdra[count] = dra[i];
					tmpdraDDRowVars[count] = draDDRowVars[i];
					tmpdraDDColVars[count] = draDDColVars[i];
					// tricky part; double check when reachExpr is set for efficiency
					tmptargetDDs.add(targetDDs.get(count));
					tmpallstatesH.add(allstatesH.get(i));
					tmpallstatesL.add(allstatesL.get(i));
					count++;
				}
			//multitargetDDs = new ArrayList<JDDNode>();
			List<List<Integer>> tmpmultitargetIDs = new ArrayList<List<Integer>>();
			
			mcLtl.findMultiConflictAcceptingStates(tmpdra, modelProduct, 
					tmpdraDDRowVars, tmpdraDDColVars, tmptargetDDs, tmpallstatesH, tmpallstatesL,
					multitargetDDs, tmpmultitargetIDs);
			// again. double check when reachExpr is set
			count = 0;
			for(i=0; i<numTargets; i++)
				if (!reachExpr[i]) {
					targetDDs.remove(count);
					targetDDs.add(count, tmptargetDDs.get(count));
					count++;
				}
			
			// deal with reachability targets
			//multitargetIDs = new ArrayList<Integer>(tmpmultitargetIDs.size());
			for(i=0; i<tmpmultitargetIDs.size(); i++) {
				multitargetIDs.add(changeToInteger(tmpmultitargetIDs.get(i)));
			}
			
			for(i=0; i<numTargets; i++)
				if (!reachExpr[i]) {
					List<JDDNode> tmpLH = allstatesH.get(i);
					for(JDDNode n : tmpLH)
						JDD.Deref(n);
					tmpLH = allstatesL.get(i);
					for(JDDNode n : tmpLH)
						JDD.Deref(n);
				}
		}

		for (JDDNode ec : allecs)
			JDD.Deref(ec);
		if (sf != null) {
			for (int t = 0; t < sf.size(); t++)
				sf.get(t).derefHK();
			for (int t = 0; t < wf.size(); t++)
				wf.get(t).derefHK();
		}
	}
	// Contents of a P operator, i.e. a path formula

	protected StateValues checkProbPathFormula(Expression expr, boolean qual, boolean min) throws PrismException
	{
		// Test whether this is a simple path formula (i.e. PCTL)
		// and then pass control to appropriate method.
		// modified by Hongyang Qu
		// if an LTL fairness constraint is specified, then we force Prism to use checkProbPathFormulaLTL
		// Fixme: is there any conflict?
		if (expr.isSimplePathFormula() && model.getFairnessConstraints() == null) {
			return checkProbPathFormulaSimple(expr, qual, min);
		} else {
			return checkProbPathFormulaLTL(expr, qual, min);
			//throw new PrismException("LTL-style path formulas are not yet supported");
		}
	}

	// Simple path formula for P operator (one temporal op, possibly negated)

	protected StateValues checkProbPathFormulaSimple(Expression expr, boolean qual, boolean min) throws PrismException
	{
		StateValues probs = null;

		// Negation/parentheses
		if (expr instanceof ExpressionUnaryOp) {
			ExpressionUnaryOp exprUnary = (ExpressionUnaryOp) expr;
			// Parentheses
			if (exprUnary.getOperator() == ExpressionUnaryOp.PARENTH) {
				// Recurse
				probs = checkProbPathFormulaSimple(exprUnary.getOperand(), qual, min);
			}
			// Negation
			else if (exprUnary.getOperator() == ExpressionUnaryOp.NOT) {
				// Flip min/max, then subtract from 1
				probs = checkProbPathFormulaSimple(exprUnary.getOperand(), qual, !min);
				probs.subtractFromOne();
			}
		}
		// Temporal operators
		else if (expr instanceof ExpressionTemporal) {
			ExpressionTemporal exprTemp = (ExpressionTemporal) expr;
			// Next
			if (exprTemp.getOperator() == ExpressionTemporal.P_X) {
				probs = checkProbNext(exprTemp, min);
			}
			// Until
			else if (exprTemp.getOperator() == ExpressionTemporal.P_U) {
				if (exprTemp.hasBounds()) {
					probs = checkProbBoundedUntil(exprTemp, min);
				} else {
					probs = checkProbUntil(exprTemp, qual, min);
				}
			}
			// Anything else - convert to until and recurse
			else {
				if(exprTemp.getOperator() == ExpressionTemporal.P_F /*&& !min*/)
					sccBased = true;
				probs = checkProbPathFormulaSimple(exprTemp.convertToUntilForm(), qual, min);
			}
		}

		if (probs == null)
			throw new PrismException("Unrecognised path operator in P operator");

		return probs;
	}

	// LTL-like path formula for P operator

	protected StateValues checkProbPathFormulaLTL(Expression expr, boolean qual, boolean min) throws PrismException
	{
		LTLModelChecker mcLtl;
		StateValues probsProduct = null, probs = null;
		Expression ltl;
		Vector<JDDNode> labelDDs;
		DRA dra;
		NondetModel modelProduct;
		NondetModelChecker mcProduct;
		JDDNode startMask;
		JDDVars draDDRowVars, draDDColVars;
		int i;
		long l;

		// Can't do LTL with time-bounded variants of the temporal operators
		try {
			expr.accept(new ASTTraverse()
			{
				public void visitPre(ExpressionTemporal e) throws PrismLangException
				{
					if (e.getLowerBound() != null)
						throw new PrismLangException(e.getOperatorSymbol());
					if (e.getUpperBound() != null)
						throw new PrismLangException(e.getOperatorSymbol());
				}
			});
		} catch (PrismLangException e) {
			String s = "Temporal operators (like " + e.getMessage() + ")";
			s += " cannot have time bounds for LTL properties";
			throw new PrismException(s);
		}

		// For LTL model checking routines
		mcLtl = new LTLModelChecker(prism);

		// Model check maximal state formulas
		labelDDs = new Vector<JDDNode>();
		ltl = mcLtl.checkMaximalStateFormulas(this, model, expr.deepCopy(), labelDDs);

		// Convert LTL formula to deterministic Rabin automaton (DRA)
		// For min probabilities, need to negate the formula
		// (But check fairness setting since this may affect min/max)
		mainLog.println("\nBuilding deterministic Rabin automaton (for " + (min && !fairness ? "!" : "") + ltl + ")...");
		l = System.currentTimeMillis();
		if (min && !fairness) {
			dra = LTL2Rabin.ltl2rabin(ltl.convertForJltl2ba().negate());
		} else {
			dra = LTL2Rabin.ltl2rabin(ltl.convertForJltl2ba());
		}
		mainLog.println("\nDRA has " + dra.size() + " states, " + dra.acceptance().size() + " pairs.");
		/*try{
			mainLog.println("Printing dra...   ");
			dra.print(System.out);
			dra.printDot(new java.io.PrintStream("dra.dot"));
			mainLog.println("done...   ");
		} catch(Exception e) {}*/
		l = System.currentTimeMillis() - l;
		mainLog.println("\nTime for Rabin translation: " + l / 1000.0 + " seconds.");

		// Build product of MDP and automaton
		mainLog.println("\nConstructing MDP-DRA product...");
		draDDRowVars = new JDDVars();
		draDDColVars = new JDDVars();
		modelProduct = mcLtl.constructProductMDP(dra, model, labelDDs, draDDRowVars, draDDColVars);
		mainLog.println();
		modelProduct.printTransInfo(mainLog, prism.getExtraDDInfo());
		/*try {
			mainLog.println("Printing modelProduct...   ");
			prism.exportTransToFile(modelProduct, true, Prism.EXPORT_DOT_STATES, new java.io.File("product.dot"));
			//prism.exportStatesToFile(modelProduct, Prism.EXPORT_PLAIN, null);
			//prism.exportTransToFile(modelProduct, true, Prism.EXPORT_PLAIN, null);
			mainLog.println("done...   ");
		} catch (Exception e) {
		}*/

		// Find accepting maximum end components
		mainLog.println("\nFinding accepting end components...");
		/*
		 * convert fairness constraints into JDDs
		 */
		Vector<JDDFairnessConstraint> sf = null;
		Vector<JDDFairnessConstraint> wf = null;
		JDDNode acc;

		if (model.getFairnessConstraints() != null) {
			ExpressionFunc fe = (ExpressionFunc) model.getFairnessConstraints();
			sf = new Vector<JDDFairnessConstraint>();
			wf = new Vector<JDDFairnessConstraint>();
			for (int t = 0; t < fe.getNumOperands(); t++) {
				ExpressionFairness operand = (ExpressionFairness) fe.getOperand(t);
				JDDFairnessConstraint ddfc = new JDDFairnessConstraint(operand, modelProduct);
				if (operand.isStrongFairness())
					sf.add(ddfc);
				else
					wf.add(ddfc);
			}

			ArrayList<JDDNode> result = mcLtl.findFairAcceptingStates(model, dra, modelProduct, draDDRowVars, draDDColVars, fairness, sf, wf);

			acc = result.get(0);
			mainLog.println("\nComputing reachability probabilities...");
			mcProduct = new NondetModelChecker(prism, modelProduct, null);
			if (result.size() > 1) {
				mcProduct.trans = result.get(1);
				mcProduct.trans01 = result.get(2);
			}
			probsProduct = mcProduct.checkProbUntil(modelProduct.getReach(), acc, qual, min && fairness);
			if (result.size() > 1) {
				JDD.Deref(result.get(1));
				JDD.Deref(result.get(2));
			}
			/*
			 * Deref JDDs for fairness constraints
			 */
			for (int t = 0; t < sf.size(); t++)
				sf.get(t).derefHK();
			for (int t = 0; t < wf.size(); t++)
				wf.get(t).derefHK();

		} else {
			acc = mcLtl.findAcceptingStates(model, dra, modelProduct, draDDRowVars, draDDColVars, fairness);

			// Compute reachability probabilities
			mainLog.println("\nComputing reachability probabilities...");
			mcProduct = new NondetModelChecker(prism, modelProduct, null);
			probsProduct = mcProduct.checkProbUntil(modelProduct.getReach(), acc, qual, min && fairness);
		}

		// subtract from 1 if we're model checking a negated formula for regular Pmin
		if (min && !fairness) {
			probsProduct.subtractFromOne();
		}

		// Convert probability vector to original model
		// First, filter over DRA start states
		// (which we can get from initial states of product model,
		// because of the way it is constructed)
		startMask = modelProduct.getStart();
		probsProduct.filter(startMask);
		// Then sum over DD vars for the DRA state
		probs = probsProduct.sumOverDDVars(draDDRowVars, model);

		// Deref, clean up
		probsProduct.clear();
		modelProduct.clear();
		for (i = 0; i < labelDDs.size(); i++) {
			JDD.Deref(labelDDs.get(i));
		}
		JDD.Deref(acc);
		draDDRowVars.derefAll();
		draDDColVars.derefAll();

		return probs;
	}

	// next

	protected StateValues checkProbNext(ExpressionTemporal expr, boolean min) throws PrismException
	{
		JDDNode b;
		StateValues probs = null;

		// model check operand first
		b = checkExpressionDD(expr.getOperand2());

		// print out some info about num states
		// mainLog.print("\nb = " + JDD.GetNumMintermsString(b,
		// allDDRowVars.n()) + " states\n");

		// compute probabilities
		probs = computeNextProbs(trans, b, min);

		// derefs
		JDD.Deref(b);

		return probs;
	}

	// bounded until

	protected StateValues checkProbBoundedUntil(ExpressionTemporal expr, boolean min) throws PrismException
	{
		int time;
		JDDNode b1, b2;
		StateValues probs = null;

		// get info from bounded until
		time = expr.getUpperBound().evaluateInt(constantValues, null);
		if (expr.upperBoundIsStrict())
			time--;
		if (time < 0) {
			String bound = expr.upperBoundIsStrict() ? "<" + (time + 1) : "<=" + time;
			throw new PrismException("Invalid bound " + bound + " in bounded until formula");
		}

		// model check operands first
		b1 = checkExpressionDD(expr.getOperand1());
		try {
			b2 = checkExpressionDD(expr.getOperand2());
		} catch (PrismException e) {
			JDD.Deref(b1);
			throw e;
		}

		// print out some info about num states
		// mainLog.print("\nb1 = " + JDD.GetNumMintermsString(b1,
		// allDDRowVars.n()));
		// mainLog.print(" states, b2 = " + JDD.GetNumMintermsString(b2,
		// allDDRowVars.n()) + " states\n");

		// compute probabilities

		// a trivial case: "U<=0"
		if (time == 0) {
			// prob is 1 in b2 states, 0 otherwise
			JDD.Ref(b2);
			probs = new StateValuesMTBDD(b2, model);
		} else {
			try {
				probs = computeBoundedUntilProbs(trans, trans01, b1, b2, time, min);
			} catch (PrismException e) {
				JDD.Deref(b1);
				JDD.Deref(b2);
				throw e;
			}
		}

		// derefs
		JDD.Deref(b1);
		JDD.Deref(b2);

		return probs;
	}

	// until (unbounded)

	// this method is split into two steps so that the LTL model checker can use the second part directly

	protected StateValues checkProbUntil(ExpressionTemporal expr, boolean qual, boolean min) throws PrismException
	{
		JDDNode b1, b2;
		StateValues probs = null;

		// model check operands first
		b1 = checkExpressionDD(expr.getOperand1());
		try {
			b2 = checkExpressionDD(expr.getOperand2());
		} catch (PrismException e) {
			JDD.Deref(b1);
			throw e;
		}

		// print out some info about num states
		// mainLog.print("\nb1 = " + JDD.GetNumMintermsString(b1,
		// allDDRowVars.n()));
		// mainLog.print(" states, b2 = " + JDD.GetNumMintermsString(b2,
		// allDDRowVars.n()) + " states\n");

		try {
			probs = checkProbUntil(b1, b2, qual, min);
		} catch (PrismException e) {
			JDD.Deref(b1);
			JDD.Deref(b2);
			throw e;
		}

		// derefs
		JDD.Deref(b1);
		JDD.Deref(b2);

		return probs;
	}

	// until (unbounded): b1/b2 are bdds for until operands

	protected StateValues checkProbUntil(JDDNode b1, JDDNode b2, boolean qual, boolean min) throws PrismException
	{
		JDDNode splus, newb1, newb2;
		StateValues probs = null;
		long l;

		// compute probabilities

		// if doing min with fairness, we solve a different problem
		// (as in christel's habilitation)
		if (min && fairness) {

			// print out reminder that we have to do conversion for fairness
			mainLog.print("\nDoing conversion for fairness...\n");
			// start timer
			l = System.currentTimeMillis();
			// convert to new problem
			newb2 = PrismMTBDD.Prob0A(trans01, reach, allDDRowVars, allDDColVars, allDDNondetVars, b1, b2);
			JDD.Ref(newb2);
			splus = JDD.Not(newb2);
			JDD.Ref(splus);
			JDD.Ref(b2);
			newb1 = JDD.And(splus, JDD.Not(b2));
			JDD.Deref(splus);
			JDD.Ref(reach);
			newb1 = JDD.And(reach, newb1);
			JDD.Ref(reach);
			newb2 = JDD.And(reach, newb2);
			// stop timer
			l = System.currentTimeMillis() - l;
			// print out conversion info
			mainLog.print("\nTime for fairness conversion: " + l / 1000.0 + " seconds.\n");
			// mainLog.print("\nb1 = " + JDD.GetNumMintermsString(newb1, allDDRowVars.n()));
			// mainLog.print(" states, b2 = " + JDD.GetNumMintermsString(newb2, allDDRowVars.n()) + " states\n");
		} else {
			JDD.Ref(b1);
			newb1 = b1;
			JDD.Ref(b2);
			newb2 = b2;
		}

		// if requested (i.e. when prob bound is 0 or 1 and precomputation algorithms are enabled),
		// compute probabilities qualitatively
		if (qual) {
			mainLog.print("\nWarning: probability bound in formula is" + " 0/1 so exact probabilities may not be computed\n");
			// for fairness, we compute max here
			probs = computeUntilProbsQual(trans01, newb1, newb2, min && !fairness);
		}
		// otherwise actually compute probabilities
		else {
			// for fairness, we compute max here
			try {
				probs = computeUntilProbs(trans, transActions, trans01, newb1, newb2, min && !fairness);
			} catch (PrismException e) {
				JDD.Deref(newb1);
				JDD.Deref(newb2);
				throw e;
			}
		}

		// if we're doing min with fairness,
		// we need to subtract the probabilities from
		// one to get the actual answer
		if (min && fairness) {
			probs.subtractFromOne();
		}

		// Derefs
		JDD.Deref(newb1);
		JDD.Deref(newb2);

		return probs;
	}

	// inst reward

	protected StateValues checkRewardInst(ExpressionTemporal expr, JDDNode stateRewards, JDDNode transRewards, boolean min) throws PrismException
	{
		int time; // time bound
		StateValues rewards = null;

		// get info from bounded until
		time = expr.getUpperBound().evaluateInt(constantValues, null);
		if (time < 0) {
			throw new PrismException("Invalid bound " + time + " in instantaneous reward property");
		}

		// compute rewards
		rewards = computeInstRewards(trans, stateRewards, time, min);

		return rewards;
	}

	// reach reward

	protected StateValues checkRewardReach(ExpressionTemporal expr, JDDNode stateRewards, JDDNode transRewards, boolean min) throws PrismException
	{
		JDDNode b;
		StateValues rewards = null;

		// model check operand first
		b = checkExpressionDD(expr.getOperand2());

		// print out some info about num states
		// mainLog.print("\nb = " + JDD.GetNumMintermsString(b,
		// allDDRowVars.n()) + " states\n");

		// compute rewards
		try {
			rewards = computeReachRewards(trans, transActions, trans01, stateRewards, transRewards, b, min);
		} catch (PrismException e) {
			JDD.Deref(b);
			throw e;
		}

		// derefs
		JDD.Deref(b);

		return rewards;
	}

	// -----------------------------------------------------------------------------------
	// probability computation methods
	// -----------------------------------------------------------------------------------

	// compute probabilities for next

	protected StateValues computeNextProbs(JDDNode tr, JDDNode b, boolean min)
	{
		JDDNode tmp;
		StateValues probs = null;

		// matrix multiply: trans * b
		JDD.Ref(b);
		tmp = JDD.PermuteVariables(b, allDDRowVars, allDDColVars);
		JDD.Ref(tr);
		tmp = JDD.MatrixMultiply(tr, tmp, allDDColVars, JDD.BOULDER);
		// (then min or max)
		if (min) {
			// min
			JDD.Ref(nondetMask);
			tmp = JDD.Apply(JDD.MAX, tmp, nondetMask);
			tmp = JDD.MinAbstract(tmp, allDDNondetVars);
		} else {
			// max
			tmp = JDD.MaxAbstract(tmp, allDDNondetVars);
		}
		probs = new StateValuesMTBDD(tmp, model);

		return probs;
	}

	// compute probabilities for bounded until

	protected StateValues computeBoundedUntilProbs(JDDNode tr, JDDNode tr01, JDDNode b1, JDDNode b2, int time, boolean min) throws PrismException
	{
		JDDNode yes, no, maybe;
		JDDNode probsMTBDD;
		DoubleVector probsDV;
		StateValues probs = null;

		// compute yes/no/maybe states
		if (b2.equals(JDD.ZERO)) {
			yes = JDD.Constant(0);
			JDD.Ref(reach);
			no = reach;
			maybe = JDD.Constant(0);
		} else if (b1.equals(JDD.ZERO)) {
			JDD.Ref(b2);
			yes = b2;
			JDD.Ref(reach);
			JDD.Ref(b2);
			no = JDD.And(reach, JDD.Not(b2));
			maybe = JDD.Constant(0);
		} else {
			// yes
			JDD.Ref(b2);
			yes = b2;
			// no
			if (yes.equals(reach)) {
				no = JDD.Constant(0);
			} else if (precomp && prob0) {
				if (min) {
					// "min prob = 0" equates to "there exists a prob 0"
					no = PrismMTBDD.Prob0E(tr01, reach, nondetMask, allDDRowVars, allDDColVars, allDDNondetVars, b1, yes);
				} else {
					// "max prob = 0" equates to "all probs 0"
					no = PrismMTBDD.Prob0A(tr01, reach, allDDRowVars, allDDColVars, allDDNondetVars, b1, yes);
				}
			} else {
				JDD.Ref(reach);
				JDD.Ref(b1);
				JDD.Ref(b2);
				no = JDD.And(reach, JDD.Not(JDD.Or(b1, b2)));
			}
			// maybe
			JDD.Ref(reach);
			JDD.Ref(yes);
			JDD.Ref(no);
			maybe = JDD.And(reach, JDD.Not(JDD.Or(yes, no)));
		}

		// print out yes/no/maybe
		mainLog.print("\nyes = " + JDD.GetNumMintermsString(yes, allDDRowVars.n()));
		mainLog.print(", no = " + JDD.GetNumMintermsString(no, allDDRowVars.n()));
		mainLog.print(", maybe = " + JDD.GetNumMintermsString(maybe, allDDRowVars.n()) + "\n");

		// if maybe is empty, we have the probabilities already
		if (maybe.equals(JDD.ZERO)) {
			JDD.Ref(yes);
			probs = new StateValuesMTBDD(yes, model);
		}
		// otherwise explicitly compute the remaining probabilities
		else {
			// compute probabilities
			try {
				switch (engine) {
				case Prism.MTBDD:
					probsMTBDD = PrismMTBDD.NondetBoundedUntil(tr, odd, nondetMask, allDDRowVars, allDDColVars, allDDNondetVars, yes, maybe, time, min);
					probs = new StateValuesMTBDD(probsMTBDD, model);
					break;
				case Prism.SPARSE:
					probsDV = PrismSparse.NondetBoundedUntil(tr, odd, allDDRowVars, allDDColVars, allDDNondetVars, yes, maybe, time, min);
					probs = new StateValuesDV(probsDV, model);
					break;
				case Prism.HYBRID:
					probsDV = PrismHybrid.NondetBoundedUntil(tr, odd, allDDRowVars, allDDColVars, allDDNondetVars, yes, maybe, time, min);
					probs = new StateValuesDV(probsDV, model);
					break;
				default:
					throw new PrismException("Unknown engine");
				}
			} catch (PrismException e) {
				JDD.Deref(yes);
				JDD.Deref(no);
				JDD.Deref(maybe);
				throw e;
			}
		}

		// derefs
		JDD.Deref(yes);
		JDD.Deref(no);
		JDD.Deref(maybe);

		return probs;
	}

	// compute probabilities for until (for qualitative properties)

	// note: this function doesn't need to know anything about fairness
	// it is just told whether to compute min or max probabilities

	protected StateValues computeUntilProbsQual(JDDNode tr01, JDDNode b1, JDDNode b2, boolean min)
	{
		JDDNode yes = null, no = null, maybe;
		StateValues probs = null;

		// note: we know precomputation is enabled else this function wouldn't
		// have been called

		// compute yes/no/maybe states
		if (b2.equals(JDD.ZERO)) {
			yes = JDD.Constant(0);
			JDD.Ref(reach);
			no = reach;
			maybe = JDD.Constant(0);
		} else if (b1.equals(JDD.ZERO)) {
			JDD.Ref(b2);
			yes = b2;
			JDD.Ref(reach);
			JDD.Ref(b2);
			no = JDD.And(reach, JDD.Not(b2));
			maybe = JDD.Constant(0);
		} else {
			// min
			if (min) {
				// no: "min prob = 0" equates to "there exists an adversary prob equals 0"
				no = PrismMTBDD.Prob0E(tr01, reach, nondetMask, allDDRowVars, allDDColVars, allDDNondetVars, b1, b2);
				// yes: "min prob = 1" equates to "for all adversaries prob equals 1"
				yes = PrismMTBDD.Prob1A(tr01, reach, nondetMask, allDDRowVars, allDDColVars, allDDNondetVars, no, b2);
			}
			// max
			else {
				// no: "max prob = 0" equates to "for all adversaries prob equals 0"
				no = PrismMTBDD.Prob0A(tr01, reach, allDDRowVars, allDDColVars, allDDNondetVars, b1, b2);
				// yes: "max prob = 1" equates to "there exists an adversary prob equals 1"
				yes = PrismMTBDD.Prob1E(tr01, reach, allDDRowVars, allDDColVars, allDDNondetVars, b1, b2, no);
			}
			// maybe
			JDD.Ref(reach);
			JDD.Ref(yes);
			JDD.Ref(no);
			maybe = JDD.And(reach, JDD.Not(JDD.Or(yes, no)));
		}

		// print out yes/no/maybe
		mainLog.print("\nyes = " + JDD.GetNumMintermsString(yes, allDDRowVars.n()));
		mainLog.print(", no = " + JDD.GetNumMintermsString(no, allDDRowVars.n()));
		mainLog.print(", maybe = " + JDD.GetNumMintermsString(maybe, allDDRowVars.n()) + "\n");

		// if maybe is empty, we have the answer already...
		if (maybe.equals(JDD.ZERO)) {
			JDD.Ref(yes);
			probs = new StateValuesMTBDD(yes, model);
		}
		// otherwise we set the probabilities for maybe states to be 0.5
		// (actual probabilities for these states are unknown but definitely >0
		// and <1)
		// (this is safe because the results of this function will only be used
		// to compare against 0/1 bounds)
		// (this is not entirely elegant but is simpler and less error prone
		// than
		// trying to work out whether to use 0/1 for all case of min/max,
		// fairness, future/global, etc.)
		else {
			JDD.Ref(yes);
			JDD.Ref(maybe);
			probs = new StateValuesMTBDD(JDD.Apply(JDD.PLUS, yes, JDD.Apply(JDD.TIMES, maybe, JDD.Constant(0.5))), model);
		}

		// derefs
		JDD.Deref(yes);
		JDD.Deref(no);
		JDD.Deref(maybe);

		return probs;
	}

	// compute probabilities for until (general case)

	// note: this function doesn't need to know anything about fairness
	// it is just told whether to compute min or max probabilities

	protected StateValues computeUntilProbs(JDDNode tr, JDDNode tra, JDDNode tr01, JDDNode b1, JDDNode b2, boolean min) throws PrismException
	{
		JDDNode yes, no, maybe;
		JDDNode probsMTBDD;
		DoubleVector probsDV;
		StateValues probs = null;

		// If required, export info about target states 
		if (prism.getExportTarget()) {
			JDDNode labels[] = { model.getStart(), b2 };
			String labelNames[] = { "init", "target" };
			try {
				mainLog.println("\nExporting target states info to file \"" + prism.getExportTargetFilename() + "\"...");
				PrismMTBDD.ExportLabels(labels, labelNames, "l", model.getAllDDRowVars(), model.getODD(), Prism.EXPORT_PLAIN, prism.getExportTargetFilename());
			} catch (FileNotFoundException e) {
				mainLog.println("\nWarning: Could not export target to file \"" + prism.getExportTargetFilename() + "\"");
			}
		}

		if(prism.getSCCMethod() == Prism.SCCFIND)
			sccBased = false;
		// compute yes/no/maybe states
		if (b2.equals(JDD.ZERO)) {
			yes = JDD.Constant(0);
			JDD.Ref(reach);
			no = reach;
			maybe = JDD.Constant(0);
		} else if (b1.equals(JDD.ZERO)) {
			JDD.Ref(b2);
			yes = b2;
			JDD.Ref(reach);
			JDD.Ref(b2);
			no = JDD.And(reach, JDD.Not(b2));
			maybe = JDD.Constant(0);
		} else {
			// no
			// if precomputation enabled
			if (precomp && prob0) {
				// min
				if (min) {
					// no: "min prob = 0" equates to "there exists an adversary prob equals 0"
					no = PrismMTBDD.Prob0E(tr01, reach, nondetMask, allDDRowVars, allDDColVars, allDDNondetVars, b1, b2);
				}
				// max
				else {
					// no: "max prob = 0" equates to "for all adversaries prob equals 0"
					no = PrismMTBDD.Prob0A(tr01, reach, allDDRowVars, allDDColVars, allDDNondetVars, b1, b2);
				}
			}
			// if precomputation not enabled
			else {
				// no
				JDD.Ref(reach);
				JDD.Ref(b1);
				JDD.Ref(b2);
				no = JDD.And(reach, JDD.Not(JDD.Or(b1, b2)));
			}
			// yes
			// if precomputation enabled (need both prob0/prob1 to be enabled to do prob1)
			if (precomp && prob0 && prob1) {
				// min
				if (min) {
					// yes: "min prob = 1" equates to "for all adversaries prob equals 1"
					yes = PrismMTBDD.Prob1A(tr01, reach, nondetMask, allDDRowVars, allDDColVars, allDDNondetVars, no, b2);
				}
				// max
				else {
					// yes: "max prob = 1" equates to "there exists an adversary prob equals 1"
					yes = PrismMTBDD.Prob1E(tr01, reach, allDDRowVars, allDDColVars, allDDNondetVars, b1, b2, no);
				}
			}
			// if precomputation not enabled
			else {
				// yes
				JDD.Ref(b2);
				yes = b2;
			}
			// maybe
			JDD.Ref(reach);
			JDD.Ref(yes);
			JDD.Ref(no);
			maybe = JDD.And(reach, JDD.Not(JDD.Or(yes, no)));
		}

		// print out yes/no/maybe
		mainLog.print("\nyes = " + JDD.GetNumMintermsString(yes, allDDRowVars.n()));
		mainLog.print(", no = " + JDD.GetNumMintermsString(no, allDDRowVars.n()));
		mainLog.print(", maybe = " + JDD.GetNumMintermsString(maybe, allDDRowVars.n()) + "\n");

		// if maybe is empty, we have the answer already...
		if (maybe.equals(JDD.ZERO)) {
			JDD.Ref(yes);
			probs = new StateValuesMTBDD(yes, model);
		}
		// otherwise we compute the actual probabilities
		else {
			// compute probabilities
			mainLog.println("\nComputing remaining probabilities...");

			try {
				switch (engine) {
				case Prism.MTBDD:
					probsMTBDD = sccBased ? 
							PrismMTBDD.SCCBasedNondetUntil(tr, trans01, odd, nondetMask, model.getStart(), 
									allDDRowVars, allDDColVars, allDDNondetVars, yes, maybe, min, 
									prism.getSCCMethod() == Prism.XIEBEEREL) :
							PrismMTBDD.NondetUntil(tr, odd, nondetMask, allDDRowVars, allDDColVars,
							allDDNondetVars, yes, maybe, min);
					probs = new StateValuesMTBDD(probsMTBDD, model);
					break;
				case Prism.SPARSE:
					probsDV = PrismSparse.NondetUntil(tr, tra, model.getSynchs(), odd, allDDRowVars, allDDColVars, allDDNondetVars, yes, maybe, min);
					probs = new StateValuesDV(probsDV, model);
					break;
				case Prism.HYBRID:
					probsDV = PrismHybrid.NondetUntil(tr, odd, allDDRowVars, allDDColVars, allDDNondetVars, yes, maybe, min);
					probs = new StateValuesDV(probsDV, model);
					break;
				default:
					throw new PrismException("Unknown engine");
				}
			} catch (PrismException e) {
				JDD.Deref(yes);
				JDD.Deref(no);
				JDD.Deref(maybe);
				throw e;
			}
		}

		// derefs
		JDD.Deref(yes);
		JDD.Deref(no);
		JDD.Deref(maybe);

		return probs;
	}

	// compute (max) probabilities for multi-objective reachability

	protected double computeMultiReachProbs(NondetModel modelProduct, LTLModelChecker mcLtl, List<JDDNode> rewards, JDDNode st, 
			List<JDDNode> targets, List<JDDNode> combinations, List<Integer> combinationIDs, List<Integer> relOpsProb, List<Double> boundsProb, 
			List<Integer> relOpsReward, List<Double> boundsReward, boolean hasconflictobjectives) throws PrismException
	{
		JDDNode yes, no, maybe, bottomec = null;
		double prob;
		int i, j, n;
		
		//JDDNode maybe_r = null; // maybe states for the reward formula
		//JDDNode trr = null; // transition rewards
		//int op1 = relOps.get(0).intValue(); // the operator of the first objective query

		n = targets.size();

		// If required, export info about target states 
		if (prism.getExportTarget()) {
			JDDNode labels[] = new JDDNode[1 + n];
			String labelNames[] = new String[1 + n];
			labels[0] = modelProduct.getStart();
			labelNames[0] = "init";
			// Build temporary DDs for combined targets
			for (i = 0; i < n; i++) {
				JDD.Ref(targets.get(i));
				JDDNode tmp = targets.get(i);
				for (j = 0; j < combinations.size(); j++) {
					if ((combinationIDs.get(j) & (1 << i)) > 0) {
						JDD.Ref(combinations.get(j));
						tmp = JDD.Or(tmp, combinations.get(j));
					}
				}
				labels[i + 1] = tmp;
				labelNames[i + 1] = "target" + (i + 1);
			}
			try {
				mainLog.print("\nExporting target states info to file \"" + prism.getExportTargetFilename() + "\"...");
				PrismMTBDD.ExportLabels(labels, labelNames, "l", modelProduct.getAllDDRowVars(), modelProduct.getODD(), Prism.EXPORT_PLAIN, prism
						.getExportTargetFilename());
			} catch (FileNotFoundException e) {
				mainLog.println("\nWarning: Could not export target to file \"" + prism.getExportTargetFilename() + "\"");
			}
			// Deref temporary target DDs created above
			for (i = 0; i < n; i++) {
				JDD.Deref(labels[i + 1]);
			}
		}

		// yes - union of targets (just to compute no)
		yes = JDD.Constant(0);
		//n = targets.size();
		for (i = 0; i < n; i++) {
			JDD.Ref(targets.get(i));
			yes = JDD.Or(yes, targets.get(i));
		}
		if(combinations != null)
			for (i = 0; i < combinations.size(); i++) {
				JDD.Ref(combinations.get(i));
				yes = JDD.Or(yes, combinations.get(i));
			}

		if(relOpsReward.size() == 0)
			no = PrismMTBDD.Prob0A(modelProduct.getTrans01(), modelProduct.getReach(), modelProduct.getAllDDRowVars(), modelProduct.getAllDDColVars(), modelProduct
					.getAllDDNondetVars(), modelProduct.getReach(), yes);
		else {
			no = JDD.Constant(0);
			bottomec = PrismMTBDD.Prob0A(modelProduct.getTrans01(), modelProduct.getReach(), modelProduct.getAllDDRowVars(), modelProduct.getAllDDColVars(), 
					modelProduct.getAllDDNondetVars(), modelProduct.getReach(), yes);
			Vector<JDDNode> becs = mcLtl.findBottomEndComponents(modelProduct, bottomec);
			bottomec = JDD.Constant(0);
			for(JDDNode ec : becs)
				bottomec = JDD.Or(bottomec, ec);
		}

		/*if(op1>2) { // the first query is about reward
			JDDNode no_r = PrismMTBDD.Prob0A(modelProduct.getTrans01(), modelProduct.getReach(), 
					modelProduct.getAllDDRowVars(), modelProduct.getAllDDColVars(), 
					modelProduct.getAllDDNondetVars(), modelProduct.getReach(), targets.get(0));
			JDD.Ref(no_r);
			maybe_r = JDD.And(modelProduct.getReach(), JDD.Not(JDD.Or(targets.get(0), no_r)));
			trr = modelProduct.getTransRewards();
		}*/

		// maybe
		JDD.Ref(modelProduct.getReach());
		JDD.Ref(yes);
		JDD.Ref(no);
		maybe = JDD.And(modelProduct.getReach(), JDD.Not(JDD.Or(yes, no)));
		

		for(i = 0; i< rewards.size(); i++) {
			JDDNode tmp = rewards.remove(i);
			JDD.Ref(no);
			tmp = JDD.Apply(JDD.TIMES, tmp, JDD.Not(no));
			rewards.add(i, tmp);
		}
		
		// print out yes/no/maybe
		mainLog.print("\nyes = " + JDD.GetNumMintermsString(yes, modelProduct.getAllDDRowVars().n()));
		mainLog.print(", no = " + JDD.GetNumMintermsString(no, modelProduct.getAllDDRowVars().n()));
		mainLog.print(", maybe = " + JDD.GetNumMintermsString(maybe, modelProduct.getAllDDRowVars().n()) + "\n");

		// compute probabilities
		mainLog.println("\nComputing remaining probabilities...");
		//modelProduct.printTransInfo(mainLog, true);
		try {
			switch (engine) {
			case Prism.SPARSE:
				prob = relOpsReward.size() > 0 ? (
						(hasconflictobjectives)?
								PrismSparse.NondetMultiReachReward1(modelProduct.getTrans(), modelProduct.getTransActions(), modelProduct.getSynchs(), modelProduct.getODD(),
										modelProduct.getAllDDRowVars(), modelProduct.getAllDDColVars(), modelProduct.getAllDDNondetVars(), targets, combinations, combinationIDs, relOpsProb, boundsProb, 
										relOpsReward, boundsReward, maybe, st, rewards, bottomec) :
											PrismSparse.NondetMultiReachReward(modelProduct.getTrans(), modelProduct.getTransActions(), modelProduct.getSynchs(), modelProduct.getODD(),
													modelProduct.getAllDDRowVars(), modelProduct.getAllDDColVars(), modelProduct.getAllDDNondetVars(), targets, relOpsProb, boundsProb, 
						relOpsReward, boundsReward, maybe, st, rewards, bottomec) ):
							((hasconflictobjectives) ?
									PrismSparse.NondetMultiReach1(modelProduct.getTrans(), modelProduct.getTransActions(), modelProduct.getSynchs(), modelProduct.getODD(),
											modelProduct.getAllDDRowVars(), modelProduct.getAllDDColVars(), modelProduct.getAllDDNondetVars(), targets, combinations, combinationIDs, relOpsProb, boundsProb, 
											maybe, st) :
												PrismSparse.NondetMultiReach(modelProduct.getTrans(), modelProduct.getTransActions(), modelProduct.getSynchs(), modelProduct.getODD(),
														modelProduct.getAllDDRowVars(), modelProduct.getAllDDColVars(), modelProduct.getAllDDNondetVars(), targets, relOpsProb, boundsProb, 
														maybe, st));
				break;
			default:
				throw new PrismException("Currently only sparse engine supports multi-objective properties");
			}
		} catch (PrismException e) {
			JDD.Deref(yes);
			JDD.Deref(no);
			JDD.Deref(maybe);
			throw e;
		}

		// derefs
		JDD.Deref(yes);
		JDD.Deref(no);
		JDD.Deref(maybe);

		return prob;
	}

	// compute rewards for inst reward

	protected StateValues computeInstRewards(JDDNode tr, JDDNode sr, int time, boolean min) throws PrismException
	{
		JDDNode rewardsMTBDD;
		DoubleVector rewardsDV;
		StateValues rewards = null;

		// a trivial case: "=0"
		if (time == 0) {
			JDD.Ref(sr);
			rewards = new StateValuesMTBDD(sr, model);
		}
		// otherwise we compute the actual rewards
		else {
			// compute the rewards
			try {
				switch (engine) {
				case Prism.MTBDD:
					rewardsMTBDD = PrismMTBDD.NondetInstReward(tr, sr, odd, nondetMask, allDDRowVars, allDDColVars, allDDNondetVars, time, min, start);
					rewards = new StateValuesMTBDD(rewardsMTBDD, model);
					break;
				case Prism.SPARSE:
					rewardsDV = PrismSparse.NondetInstReward(tr, sr, odd, allDDRowVars, allDDColVars, allDDNondetVars, time, min, start);
					rewards = new StateValuesDV(rewardsDV, model);
					break;
				case Prism.HYBRID:
					throw new PrismException("Hybrid engine does not yet supported this type of property (use sparse or MTBDD engine instead)");
				default:
					throw new PrismException("Unknown engine");
				}
			} catch (PrismException e) {
				throw e;
			}
		}

		return rewards;
	}

	// compute rewards for reach reward

	protected StateValues computeReachRewards(JDDNode tr, JDDNode tra, JDDNode tr01, JDDNode sr, JDDNode trr, JDDNode b, boolean min) throws PrismException
	{
		JDDNode inf, maybe, prob1, no;
		JDDNode rewardsMTBDD;
		DoubleVector rewardsDV;
		StateValues rewards = null;

		Vector<JDDNode> zeroCostEndComponents = null;

		if(prism.getSCCMethod() == Prism.SCCFIND)
			sccBased = false;
		// compute states which can't reach goal with probability 1
		if (b.equals(JDD.ZERO)) {
			JDD.Ref(reach);
			inf = reach;
			maybe = JDD.Constant(0);
		} else if (b.equals(reach)) {
			inf = JDD.Constant(0);
			maybe = JDD.Constant(0);
		} else {
			if (!min) {
				// compute states for which some adversaries don't reach goal with probability 1
				// note that prob1a (unlike prob1e) requires no/b2, not b1/b2
				// hence we have to call prob0e first
				no = PrismMTBDD.Prob0E(tr01, reach, nondetMask, allDDRowVars, allDDColVars, allDDNondetVars, reach, b);
				prob1 = PrismMTBDD.Prob1A(tr01, reach, nondetMask, allDDRowVars, allDDColVars, allDDNondetVars, no, b);
				JDD.Deref(no);
				JDD.Ref(reach);
				inf = JDD.And(reach, JDD.Not(prob1));
			} else {

				if (prism.getCheckZeroLoops()) {
					// find states transitions that have no cost
					JDD.Ref(sr);
					JDD.Ref(reach);
					JDDNode zeroReach = JDD.And(reach, JDD.Apply(JDD.EQUALS, sr, JDD.Constant(0)));
					JDD.Ref(b);
					zeroReach = JDD.And(zeroReach, JDD.Not(b));
					JDD.Ref(trr);
					JDDNode zeroTrr = JDD.Apply(JDD.EQUALS, trr, JDD.Constant(0));
					JDD.Ref(trans);
					JDD.Ref(zeroTrr);
					JDDNode zeroTrans = JDD.And(trans, zeroTrr);
					JDD.Ref(trans01);
					JDDNode zeroTrans01 = JDD.And(trans01, zeroTrr);

					ECComputer ecComp = new ECComputerDefault(prism, zeroReach, zeroTrans, zeroTrans01, model.getAllDDRowVars(), model.getAllDDColVars(), model
							.getAllDDNondetVars());
					ecComp.computeECs();

					zeroCostEndComponents = ecComp.getVectECs();

					JDD.Deref(zeroReach);
					JDD.Deref(zeroTrans);
					JDD.Deref(zeroTrans01);
				}

				// compute states for which all adversaries don't reach goal with probability 1
				no = PrismMTBDD.Prob0A(tr01, reach, allDDRowVars, allDDColVars, allDDNondetVars, reach, b);
				prob1 = PrismMTBDD.Prob1E(tr01, reach, allDDRowVars, allDDColVars, allDDNondetVars, reach, b, no);
				JDD.Deref(no);
				JDD.Ref(reach);
				inf = JDD.And(reach, JDD.Not(prob1));
			}
			JDD.Ref(reach);
			JDD.Ref(inf);
			JDD.Ref(b);
			maybe = JDD.And(reach, JDD.Not(JDD.Or(inf, b)));
		}

		if (prism.getCheckZeroLoops()) {
			// need to deal with zero loops yet
			if (min && zeroCostEndComponents != null && zeroCostEndComponents.size() > 0) {
				mainLog.println("\nWarning: PRISM detected your model contains " + zeroCostEndComponents.size() + " zero-reward "
						+ ((zeroCostEndComponents.size() == 1) ? "loop." : "loops."));
				mainLog.println("Your minimum rewards may be too low...");
			}
		} else if (min) {
			mainLog.println("\nWarning: PRISM hasn't checked for zero-reward loops.");
			mainLog.println("Your minimum rewards may be too low...");
		}

		// print out yes/no/maybe
		mainLog.print("\ngoal = " + JDD.GetNumMintermsString(b, allDDRowVars.n()));
		mainLog.print(", inf = " + JDD.GetNumMintermsString(inf, allDDRowVars.n()));
		mainLog.print(", maybe = " + JDD.GetNumMintermsString(maybe, allDDRowVars.n()) + "\n");

		// if maybe is empty, we have the rewards already
		if (maybe.equals(JDD.ZERO)) {
			JDD.Ref(inf);
			rewards = new StateValuesMTBDD(JDD.ITE(inf, JDD.PlusInfinity(), JDD.Constant(0)), model);
		}
		// otherwise we compute the actual rewards
		else {
			// compute the rewards
			try {
				switch (engine) {
				case Prism.MTBDD:
					rewardsMTBDD = sccBased? 
							PrismMTBDD.SCCBasedNondetReachReward(tr, trans01, sr, trr, odd, nondetMask, model.getStart(), 
									allDDRowVars, allDDColVars, allDDNondetVars, b, inf, maybe, min) :
							PrismMTBDD.NondetReachReward(tr, sr, trr, odd, nondetMask, allDDRowVars,
							allDDColVars, allDDNondetVars, b, inf, maybe, min);
					rewards = new StateValuesMTBDD(rewardsMTBDD, model);
					break;
				case Prism.SPARSE:
					rewardsDV = PrismSparse.NondetReachReward(tr, tra, model.getSynchs(), sr, trr, odd, allDDRowVars, allDDColVars, allDDNondetVars, b, inf, maybe, min);
					rewards = new StateValuesDV(rewardsDV, model);
					break;
				case Prism.HYBRID:
					throw new PrismException("Hybrid engine does not yet supported this type of property (use sparse or MTBDD engine instead)");
					// rewardsDV = PrismHybrid.NondetReachReward(tr, sr, trr,
					// odd, allDDRowVars, allDDColVars, allDDNondetVars, b, inf,
					// maybe, min);
					// rewards = new StateValuesDV(rewardsDV, model);
					// break;
				default:
					throw new PrismException("Unknown engine");
				}
			} catch (PrismException e) {
				JDD.Deref(inf);
				JDD.Deref(maybe);
				throw e;
			}
		}

		if (zeroCostEndComponents != null)
			for (JDDNode zcec : zeroCostEndComponents)
				JDD.Deref(zcec);

		// derefs
		JDD.Deref(inf);
		JDD.Deref(maybe);

		return rewards;
	}

	/**
	 * Check whether state set represented by dd is "weak absorbing",
	 * i.e. whether set is closed under all successors.
	 * Note: does not deref dd.
	 */
	private boolean checkWeakAbsorbing(JDDNode dd, NondetModel model)
	{
		boolean result;

		JDD.Ref(model.getTrans01());
		JDD.Ref(dd);
		JDDNode tmp = JDD.And(model.getTrans01(), dd);
		tmp = JDD.SwapVariables(tmp, model.getAllDDRowVars(), model.getAllDDColVars());
		tmp = JDD.ThereExists(tmp, model.getAllDDNondetVars());
		tmp = JDD.ThereExists(tmp, model.getAllDDColVars());
		JDD.Ref(model.getReach());
		tmp = JDD.And(model.getReach(), tmp);
		JDD.Ref(dd);
		tmp = JDD.And(tmp, JDD.Not(dd));
		result = tmp.equals(JDD.ZERO);
		JDD.Deref(tmp);

		return result;
	}
	
	private int changeToInteger(List<Integer> ids) {
		int k = 0;
		for(int i=0; i<ids.size(); i++) {
			int j = 1;
			if(ids.get(i) > 0)
				j = j << ids.get(i);
			k += j;
		}
		return k;
	}
	
}

// ------------------------------------------------------------------------------
